/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.bridge;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * Basic message to handle communications between C++ and Java programs
 * 
 * @author K Ericson
 * 
 */

public class JavaByteMessage {
    private int controlMessageType; // 4 bytes
    private byte[] input; // 4 bytes for length, then actual byte array
    private byte[] output; // same as above
    private String streamIdentifier; // 4 bytes for length, then actual
                                     // characters - each character is 1 byte
    private String description; // same as above
    private boolean isFromJava; // 1 byte
    private ByteArrayOutputStream os; // not being sent
    private DataOutputStream ds; // not being sent
    // these 2 needed for going back from the byte array to remake the class
    private ByteArrayInputStream is;
    private DataInputStream dis;


    /***
     * Initializes a <code>JavaCppMessage</code>, message type is set at
     * creation, and cannot be changed.
     * 
     * @param control
     *            Type of message as specified in <code>ControlMessage</code>
     */
    public JavaByteMessage(int control) // force control to be set at creation,
                                        // and don't allow it to change
    {
        controlMessageType = control;
        os = new ByteArrayOutputStream();
        ds = new DataOutputStream(os);
        isFromJava = true;
        input = new byte[0];
        output = new byte[0];
        streamIdentifier = "";
        description = "";
    }


    /***
     * Initializes a <code>JavaCppMessage</code>, based off of a
     * <code>byte[]</code>, used to reconstruct a message sent from a c++
     * program
     * 
     * @param incoming
     *            The <code>byte[]</code> recieved from an external program
     */
    public JavaByteMessage(byte[] incoming) {
        is = new ByteArrayInputStream(incoming);
        dis = new DataInputStream(is);

        os = new ByteArrayOutputStream();
        ds = new DataOutputStream(os);

        try {
            controlMessageType = dis.readInt();
            int read;
            int inputSize = dis.readInt();
            if (inputSize > 0) {
                input = new byte[inputSize];
                read = dis.read(input); // should do some check of read
            } else
                input = new byte[0];
            int outputSize = dis.readInt();
            output = new byte[outputSize];
            read = dis.read(output); // again, should check read
            int streamLength = dis.readInt(); // length of streamIdentifier
            //System.out.println("Stream length: " + streamLength);
            byte[] stream = new byte[streamLength];
            read = dis.read(stream); // again, should check
            streamIdentifier = new String(stream);
            int descriptionLength = dis.readInt(); // length of description
            byte[] desc = new byte[descriptionLength];
            read = dis.read(desc); // again with the checking
            //System.out.println("descriptionLength: " + descriptionLength);
            description = new String(desc);
            //System.out.println("description: " + description);
            isFromJava = dis.readBoolean();

        } catch (IOException e) {
            System.out.println("Problem Decoding Message: " + e.getMessage());
            e.printStackTrace();
        }

    }


    // Getters and setters
    public int getCONTROL() {
        return controlMessageType;
    }


    public void setInput(byte[] input) {
        this.input = input;
    }


    public byte[] getInput() {
        return input;
    }


    public void setOutput(byte[] output) {
        this.output = output;
    }


    public byte[] getOutput() {
        return output;
    }


    public void setStreamIdentifier(String streamIdentifier) {
        this.streamIdentifier = streamIdentifier;
    }


    public String getStreamIdentifier() {
        return streamIdentifier;
    }


    public void setDescription(String description) {
        this.description = description;
    }


    public String getDescription() {
        return description;
    }


    public void setFromJava(boolean isFromJava) {
        this.isFromJava = isFromJava;
    }


    public boolean isFromJava() {
        return isFromJava;
    }


    // done with accessors/setters

    /***
     * Function to convert the JavaCppMessage to a <code>byte[]</code> to be
     * sent to another program
     * 
     * @return byte[] A <code>byte[]</code> representation of this
     *         JavaCppMessage
     */
    public byte[] toByteArray() {
        try {
            ds.writeInt(controlMessageType); // 4 bytes for type
            ds.writeInt(input.length); // 4 bytes for size of input
            ds.write(input, 0, input.length); // ? bytes for input
            ds.writeInt(output.length); // 4 bytes for size of output
            ds.write(output, 0, output.length); // ? bytes for output
            byte[] t = streamIdentifier.getBytes();
            ds.writeInt(t.length); // 4 bytes for size of streamID
            ds.writeBytes(streamIdentifier); // ? bytes for streamID
            byte[] t2 = description.getBytes();
            ds.writeInt(t2.length); // 4 bytes for size of description
            ds.writeBytes(description); // ? bytes for description
            ds.writeBoolean(isFromJava); // 1 byte for isFromJava
            // ds.writeChar('\n'); //end char for c++?
            ds.flush();
            os.flush();
        } catch (IOException e) {
            System.out.println("Problem writing to byte stream");
            e.printStackTrace();
        }

        byte[] toRet = os.toByteArray();
        os.reset();

        return toRet;
    }


    /***
     * Helper to view the contents of a JavaCppMessage - useful for debugging
     * 
     * @return String String representation of the JavaCppMessage.
     */
    public String toString() {
        String val = "";

        val += "message type: " + controlMessageType + "\n";
        val += "input length: " + input.length + "\n";
        val += "output length: " + output.length + "\n";
        val += "stream ID: " + streamIdentifier + "\n";
        val += "description: " + description + "\n";
        val += "from Java? " + isFromJava;

        return val;
    }

}
