/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.bridge;

/**
 * Generic Computation for bridges.  Should be flexible enough to handle any language.
 * Works independently from the adaptive code.  Keeps track of sockets when necessary.
 * 
 * @author ericson
 */

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;

import ds.granules.bridge.ControlMessage;
import ds.granules.bridge.JavaByteMessage;

public class JavaGenericComputation {
    Process p;
    String programName;
    Runtime r;
    StreamReader sr;
    StreamWriter sw;
    StreamStringReader se;
    ServerSocket ss;
    Socket prog;
    JavaMessageHandler j;


    /**
     * Constructor, simply grabs the runtime, needs a program name
     * 
     * @param progName
     *            Identifying name for the program - does not directly effect
     *            execution.
     */
    public JavaGenericComputation(String progName) {
        programName = progName;
        r = Runtime.getRuntime();
    }


    /**
     * Starts execution of a bridged program, needs actual execution command,
     * type of communication, and a link to the appropriate message handler As a
     * note, error stream is used even with TCP communications
     * 
     * @param executeCommand
     *            command to be executed to start the program
     * @param commType
     *            as specified in ControlMessage - UPIPES and TCP currently
     *            supported, default is UPIPES
     * @param jmh
     *            the message handler for this program
     * @throws IOException
     */
    public void startExecution(String executeCommand, int commType,
                               JavaMessageHandler jmh) throws IOException {
        j = jmh;
        System.out.println("Execution command: " + executeCommand);
        switch (commType) {
        case ControlMessage.UPIPES: // use unnamed pipes
            p = r.exec(executeCommand);
            se = new StreamStringReader(p.getErrorStream());
            sr = new StreamReader(p.getInputStream(), j);
            sw = new StreamWriter(p.getOutputStream());
            break;
        case ControlMessage.TCP: // set up socket for tcp
            ss = new ServerSocket(0); // choose a free port
            int port = ss.getLocalPort();
            System.out.println("Port is: " + port);
            // Assumes that any program using the bridge takes the port# as a
            // command line argument
            p = r.exec(executeCommand + " " + port);
            prog = ss.accept();
            sr = new StreamReader(prog.getInputStream(), j);
            sw = new StreamWriter(prog.getOutputStream());
            se = new StreamStringReader(p.getErrorStream());
            break;
        default:
            System.out.println("Unhandled communication type, using unnamed pipes");
            p = r.exec(executeCommand);
            se = new StreamStringReader(p.getErrorStream());
            sr = new StreamReader(p.getInputStream(), j);
            sw = new StreamWriter(p.getOutputStream());
            break;
        }
    }


    /**
     * Allows the user to switch communication types from Java. The connected
     * computation needs to support communication switching for full
     * functionality.
     * 
     * @param switchMessage
     *            Actual message for switching communications - should be
     *            formatted in an appropriate manner for the connected
     *            computation
     * @param newCommType
     *            From ControlMessage - UPIPES and TCP currently supported
     * @return Renews connections.
     * @throws IOException
     */
    public JavaGenericComputation switchCommunication(
                                                      JavaByteMessage switchMessage,
                                                      int newCommType)
        throws IOException {
        // fist switch over the StreamReader to the new method - don't hook up
        // yet, but grab it
        switch (newCommType) {
        case ControlMessage.UPIPES: // switching to pipes...
            StreamReader old = sr;
            old.setName("old reader");
            old.interrupt();
            sr = new StreamReader(p.getInputStream(), j);
            this.sendMessage(switchMessage);
            old.closingStream();
            sw = new StreamWriter(p.getOutputStream());
            if (prog.isConnected())
                System.out.println("Socket is still connected");
            ss.close();
            prog.close();
            break;
        case ControlMessage.TCP: // switching to sockets
            ss = new ServerSocket(0);
            int port = ss.getLocalPort();
            System.out.println("Port is: " + port);
            // modify switch message to get the port number...
            switchMessage = appendPort(switchMessage, port);
            this.sendMessage(switchMessage);
            prog = ss.accept();
            StreamReader old2 = sr;
            sr = new StreamReader(prog.getInputStream(), j);
            sw = new StreamWriter(prog.getOutputStream());
            old2.closingStream();
            break;
        default:
            System.out.println("Default comm type"); // switch to pipes
            sr.switchStream(p.getInputStream());
            this.sendMessage(switchMessage);
            sw.switchStream(p.getOutputStream());
            // check if old communication used sockets, etc - close anything
            // needed
            if (prog.isConnected())
                System.out.println("Socket is still connected");
            else
                prog.close();
            ss.close();
            break;
        }
        return this;
    }


    /**
     * Helper function - adds dynamically assigned port number to an existant
     * JavaByteMessage message
     * 
     * @param jm
     *            Original message
     * @param port
     *            port number to append to message
     * @return Modified message
     */
    private JavaByteMessage appendPort(JavaByteMessage jm, int port) {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        DataOutputStream dos = new DataOutputStream(bos);
        try {
            dos.writeInt(ControlMessage.INT);
            dos.writeInt(port);
            dos.flush();
            bos.flush();
        } catch (IOException e) {
            System.out.println("Error: " + e.getMessage());
            e.printStackTrace();
        }
        byte[] portInfo = bos.toByteArray();
        byte[] oldInput = jm.getInput();
        byte[] newInput = new byte[oldInput.length + portInfo.length];
        System.arraycopy(oldInput, 0, newInput, 0, oldInput.length);
        System.arraycopy(portInfo, 0, newInput, oldInput.length,
                         portInfo.length);
        jm.setInput(newInput); // should work...

        return jm;
    }


    /**
     * Function which allows the resource to send messages to an attached
     * computation
     * 
     * @param j
     *            Message to send
     */
    public void sendMessage(JavaByteMessage j) {
        sw.sendMessage(j.toByteArray());
    }
    
    /**
     * Function to close all open connections
     * @throws IOException 
     */
    public void closeConnection() throws IOException {
    	sr.setName("old reader");
    	sr.interrupt();
    	sr.closingStream();
    	sw.finish();
    	if (!ss.equals(null) && ss.isBound())
    		ss.close();
    }

}
