/***
 * Defines the format of messages to handle communications between C++ and Granules
 * Java programs.
 * Contains several more attributes than the Java equivalent, to help
 * with arrays.
 *
 * @author K Ericson
 */
#include <string>
#include <iostream>
#include <stdio.h>
#include <memory.h>
#include <stdlib.h>
#include "helpers.cpp"

#ifndef _JAVAMESSAGES_CPP_
#define _JAVAMESSAGES_CPP_

class javaMessage {
		int mType;
		char* input;
		char* output;
		std::string stream;
		std::string description;
		bool fromJava;
		//helpers
		int inputSize;
		int outputSize;
	public:
		javaMessage(int);
		javaMessage(char*);
		~javaMessage();
		int getType();
		void setInputSize(int);
		int getInputSize();
		void setInput( char*);
		 char* getInput();
		void setOutputSize(int);
		int getOutputSize();
		void setOutput( char*);
		 char* getOutput();
		void setStream(std::string);
		std::string getStream();
		void setDescription(std::string);
		std::string getDescription();
		void setFromJava(bool);
		bool isFromJava();
		int byteSize();
		void getBytes( char*);
};

/***
 * As with the Java version, a <code>javaMessage</code> has the type
 * set at creation, and it cannot be changed later.
 * @param type
 * 		Message type as declared in Constants.h
 */
javaMessage::javaMessage(int type) {
	mType = type;
	fromJava = false;
}

/***
 * Generates a <code>javaMessage</code> from a <code>char*</code>
 * @param blob
 * 		Message in char* format
 */
javaMessage::javaMessage(char* blob) {
	char* temp = new char[4];//temp holder for int bytes
	int start = 0; //index to keep track of where we are in the char*

	memcpy(temp,blob+start,4);
	mType = readInt(temp);
	start +=4;

	int isize, osize;//input and output size respectively
	isize = -1;
	osize = -1;
	memcpy(temp,blob+start,4);
	isize = readInt(temp);
	inputSize = isize;
	start +=4;
	if(isize > 0){
		input = ( char*)malloc(sizeof(char)*isize);
		memcpy(input, blob+start,isize);
	}
	else
		input = NULL;
	start+=isize;

	memcpy(temp,blob+start,4);
	osize = readInt(temp);
	outputSize = osize;
	start+=4;
	if(osize > 0)
	{
		output = (char *)malloc(sizeof(char) * osize);
		memcpy(output,blob+start,osize);
	}
	else
		output = NULL;
	start+=osize;

	int ssize, dsize; //keeps track of length of stream and description strings
	memcpy(temp,blob+start,4);
	ssize = readInt(temp);
	start+=4;

	stream.reserve(ssize+1); //need the extra space for null character
	char* t = new char[ssize+1];
	memcpy(t,blob+start,ssize);
	stream.assign(t);
	start+=ssize;

	memcpy(temp,blob+start,4);
	start+=4;
	dsize = readInt(temp);
	description.reserve(dsize+1);
	t = new char[dsize +1];
	memcpy(t,blob+start,dsize);
	description.assign(t);
	start+=dsize;

	memcpy(&fromJava,blob+start,sizeof(bool));
}

javaMessage::~javaMessage(){
}


//accessor functions
int javaMessage::getType(){
	return mType;
}

void javaMessage::setInputSize(int i){
	inputSize = i;
}

int javaMessage::getInputSize(){
	return inputSize;
}

void javaMessage::setInput(char* i){
	input = i;
}

char* javaMessage::getInput(){
	return input;
}

void javaMessage::setOutputSize(int i) {
	outputSize = i;
}

int javaMessage::getOutputSize(){
	return outputSize;
}

void javaMessage::setOutput( char* o) {
	output = o;
}

 char* javaMessage::getOutput() {
	return output;
}

void javaMessage::setStream(std::string s) {
	stream = s;
}

std::string javaMessage::getStream() {
	return stream;
}

void javaMessage::setDescription(std::string s) {
	description = s;
}

std::string javaMessage::getDescription() {
	return description;
}

void javaMessage::setFromJava(bool b) {
	fromJava = b;
}

bool javaMessage::isFromJava() {
	return fromJava;
}

//end accessor functions

/***
 * Helper function to convert the message into a char* - calculates the size
 * that the byte array will be.
 * @returns int
 * 		returns the size that the char* version of this javaMessage is
 */
int javaMessage::byteSize()
{
	int size = 5 * 4;	 //basic size - includes int types, as well as needed ints for sizes of other fields - too hardcoded?
	size += stream.length();
	size += description.length();
	size += inputSize;
	size += outputSize;
	size += sizeof(bool);
	return size;
}


/***
 * generates the bytes to return the message
 * assumes the char* blob has already been initialized to the correct size
 * correct size is found with <code>byteSize()</code>
 */
void javaMessage::getBytes( char* blob){
	int size = this->byteSize();
	int start = 0;
	int s;
	
	writeInt(blob + start,mType,start);

	if(inputSize >0)
	{
		s = inputSize;
		writeInt(blob + start,s,start);
		memcpy(blob + start,input, s);
		start+=s;
	}
	else
	{
		s = 0;
		writeInt(blob + start,s,start);
	}
	

	if (outputSize > 0)
	{
		s = outputSize;
		writeInt(blob + start,s,start);
		memcpy(blob+start,output, s);
		start+=s;
	}
	else
	{
		s = 0;
		writeInt(blob + start,s,start);
	}
	
	s = stream.length();
	writeInt(blob + start,s,start);
	memcpy(blob+start,stream.c_str(), s);
	start+=s;
	

	s = description.length();
	writeInt(blob + start,s,start);
	memcpy(blob+start,description.c_str(), s);
	start+=s;

	
	memcpy(blob+start,&fromJava, sizeof(bool));
}

#endif //_JAVAMESSAGES_CPP_
