/**
 * Software License, Version 1.0
 * 
 * Copyright 2008 The Trustees of Indiana University.  All rights reserved.
 * 
 *
 *Redistribution and use in source and binary forms, with or without 
 *modification, are permitted provided that the following conditions are met:
 *
 *1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 *2) All redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 *3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Community Grids Lab. For 
 * further information contact the Community Grids Lab at 
 * http://communitygrids.iu.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 *4) The name Indiana University or Community Grids Lab or Granules, 
 * shall not be used to endorse or promote products derived from this software 
 * without prior written permission from Indiana University.  For written 
 * permission, please contact the Advanced Research and Technology Institute 
 * ("ARTI") at 351 West 10th Street, Indianapolis, Indiana 46202.
 *5) Products derived from this software may not be called Granules, 
 * nor may Indiana University or Community Grids Lab or Granules appear
 * in their name, without prior written permission of ARTI.
 * 
 *
 * Indiana University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Indiana University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.exchange;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import ds.granules.exception.MarshallingException;
import ds.granules.scheduler.TaskLifecycleMetrics;
import ds.granules.util.MarshallingUtil;

/**
 * This class encapsulates exchanges related to retrieving the status of a
 * message.
 * 
 * @author Shrideep Pallickara
 * 
 */
public class ExecutionStatus extends Exchange {
    private String jobIdentifier;

    private String taskIdentifier;

    private TaskLifecycleMetrics taskLifecycleMetrics;


    protected ExecutionStatus() {
    }


    protected ExecutionStatus(String jobIdentifier, String taskIdentifier,
            String respondTo) {
        if (jobIdentifier == null) {
            throw new NullPointerException("Specified JobIdentifier is NULL");
        }
        if (taskIdentifier == null) {
            throw new NullPointerException("Specified TaskIdentifier is NULL");
        }
        if (respondTo == null) {
            throw new NullPointerException("Specified RespondTo stream is NULL");
        }

        this.jobIdentifier = jobIdentifier;
        this.taskIdentifier = taskIdentifier;

        setExchangeType(ExchangeType.EXECUTION_STATUS);
        setRequest();
        setRespondTo(respondTo);
    }


    protected ExecutionStatus(String jobIdentifier, String taskIdentifier,
            int conditionFlag, String description,
            TaskLifecycleMetrics taskLifecycleMetrics, String respondTo) {
        if (jobIdentifier == null) {
            throw new NullPointerException("Specified JobIdentifier is NULL");
        }
        if (taskIdentifier == null) {
            throw new NullPointerException("Specified TaskIdentifier is NULL");
        }
        if (conditionFlag != ConditionFlag.PROBLEMS_ENCOUNTERED) {
            if (taskLifecycleMetrics == null) {
                throw new NullPointerException("Specified Metrics is NULL");
            }
        }
        this.jobIdentifier = jobIdentifier;
        this.taskIdentifier = taskIdentifier;
        this.taskLifecycleMetrics = taskLifecycleMetrics;

        setRespondTo(respondTo);
        setExchangeType(ExchangeType.EXECUTION_STATUS);
        setResponse();
        setDescription(description);
        setConditionFlag(conditionFlag);
    }


    @Override
    public boolean isValidExchangeSubType() {
        return true;
    }


    /**
     * @return the jobIdentifier
     */
    public String getJobIdentifier() {
        return jobIdentifier;
    }


    /**
     * 
     * @return the taskIdentifier
     */
    public String getTaskIdentifier() {
        return taskIdentifier;
    }


    /**
     * Retrieve the <code>TaskLifecycleMetrics</code> associated with the
     * RESPONSE.
     * 
     * @return The taskLifecycle metrics for a task.
     */
    public TaskLifecycleMetrics getTaskLifecycleMetrics() {
        return taskLifecycleMetrics;
    }


    @Override
    public byte[] marshallExchangeSpecificElements() throws IOException,
        MarshallingException {
        byte[] marshalledBytes = null;
        ByteArrayOutputStream baOutputStream = new ByteArrayOutputStream();
        DataOutputStream dout =
            new DataOutputStream(new BufferedOutputStream(baOutputStream));
        MarshallingUtil marshaller = MarshallingUtil.getInstance();

        marshaller.writeString(dout, jobIdentifier);
        marshaller.writeString(dout, taskIdentifier);

        if (isResponse()) {
            if (getConditionFlag() != ConditionFlag.PROBLEMS_ENCOUNTERED) {
                byte[] metricBytes = taskLifecycleMetrics.getBytes();
                dout.writeInt(metricBytes.length);
                dout.write(metricBytes);
            }
        }

        /** check to see if this is a request or response and do accordingly */
        dout.flush();
        marshalledBytes = baOutputStream.toByteArray();
        return marshalledBytes;
    }


    @Override
    public void unmarshallExchangeSpecificElements(byte[] exchangeSpecificBytes)
        throws IOException, MarshallingException {
        ByteArrayInputStream baInputStream =
            new ByteArrayInputStream(exchangeSpecificBytes);
        DataInputStream din =
            new DataInputStream(new BufferedInputStream(baInputStream));
        MarshallingUtil marshaller = MarshallingUtil.getInstance();

        jobIdentifier = marshaller.readString(din);
        taskIdentifier = marshaller.readString(din);
        if (isResponse()) {
            if (getConditionFlag() != ConditionFlag.PROBLEMS_ENCOUNTERED) {
                byte[] metricBytes = new byte[din.readInt()];
                din.readFully(metricBytes);
                taskLifecycleMetrics =
                    new TaskLifecycleMetricsReader(metricBytes);
            }
        }
    }


    @Override
    public String toString() {
        String representation = "ExecutionStatus: ";

        if (isRequest()) {
            representation += "(Request)";
        }

        if (isResponse()) {
            representation += "(Response)";
        }

        representation +=
            (" Job Identifier = " + jobIdentifier + ", Task Identifier = " + taskIdentifier);

        if (getDescription() != null) {
            representation += ("  " + getDescription());
        }
        return representation;
    }

}
