/**
 * Software License, Version 1.0 Copyright 2010 Colorado State University. All
 * rights reserved. Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following conditions
 * are met: 1) All redistributions of source code must retain the above
 * copyright notice, the list of authors in the original source code, this list
 * of conditions and the disclaimer listed in this license; 2) All
 * redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the disclaimer listed in this license in the
 * documentation and/or other materials provided with the distribution; 3) Any
 * documentation included with all redistributions must include the following
 * acknowledgment: "This product includes software developed by the Department
 * of Computer Science at Colorado State University. For further information
 * contact the department at http://www.cs.colostate.edu/." Alternatively, this
 * acknowledgment may appear in the software itself, and wherever such
 * third-party acknowledgments normally appear. 4) The name Colorado State
 * University or Granules shall not be used to endorse or promote products
 * derived from this software without prior written permission from Colorado
 * State University. For written permission, please contact the Computer Science
 * Department at 1100 Center Avenue, Fort Collins, CO 80523-1873. 5) Products
 * derived from this software may not be called Granules, nor may Colorado State
 * University or Granules appear in their name, without prior written permission
 * of ARTI. Colorado State University provides no reassurances that the source
 * code provided does not infringe the patent or any other intellectual property
 * rights of any other entity. Colorado State University disclaims any liability
 * to any recipient for claims brought by any other entity based on infringement
 * of intellectual property rights or otherwise. LICENSEE UNDERSTANDS THAT
 * SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO WARRANTIES AS TO CAPABILITIES OR
 * ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES NO WARRANTIES AND MAKES NO
 * REPRESENTATION THAT SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT,
 * COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. COLORADO STATE UNIVERSITY MAKES NO
 * WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES",
 * "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE. LICENSEE ASSUMES THE ENTIRE
 * RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS, AND TO
 * THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING SOFTWARE.
 */
package ds.granules.learning.classification.bayes;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Properties;
import java.util.StringTokenizer;

import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.DatasetFactory;
import ds.granules.dataset.FileAccess;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.CommunicationsException;
import ds.granules.exception.DeploymentException;
import ds.granules.exception.MarshallingException;
import ds.granules.learning.classification.common.BayesPrepStage1;
import ds.granules.learning.classification.common.BayesPrepStage2;
import ds.granules.learning.classification.common.BayesPrepStage3;
import ds.granules.operation.InstanceDeployer;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProgressTracker;
import ds.granules.samples.iterative.IterativeDeployer;

/**
 * The deployer to use the Mahout implementation of Bayes classification.
 * 
 * @author Kathleen Ericson
 *
 */
public class BayesDeployer extends InstanceDeployer {

  private ProgressTracker progressTracker;

  /**
   * Initializes the deployer.
   * 
   * @param streamingProperties
   *          properties for streaming
   * @throws DeploymentException 
   * @throws MarshallingException 
   * @throws IOException 
   * @throws CommunicationsException 
   */
  public BayesDeployer(Properties streamingProperties) throws CommunicationsException, IOException, MarshallingException, DeploymentException {
    initialize(streamingProperties);
  }

  private BayesPrepStage1[] initializePrep1(BayesPrepStage2[] stage2s, String fileList, String directory, int mapStage1, int numOfFiles) throws IOException, DatasetException {
    BayesPrepStage1[] stage1s = new BayesPrepStage1[mapStage1];

    for(int i = 0; i < mapStage1; i++) {
      stage1s[i] = new BayesPrepStage1();

      Properties props = new Properties();
      String worker = i + "";
      props.put("Worker Number", worker);
      props.put("Input Directory", directory);

      stage1s[i].setProcessingDirectives(props );

      DatasetFactory datasetFactory = DatasetFactory.getInstance();
      FileAccess fileAccess = datasetFactory.createFileDataset("FileDataset-ID-BayesPrep", "Bayes classification example");

      BufferedReader in = new BufferedReader(new FileReader(fileList), 65536);
      String filename = "";
      for(int k = 0; k < numOfFiles; k++) {
        filename = in.readLine();

        if (k % mapStage1 == i) {
          filename = directory + filename;
          fileAccess.addFile(filename, "r");
        }
      }

      in.close();

      DatasetCollection dataCollection = stage1s[i].getDatasetCollection();
      dataCollection.addDataset(fileAccess);

      StreamingAccess streamingAccess = datasetFactory.createStreamingDataset(
          "StreamingDataset-ID-Bayes", "Bayes Dataset");
      streamingAccess.addOutputStream("LabelMap", StreamingAccess.STRING_SYNOPSIS);
      streamingAccess.addOutputStream("FeatureMap", StreamingAccess.STRING_SYNOPSIS);
      streamingAccess.addOutputStream("FeatureFreqMap", StreamingAccess.STRING_SYNOPSIS);
      streamingAccess.addOutputStream("DocFreqMap", StreamingAccess.STRING_SYNOPSIS);
      streamingAccess.addOutputStream("WeightMap", StreamingAccess.STRING_SYNOPSIS);

      streamingAccess.addOutputStream("StartTime", StreamingAccess.STRING_SYNOPSIS);

      dataCollection.addDataset(streamingAccess);

      stage1s[i].setAsExecuteOnce();
      
      System.out.println("Props for Stage 1 worker " + i + ":\n" + stage1s[i].getProcessingDirectives().toString());

    }

    return stage1s;
  }

  private BayesPrepStage2[] initializePrep2(BayesPrepStage3[] reducers, String fileList, String directory, int mapStage1, int mapStage2, int numOfFiles) throws DatasetException {
    BayesPrepStage2[] stage2s = new BayesPrepStage2[mapStage2];

    for(int i = 0; i < mapStage2; i++) {
      stage2s[i] = new BayesPrepStage2();

      Properties props = new Properties();
      props.put("Worker Number", i + "");
      props.put("Number Inputs", mapStage1 + "");

      stage2s[i].setProcessingDirectives(props);

      DatasetFactory datasetFactory = DatasetFactory.getInstance();     
      DatasetCollection dataCollection = stage2s[i].getDatasetCollection();
      StreamingAccess streamingAccess = datasetFactory.createStreamingDataset("StreamingDataset-ID-Bayes", "Bayes Dataset");
      streamingAccess.addInputStream("LabelMap", StreamingAccess.STRING_SYNOPSIS);
      streamingAccess.addInputStream("FeatureMap", StreamingAccess.STRING_SYNOPSIS);
      streamingAccess.addInputStream("FeatureFreqMap", StreamingAccess.STRING_SYNOPSIS);
      streamingAccess.addInputStream("DocFreqMap", StreamingAccess.STRING_SYNOPSIS);
      streamingAccess.addInputStream("WeightMap", StreamingAccess.STRING_SYNOPSIS);

      streamingAccess.addOutputStream("DocFreqWeightMap", StreamingAccess.STRING_SYNOPSIS);
      streamingAccess.addOutputStream("CombinedWeightMap", StreamingAccess.STRING_SYNOPSIS);

      dataCollection.addDataset(streamingAccess);

      stage2s[i].setAsExecuteWhenDataAvailable();
      
      System.out.println("Props for Stage 2 worker " + i + ":\n" + stage2s[i].getProcessingDirectives().toString());
    }

    return stage2s;
  }

  private BayesPrepStage3[] initializePrep3(BayesFinalStage reducer, String fileList, String directory, int mapStage1, int mapStage2, int mapStage3, int numOfFiles) throws DatasetException {
    BayesPrepStage3[] stage3s = new BayesPrepStage3[mapStage3];

    for(int i = 0; i < mapStage3; i++) {
      stage3s[i] = new BayesPrepStage3();

      Properties props = new Properties();
      props.put("Worker Number", i + "");
      props.put("Output Directory", directory);

      stage3s[i].setProcessingDirectives(props);

      DatasetFactory datasetFactory = DatasetFactory.getInstance();     
      DatasetCollection dataCollection = stage3s[i].getDatasetCollection();
      StreamingAccess streamingAccess = datasetFactory.createStreamingDataset("StreamingDataset-ID-Bayes", "Bayes Dataset");
      streamingAccess.addInputStream("DocFreqWeightMap", StreamingAccess.STRING_SYNOPSIS);
      streamingAccess.addInputStream("CombinedWeightMap", StreamingAccess.STRING_SYNOPSIS);

      streamingAccess.addOutputStream("SigmaJ", StreamingAccess.STRING_SYNOPSIS);
      streamingAccess.addOutputStream("SigmaK", StreamingAccess.STRING_SYNOPSIS);
      streamingAccess.addOutputStream("SigmaKSigmaJ", StreamingAccess.STRING_SYNOPSIS);

      dataCollection.addDataset(streamingAccess);


      stage3s[i].setAsExecuteWhenDataAvailable();
      
      System.out.println("Props for Stage 3 worker " + i + ":\n" + stage3s[i].getProcessingDirectives().toString());
    }

    return stage3s;
  }

  public void prepareInstancesAndDeploy(int mapStage1, int mapStage2, int mapStage3, String inDirectory,
      String outDirectory, int numFiles, String fileList) throws CommunicationsException, DeploymentException, MarshallingException, DatasetException, IOException, MapReduceException {

    BayesFinalStage reducer = initializeReducer(outDirectory);
    BayesPrepStage3[] s3Maps = initializePrep3(reducer, fileList, outDirectory,
        mapStage1, mapStage2, mapStage3, numFiles);

    BayesPrepStage2[] s2Maps = initializePrep2(s3Maps, fileList, outDirectory, mapStage1, mapStage2, numFiles);

    BayesPrepStage1[] s1Maps = initializePrep1(s2Maps, fileList, inDirectory, mapStage1, numFiles);

 /*   for (int i = 0; i < mapStage1; i++) {
      s1Maps[i].addReduce(s2Maps[0]);
    }*/
    
    // worry about not having traditional map/reduce hookups...
    BayesFinalStage[] reducers = {reducer};
    progressTracker = deployOperations(reducers);   
    String jobId = progressTracker.getIdentifier();
    deployOperations(s3Maps, jobId);
    deployOperations(s2Maps, jobId);
    deployOperations(s1Maps, jobId);
     
  }

  private BayesFinalStage initializeReducer(String outDirectory) throws DatasetException {
    BayesFinalStage red = new BayesFinalStage();

    Properties props = new Properties();
    props.put("Output Directory", outDirectory);

    red.setProcessingDirectives(props);

    DatasetFactory datasetFactory = DatasetFactory.getInstance();     
    DatasetCollection dataCollection = red.getDatasetCollection();
    StreamingAccess streamingAccess = datasetFactory.createStreamingDataset("StreamingDataset-ID-Bayes", "Bayes Dataset");
    streamingAccess.addInputStream("DocFreqWeightMap", StreamingAccess.STRING_SYNOPSIS);
    streamingAccess.addInputStream("SigmaK", StreamingAccess.STRING_SYNOPSIS);
    streamingAccess.addInputStream("SigmaKSigmaJ", StreamingAccess.STRING_SYNOPSIS);
    streamingAccess.addInputStream("StartTime", StreamingAccess.STRING_SYNOPSIS);

    dataCollection.addDataset(streamingAccess);

    red.setAsExecuteWhenDataAvailable();

    return red;
  }

  public static void main(String[] args) {
    String module = "BayesDeployer.main()->";
    if (args.length < 2) {
      System.out.println("Usage: java ds.granules.learning.classification.bayes"
          + "BayesDeployer <hostname> <port>");
      System.exit(0);
    }
    String hostname = args[0];
    String portnum = args[1];
    int map1 = 20;
    int map2 = 1;
    int map3 = 1;
    String infile = "/tmp/ericson/";
    String outfile = "/tmp/ericson/";
    String fileList = "/tmp/ericson/inFiles.txt";
    int numFiles = 20;

    Properties streamingProperties = new Properties();
    streamingProperties.put("hostname", hostname);
    streamingProperties.put("portnum", portnum);

    String helpInfo =
      "\n\nThis is the Mahout Bayes Classification program that uses Granules."
      + "The current list of commands include: \n"
      + "submit <indir> <outdir> <stage1 maps> <stage2 maps> <stage3 maps> <file list> <numFiles> \n"
      + "status       : Lists the status of tasks that comprise Job"
      + "h            : List the set of commands \n\n";
    try {
      BayesDeployer deployer =
        new BayesDeployer(streamingProperties);
      StringTokenizer tok;

      InputStreamReader isr = new InputStreamReader(System.in);
      BufferedReader commandLine = new BufferedReader(isr);

      System.out.println(helpInfo);

      while (true) {
        String s = commandLine.readLine();
        if (s == null) {
          System.out.println(module + "String is null!!!");
          break;
        }

        if (s.startsWith("submit")) {
          System.out.println(module
              + "Processing the [submit] command");

          tok = new StringTokenizer(s, " ");
          int numOfTokens = tok.countTokens();
          tok.nextToken();/* skip the command */

          if (numOfTokens > 1) {
            infile = tok.nextToken();
          }

          if (numOfTokens > 2) {
            outfile = tok.nextToken();
          }

          if (numOfTokens > 3) {
            map1 = Integer.parseInt(tok.nextToken());
          }

          if (numOfTokens > 4) {
            map2 = Integer.parseInt(tok.nextToken());
          }

          if (numOfTokens > 5) {
            map3 = Integer.parseInt(tok.nextToken());
          }

          if (numOfTokens > 6) {
            fileList = tok.nextToken();
          }

          if (numOfTokens > 7) {
            numFiles = Integer.parseInt(tok.nextToken());
          }



          deployer.prepareInstancesAndDeploy(map1, map2, map3, infile, outfile, numFiles, fileList);

          continue;
        }

        if (s.startsWith("status")) {
          deployer.getStatus();
          continue;
        }
      }

    } catch (Exception e) {
      System.out.println(e.getMessage());
      e.printStackTrace();
    }
  }
}

