/**
 * Software License, Version 1.0 Copyright 2010 Colorado State University. All
 * rights reserved. Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following conditions
 * are met: 1) All redistributions of source code must retain the above
 * copyright notice, the list of authors in the original source code, this list
 * of conditions and the disclaimer listed in this license; 2) All
 * redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the disclaimer listed in this license in the
 * documentation and/or other materials provided with the distribution; 3) Any
 * documentation included with all redistributions must include the following
 * acknowledgment: "This product includes software developed by the Department
 * of Computer Science at Colorado State University. For further information
 * contact the department at http://www.cs.colostate.edu/." Alternatively, this
 * acknowledgment may appear in the software itself, and wherever such
 * third-party acknowledgments normally appear. 4) The name Colorado State
 * University or Granules shall not be used to endorse or promote products
 * derived from this software without prior written permission from Colorado
 * State University. For written permission, please contact the Computer Science
 * Department at 1100 Center Avenue, Fort Collins, CO 80523-1873. 5) Products
 * derived from this software may not be called Granules, nor may Colorado State
 * University or Granules appear in their name, without prior written permission
 * of ARTI. Colorado State University provides no reassurances that the source
 * code provided does not infringe the patent or any other intellectual property
 * rights of any other entity. Colorado State University disclaims any liability
 * to any recipient for claims brought by any other entity based on infringement
 * of intellectual property rights or otherwise. LICENSEE UNDERSTANDS THAT
 * SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO WARRANTIES AS TO CAPABILITIES OR
 * ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES NO WARRANTIES AND MAKES NO
 * REPRESENTATION THAT SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT,
 * COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. COLORADO STATE UNIVERSITY MAKES NO
 * WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES",
 * "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE. LICENSEE ASSUMES THE ENTIRE
 * RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS, AND TO
 * THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING SOFTWARE.
 */

package ds.granules.learning.classification.bayes;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.Properties;


import cgl.narada.event.NBEvent;

import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.learning.classification.common.GranTuple;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProcessingException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * 
 * This is the final stage of Bayes Classification, previous stages are similar for both Bayes
 * and Complimentary Bayes, found in ds.granules.learning.classification.common
 * 
 * @author Kathleen Ericson
 *
 */
public class BayesFinalStage extends MapReduceBase{

  private String datasetIdentifier = "StreamingDataset-ID-Bayes";

  private double alpha_i = 1.0; //passed in, 1.0 is default value
  private HashMap<String, Double> thetaNormalizedWeights = new HashMap<String, Double>();

  // what we're looping through.
  // may be passed in piece-meal
  private HashMap<GranTuple, Double> docFreqWeightMap = new HashMap<GranTuple, Double>();
  //passed in -> Sigma_k
  private HashMap<String, Double> labelWeights = new HashMap<String, Double>();
  //passed in -> trainer-tfIdf/trainer-vocabCount
  private double vocabCount = 0.0;
  private long startTime;

  private String outputDir;
  private boolean initialized;
  private boolean gotTfidf = false;
  // to be more flexible, would need a count of how many may come in
  private boolean gotSigmaKSigmaJ = false;
  private boolean gotSigmaK = false;

  public BayesFinalStage() {
    setDomain("Classification");
    setVersionInformation(075);
    setOperationIdentifier("BayesFinalStage");
  }

  private void initialize() {
    Properties props = this.getProcessingDirectives();
    outputDir = props.getProperty("Output Directory");
  }

  @Override
  public void execute() throws ProcessingException {
    // loop through trainer-tfIdf/trainer-tfIdf
    // will get that first, but need to wait for stage 3 to finish so we can have our lookup table
    System.out.println("Hit Execute in Final Stage!");
    // and total sum
    if(!initialized) {
      initialize();
    }

    //figure out what this should be...
    DatasetCollection datasetCollection = getDatasetCollection();
    StreamingAccess access = null;
    try {
      Dataset dataset = datasetCollection.getDataset(datasetIdentifier);

      if (dataset.getDatasetType() == Dataset.STREAMS) {
        access = (StreamingAccess) dataset;
      } else {
        System.out.println("Incorrect datatset: "
            + dataset.getDatasetType() + " initialized. Returning ...");
        return;
      }

      if (access != null) {

        while (dataset.isDataAvailable()) {
          NBEvent nbEvent = access.getStreamingData();
          processStreamPacket(nbEvent);
        }

      } else {
        System.out.println("\nThe INPUT Dataset is NULL!\n");
      }

    } catch (DatasetException e) {
      e.printStackTrace();

    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (MarshallingException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

  }

  private void processStreamPacket(NBEvent nbEvent) throws IOException, MarshallingException {
    String synopsis = (String) nbEvent.getContentSynopsis();
    byte[] resultBytes = nbEvent.getContentPayload();
    Results results = ResultsFactory.getInstance().getResults(resultBytes);
    byte[] resultPayloadBytes = results.getResultPayload();
    // may need this later...
    //byte[] decompressedBytes = decompressPayload(resultPayloadBytes, Integer
    //    .parseInt(results.getRespondTo()));
    ResultPayload resultPayload = ResultPayload.getInstance();
    if (synopsis.equals("StartTime")) {
      if(results.getResultType() != ResultPayload.LONG) {
        System.out
        .println("BayesFinalStage: Incorrect ResultType received on Synopsis: " + synopsis);
        return;
      }
      startTime = resultPayload.getLong(resultPayloadBytes);
    }
    else if(synopsis.equals("DocFreqWeightMap")) {
      if (results.getResultType() != ResultPayload.HASHMAP) {
        System.out
        .println("BayesFinalStage: Incorrect ResultType received on Synopsis: " + synopsis);
        return;
      }
      HashMap<GranTuple, Double> tmpMap = (HashMap<GranTuple, Double>)resultPayload.getHashMap(resultPayloadBytes);
      addInput(tmpMap);
      this.gotTfidf = true;
    }
    else if(synopsis.equals("SigmaK")) {
      if (results.getResultType() != ResultPayload.HASHMAP) {
        System.out
        .println("BayesFinalStage: Incorrect ResultType received on Synopsis: " + synopsis);
        return;
      }
      HashMap<String, Double> tmpMap = (HashMap<String, Double>)resultPayload.getHashMap(resultPayloadBytes);
      addLabelMap(tmpMap);
      this.gotSigmaK = true;
    }
    else if(synopsis.equals("SigmaKSigmaJ")) {
      if (results.getResultType() != ResultPayload.DOUBLE) {
        System.out
        .println("BayesFinalStage: Incorrect ResultType received on Synopsis: " + synopsis);
        return;
      }
      double tmp = resultPayload.getDouble(resultPayloadBytes);
      vocabCount += tmp;
      this.gotSigmaKSigmaJ = true;
    }

    // after getting the last thing...
    if(this.gotSigmaK && this.gotSigmaKSigmaJ && this.gotTfidf) {
      //loop through the LabelWeightMap to grab inputs
      for(GranTuple key : docFreqWeightMap.keySet()) {
        processInput(docFreqWeightMap.get(key), key);
      }
      saveOut();
      long time = System.nanoTime() - startTime;
      FileWriter fw = new FileWriter(new File("BayesFinalStage.timing"), true);
      fw.write(time +"\n");
      fw.flush();
      fw.close();
      System.out.println("Saved out timing, done");
    }
    
  }

  private void addLabelMap(HashMap<String, Double> tmpMap) {
    for(String key : tmpMap.keySet()) {
      if(labelWeights.containsKey(key)) {
        labelWeights.put(key, labelWeights.get(key) + tmpMap.get(key));
      }
      else {
        labelWeights.put(key, tmpMap.get(key));
      }
    }
  }

  private void addInput(HashMap<GranTuple, Double> tmpMap) {
    //loop through in case the docFreqWeightMap doesn't come through all together
    for (GranTuple key : tmpMap.keySet()) {
      if(docFreqWeightMap.containsKey(key)) {
        docFreqWeightMap.put(key, docFreqWeightMap.get(key) + tmpMap.get(key));
      }
      else {
        docFreqWeightMap.put(key, tmpMap.get(key));
      }
    }

  }

  // loop through each weight in tfIdf output
  private void processInput(double val, GranTuple labelFeaturePair) {
    String label = labelFeaturePair.stringAt(0);
    double weight = Math.log(val + alpha_i)/ (labelLookup(label) + vocabCount);
    if(thetaNormalizedWeights.containsKey(label)){
      thetaNormalizedWeights.put(label, thetaNormalizedWeights.get(label) + weight);
    }
    else {
      thetaNormalizedWeights.put(label, weight);
    }
  }

  private double labelLookup(String label) {
    if(labelWeights.containsKey(label))
      return labelWeights.get(label);
    else
      return 0;
  }

  private void saveOut() {
    // do stuff here...
  }

}
