/**
 * Software License, Version 1.0 Copyright 2010 Colorado State University. All
 * rights reserved. Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following conditions
 * are met: 1) All redistributions of source code must retain the above
 * copyright notice, the list of authors in the original source code, this list
 * of conditions and the disclaimer listed in this license; 2) All
 * redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the disclaimer listed in this license in the
 * documentation and/or other materials provided with the distribution; 3) Any
 * documentation included with all redistributions must include the following
 * acknowledgment: "This product includes software developed by the Department
 * of Computer Science at Colorado State University. For further information
 * contact the department at http://www.cs.colostate.edu/." Alternatively, this
 * acknowledgment may appear in the software itself, and wherever such
 * third-party acknowledgments normally appear. 4) The name Colorado State
 * University or Granules shall not be used to endorse or promote products
 * derived from this software without prior written permission from Colorado
 * State University. For written permission, please contact the Computer Science
 * Department at 1100 Center Avenue, Fort Collins, CO 80523-1873. 5) Products
 * derived from this software may not be called Granules, nor may Colorado State
 * University or Granules appear in their name, without prior written permission
 * of ARTI. Colorado State University provides no reassurances that the source
 * code provided does not infringe the patent or any other intellectual property
 * rights of any other entity. Colorado State University disclaims any liability
 * to any recipient for claims brought by any other entity based on infringement
 * of intellectual property rights or otherwise. LICENSEE UNDERSTANDS THAT
 * SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO WARRANTIES AS TO CAPABILITIES OR
 * ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES NO WARRANTIES AND MAKES NO
 * REPRESENTATION THAT SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT,
 * COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. COLORADO STATE UNIVERSITY MAKES NO
 * WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES",
 * "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE. LICENSEE ASSUMES THE ENTIRE
 * RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS, AND TO
 * THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING SOFTWARE.
 */

package ds.granules.learning.clustering.dirichlet;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.FileUtil;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.fs.PathFilter;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.mahout.clustering.dirichlet.DirichletCluster;
import org.apache.mahout.clustering.dirichlet.DirichletState;
import org.apache.mahout.clustering.dirichlet.UncommonDistributions;
import org.apache.mahout.clustering.dirichlet.models.DistributionDescription;
import org.apache.mahout.common.distance.DistanceMeasure;
import org.apache.mahout.math.SequentialAccessSparseVector;
import org.apache.mahout.math.Vector;
import org.apache.mahout.math.VectorWritable;

import cgl.narada.event.NBEvent;
import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.FileAccess;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * An example demonstrating the <code>Map</code> role within a Word Count
 * operation.
 * 
 * @author Shrideep Pallickara
 */
public class DirichletMapper extends MapReduceBase {
  // from fuzzykmeans
  private static final double MINIMAL_VALUE = 0.0000000001;
  private double m = 2.0; // default value
  private int modelNum;
  private String distanceMeasure;

  private ResultPayload resultPayload;
  private String datasetIdentifier = "FileDataset-ID-MKMeans";
  private String streamingIdentifier = "StreamingDataset-ID-MKMeans";
  private int WorkerNumber;
  private boolean initialized = false;
  private String initClusters; // filename where initial clusters are located
  double distanceThreshold; // will be entered with list of points/clusters

  DistanceMeasure measure; // will be entered with initialization
  ArrayList<VectorWritable> points2Cluster = new ArrayList<VectorWritable>(); // the
                                                                              // points
                                                                              // it's
                                                                              // clustering
  long runTime; // quick and dirty timer...
  int runCount = 0; // loop and gather times
  int maxRun = 20; // how many times to run
  DirichletState state;
  String modelFactory;
  String modelPrototype;
  long totalRunTime;

  public DirichletMapper() {
    setDomain("Generic");
    setVersionInformation(200);
    setOperationIdentifier("MahoutMapper");
    resultPayload = ResultPayload.getInstance();
  }

  /**
   * Gets inputs from the deployer.
   * Reads in the points it is responsible for clustering
   */
  private void initializeMapper() {
    // get distance measure here, threshold here
    // read in points it is responsible for from wherever it is supposed to...
    Properties props = this.getProcessingDirectives();
    distanceMeasure = props.getProperty("Distance Measure");
    ClassLoader ccl = Thread.currentThread().getContextClassLoader();
    try {
      measure = ccl.loadClass(distanceMeasure).asSubclass(DistanceMeasure.class)
          .newInstance();
    } catch (InstantiationException e) {
      e.printStackTrace();
    } catch (IllegalAccessException e) {
      e.printStackTrace();
    } catch (ClassNotFoundException e) {
      e.printStackTrace();
    }
    modelFactory = props.getProperty("Model Distribution");
    modelPrototype = props.getProperty("Model Prototype");
    WorkerNumber = Integer.parseInt(props.getProperty("Worker Number"));
    distanceThreshold = Double.parseDouble(props
        .getProperty("Distance Threshold"));
    initClusters = props.getProperty("Initial Clusters");
    modelNum = Integer.parseInt(props.getProperty("Num Models"));
    m = Double.parseDouble(props.getProperty("Alpha"));
    loadState(initClusters);
 //   System.out.println("Mapper done initializing!");
    initialized = true;
    runTime = 0;
    totalRunTime = 0;
  }

  /**
   * Loads initial state information
   * @param initClusters2
   *             File location of the initial cluster state.
   */
  private void loadState(String initClusters2) {
//    System.out.println("Loading clusters in mapper");
    Configuration conf = new Configuration();
    Path clusterPath = new Path(initClusters2, "*");

    try {
 //     System.out.println("getting model distribution");
 /*     ModelDistribution<VectorWritable> modelDistribution;
      ClassLoader ccl = Thread.currentThread().getContextClassLoader();
      System.out.println("got ccl");
      Class<? extends AbstractVectorModelDistribution> cl = ccl.loadClass(
          modelFactory).asSubclass(AbstractVectorModelDistribution.class);
      modelDistribution = cl.newInstance();
      System.out.println("got modelDistribution");

      Class<? extends Vector> vcl = ccl.loadClass(modelPrototype).asSubclass(
          Vector.class);
      Constructor<? extends Vector> v = vcl.getConstructor(int.class);
      ((AbstractVectorModelDistribution) modelDistribution)
          .setModelPrototype(new VectorWritable(v.newInstance(modelNum)));
      System.out.println("Got Vector class");*/
      DistributionDescription description = new DistributionDescription(modelFactory, modelPrototype, distanceMeasure, modelNum);
//      System.out.println("got description");
//      System.out.println("Description: " + description.toString());
      try {
      state = new DirichletState(description, modelNum, m);
      }
      catch (Exception e) {
        System.out.println("caught something!!!");
        e.printStackTrace();
      }
 //     System.out.println("made a new state");
      // get all filtered file names in result list   
      FileSystem fs = clusterPath.getFileSystem(conf);
      PathFilter clusterFileFilter = new PathFilter() {
        @Override
        public boolean accept(Path path) {
          return path.getName().startsWith("part");
        }
      };
      FileStatus[] matches = fs.listStatus(FileUtil.stat2Paths(fs.globStatus(
          clusterPath, clusterFileFilter)), clusterFileFilter);
      System.out.println("There are " + matches.length + " matches in " + clusterPath + "!");

      for (FileStatus match : matches) {
      //  System.out.println("working on match: " + match.getPath());
        SequenceFile.Reader reader = null;
        try {
          reader = new SequenceFile.Reader(fs, match.getPath(), conf);
          Writable key = new Text();
          DirichletCluster c = new DirichletCluster();
          while (reader.next(key, c)) {
            // get the cluster info
            state.getClusters().set(Integer.parseInt(key.toString()), c);
            c = new DirichletCluster();
          }
        } catch (Exception e) {
          System.out.println("Error: " + e.getMessage());
          e.printStackTrace();
        } finally {
          reader.close();
        }

      }
    } catch (IOException e) {
      System.out.println("Problem with IO");
      throw new IllegalStateException(e);
    }
    state.setMixture(UncommonDistributions.rDirichlet(state.totalCounts(), m));
  }

  public void execute() {
    try {
      DatasetCollection dc = getDatasetCollection();
      if (!initialized) {
        System.out.println("Mapper not yet initialized");
        FileAccess k = null;
        Dataset d = dc.getDataset(datasetIdentifier);
        if (d.getDatasetType() == Dataset.FILES)
          k = (FileAccess) d;
        else {
          System.out.println(WorkerNumber + ": Dataset issue: "
              + d.getDatasetType() + "initialized.  Returning...");
          return;
        }

        Enumeration fnames = k.getFilenames();

        while (fnames.hasMoreElements()) {
          String f = (String) fnames.nextElement();
          getPoints(f);
        }

        initializeMapper();
        this.changeToExecuteWhenDataAvailable();
        totalRunTime = System.nanoTime();
        HashMap<Integer, GranulesDirichletCluster> forReducer = DataforReducer(runIteration());
        writeResults(forReducer);
      }
      // now listen for streaming stuff

      StreamingAccess s = null;
      Dataset t = dc.getDataset(streamingIdentifier);
      if (t.getDatasetType() == Dataset.STREAMS)
        s = (StreamingAccess) t;
      else {
        System.out.println(WorkerNumber + ": Dataset issue: "
            + t.getDatasetType() + "initialized.  Returning...");
        return;
      }

      while (s.isDataAvailable()) {
        NBEvent nbEvent = s.getStreamingData();
        processResultsFromMap(nbEvent);
      }

    } catch (DatasetException e) {
      e.printStackTrace();
    } catch (IOException e) {
      System.out.println("Problems converting bytes to longs");
      e.printStackTrace();
    } catch (MarshallingException e) {
      System.out.println("Problems Marshalling: " + e.getMessage());
      e.printStackTrace();
    } catch (MapReduceException e) {
      System.out.println("Problems writing Results: " + e.getMessage());
      e.printStackTrace();
    }

  }

  private void getPoints(String f) {
    // add point vectors to the point list. File f contains several points
    try {
//      System.out.println("hello world");
      Configuration c = new Configuration();
      Path p = new Path(f);
      SequenceFile.Reader reader = new SequenceFile.Reader(p.getFileSystem(c),
          p, c);
 //     System.out.println("after reader");
      Writable key;
      key = (Writable) reader.getKeyClass().newInstance();
      VectorWritable value = new VectorWritable();
      while (reader.next(key, value)) {
        // get the point info
        points2Cluster.add(value);
        value = new VectorWritable();
      }
      reader.close();
      System.out.println("finished with reader on file " + f + ", have " + points2Cluster.size() + " points");
    } catch (Exception e) {
      System.out.println("Problem reading in vector file");
      e.printStackTrace();
    }
  }

  private void startRun(NBEvent nbEvent) throws IOException,
      MarshallingException {
    // contains clusters
    if (!initialized)
      initializeMapper();
    byte[] data = nbEvent.getContentPayload();
    Results results = ResultsFactory.getInstance().getResults(data);
    byte[] resultPayloadBytes = results.getResultPayload();
    ResultPayload rp = ResultPayload.getInstance();

    Object[] objArr = rp.getObjectArray(resultPayloadBytes);

    state.getClusters().clear();
    for (Object c : objArr) {
      state.getClusters().add(((GranulesDirichletCluster) c).getCluster());
    }
  }

  private synchronized void processResultsFromMap(NBEvent nbEvent)
      throws IOException, MarshallingException, MapReduceException {

    String synopsis = (String) nbEvent.getContentSynopsis();
    if (synopsis.startsWith("Initialize") || synopsis.startsWith("ClusterInfo")) // contains
                                                                                 // clusters,
                                                                                 // maxIterations
                                                                                 // and
                                                                                 // distance
                                                                                 // measure,
    {
      startRun(nbEvent);
      HashMap<Integer, GranulesDirichletCluster> forReducer = DataforReducer(runIteration());
      writeResults(forReducer);
    } else // assume termination
    {
      //Code to save out timing information.  Uncomment to measure runtimes.
      long tmpTime = System.nanoTime() - totalRunTime;
      FileWriter fw = new FileWriter(new File("worker"
          + WorkerNumber + ".timing"), true);
      fw.write(runTime + "\n");
      fw.flush();
      fw.close();
      if (WorkerNumber == 0) { // save out totalRunTime
        fw = new FileWriter(new File("overall_dirichlet.timing"),
            true);
        fw.write(tmpTime + "\n");
        fw.flush();
        fw.close();
      }
      runTime = 0;

      if (runCount < maxRun) {// start again
        initialized = false;
        state = null;
        System.gc();
        runCount++;
        initializeMapper();
        System.out.println("starting iteration on " + points2Cluster.size()
            + " points...");
        totalRunTime = System.nanoTime();
        HashMap<Integer, GranulesDirichletCluster> forReducer = DataforReducer(runIteration());

        writeResults(forReducer);
        System.out.println("Sent results to reducer");
      }
    }
  }

  // to actually run clustering
  private List<DirichletCluster> runIteration() {
 //   System.out.println("in runIteration()");
    long startTime = System.nanoTime();
    List<DirichletCluster> results = new ArrayList<DirichletCluster>();
    int loopCt = 0;
   
    for (VectorWritable point : points2Cluster) {
//      System.out.println("Starting on loop " + loopCt + " out of " + points2Cluster.size());
      Vector pi = new SequentialAccessSparseVector(state.getNumClusters());
 //     System.out.println("made pi!");
      for (int i = 0; i < state.getNumClusters(); i++) {
        pi.set(i, state.adjustedProbability(point, i));
      }
  //    System.out.println("done setting pi!");

      int k = UncommonDistributions.rMultinom(pi);
 //     System.out.println("got k!: " + k);
      
 //     System.out.println("State cluster length: " + state.getClusters().size());
      DirichletCluster c = state.getClusters().get(k);
//      System.out.println("got c!");
      c.observe(point);
      loopCt++;
    }
 //   System.out.println("outta loop!");
    results.addAll(state.getClusters());
    runTime += System.nanoTime() - startTime;
    return results;
  }

  private HashMap<Integer, GranulesDirichletCluster> DataforReducer(
      List<DirichletCluster> list) {
    HashMap<Integer, GranulesDirichletCluster> results = new HashMap<Integer, GranulesDirichletCluster>();
//    System.out.println("getting hashmap, list size: " + list.size());
    for (int i = 0; i < list.size(); i++) {
      // if(c.getNumPoints())
      results.put(i, new GranulesDirichletCluster(list.get(i)));
    }

    return results;
  }

  private void writeResults(HashMap<Integer, GranulesDirichletCluster> forReducer)
      throws MarshallingException, MapReduceException {
    boolean completeResult = true;
    ResultPayload resultPayload = ResultPayload.getInstance();

//    System.out.println("here?");
//    System.out.println("here?0.25");
//    Collection<GranulesDirichletCluster> test = forReducer.values();
//    System.out.println("test length: " + test.size());

    byte[] resultPayloadBytes = resultPayload.getBytes(forReducer);
//    System.out.println("here?0.75");
//    System.out.println("here?1");
    Results results = createResults(completeResult, false);
//  System.out.println("here?2");
    results.setResultPayload(resultPayloadBytes);
//    System.out.println("here?3");
    results.setResultType(ResultPayload.HASHMAP);
//    System.out.println("Does it get here?");
    writeResults(results);
  }

  /** 
   * Computes the probability of a point belonging to a cluster
   */
  public double computeProbWeight(double clusterDistance,
      Iterable<Double> clusterDistanceList) {
    if (clusterDistance == 0) {
      clusterDistance = MINIMAL_VALUE;
    }
    double denom = 0.0;
    for (double eachCDist : clusterDistanceList) {
      if (eachCDist == 0.0) {
        eachCDist = MINIMAL_VALUE;
      }
      denom += Math.pow(clusterDistance / eachCDist, 2.0 / (m - 1));
    }
    return 1.0 / denom;
  }

}
