/**
 * Software License, Version 1.0 Copyright 2010 Colorado State University. All
 * rights reserved. Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following conditions
 * are met: 1) All redistributions of source code must retain the above
 * copyright notice, the list of authors in the original source code, this list
 * of conditions and the disclaimer listed in this license; 2) All
 * redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the disclaimer listed in this license in the
 * documentation and/or other materials provided with the distribution; 3) Any
 * documentation included with all redistributions must include the following
 * acknowledgment: "This product includes software developed by the Department
 * of Computer Science at Colorado State University. For further information
 * contact the department at http://www.cs.colostate.edu/." Alternatively, this
 * acknowledgment may appear in the software itself, and wherever such
 * third-party acknowledgments normally appear. 4) The name Colorado State
 * University or Granules shall not be used to endorse or promote products
 * derived from this software without prior written permission from Colorado
 * State University. For written permission, please contact the Computer Science
 * Department at 1100 Center Avenue, Fort Collins, CO 80523-1873. 5) Products
 * derived from this software may not be called Granules, nor may Colorado State
 * University or Granules appear in their name, without prior written permission
 * of ARTI. Colorado State University provides no reassurances that the source
 * code provided does not infringe the patent or any other intellectual property
 * rights of any other entity. Colorado State University disclaims any liability
 * to any recipient for claims brought by any other entity based on infringement
 * of intellectual property rights or otherwise. LICENSEE UNDERSTANDS THAT
 * SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO WARRANTIES AS TO CAPABILITIES OR
 * ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES NO WARRANTIES AND MAKES NO
 * REPRESENTATION THAT SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT,
 * COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. COLORADO STATE UNIVERSITY MAKES NO
 * WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES",
 * "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE. LICENSEE ASSUMES THE ENTIRE
 * RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS, AND TO
 * THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING SOFTWARE.
 */
package ds.granules.learning.clustering.kmeans;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Properties;
import java.util.StringTokenizer;

import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.DatasetFactory;
import ds.granules.dataset.FileAccess;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.CommunicationsException;
import ds.granules.exception.DeploymentException;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.InstanceDeployer;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProgressTracker;

/**
 * The deployer for the Mahout implementation of K-Means Clustering
 * 
 * @author ericson
 */
public class KMeansDeployer extends InstanceDeployer {
  
  private ProgressTracker progressTracker;

  /**
   * Initializes the deployer.
   * 
   * @param streamingProperties
   *          The properties for streaming
   * @throws CommunicationsException
   *           If there are network problems during communications
   * @throws IOException
   *           If there are IO problems
   * @throws MarshallingException
   *           If there are marshalling erros
   * @throws DeploymentException
   *           If there are deployment related problems
   */
  public KMeansDeployer(Properties streamingProperties)
      throws CommunicationsException, IOException, MarshallingException,
      DeploymentException {
    initialize(streamingProperties);
  }

  private KMeansMapper initializeMKMMap(int mapNum, String distanceMeasure,
      String fileList, String Directory, int numOfMaps, int numOfFiles,
      String distanceThreshold, KMeansReducer r, String initClusters)
      throws DatasetException, MapReduceException, IOException {
    KMeansMapper map = new KMeansMapper();

    Properties props = new Properties();
    String value = mapNum + "";
    String property = "Worker Number";
    props.put(property, value);
    props.put("Distance Measure", distanceMeasure);
    props.put("Distance Threshold", distanceThreshold);
    props.put("Initial Clusters", initClusters);

    map.setProcessingDirectives(props);

    DatasetFactory datasetFactory = DatasetFactory.getInstance();
    FileAccess fileAccess = datasetFactory.createFileDataset(
        "FileDataset-ID-MKMeans", "Mahout Clustering example");

    BufferedReader in = new BufferedReader(new FileReader(fileList), 65536);
    String filename = "";
    for (int i = 0; i < numOfFiles; i++) {
      filename = in.readLine();

      if (i % numOfMaps == mapNum) {
        filename = Directory + filename;
        fileAccess.addFile(filename, "r");
      }
    }

    in.close();

    DatasetCollection datasetCollection = map.getDatasetCollection();
    datasetCollection.addDataset(fileAccess);

    StreamingAccess streamingAccess = datasetFactory.createStreamingDataset(
        "StreamingDataset-ID-MKMeans", "Cluster Info");
    streamingAccess.addInputStream("ClusterInfo",
        StreamingAccess.STRING_SYNOPSIS);
    streamingAccess.addInputStream("Termination",
        StreamingAccess.STRING_SYNOPSIS);
    datasetCollection.addDataset(streamingAccess);

    System.out.println("Initialized Map (" + mapNum + ")");
    map.setAsExecuteOnce();
    map.addReduce(r);

    return map;
  }

  /**
   * Prepares and deploys Mappers and a Reducer
   * 
   * @param numMaps
   *          The number of <code>MKMeansMapper</code>s that need to be
   *          initialized.
   * @param filenameBase
   *          The directory in which the files are stored.
   * @param numOfFiles
   *          The total number of files that need to be processed
   * @param fileList
   *          List of Files to be processessed.
   * @throws DatasetException
   *           If there are problems initializing the datasets
   * @throws MapReduceException
   *           If there is a MapReduceExcecution because of problems when
   *           setting up the roles.
   * @throws IOException
   *           If there are IO problems.
   * @throws CommunicationsException
   *           If there are network problems during communications
   * @throws DeploymentException
   *           If there are deployment related problems
   * @throws MarshallingException
   *           If there are marshalling problems.
   * @throws InterruptedException
   *           If there are problems sleeping
   */
  public void prepareInstancesAndDeploy(int numMaps, String directory,
      int numOfFiles, String fileList, int iters, String delta, String dm,
      String clusterFile) throws MapReduceException, CommunicationsException,
      DeploymentException, MarshallingException, DatasetException,
      InterruptedException, IOException {

    KMeansReducer reduce = initializeReducer(iters, delta, dm, clusterFile);

    KMeansMapper[] maps = new KMeansMapper[numMaps];

    for (int i = 0; i < numMaps; i++) {
      maps[i] = initializeMKMMap(i, dm, fileList, directory, numMaps,
          numOfFiles, delta, reduce, clusterFile);
    }

    KMeansReducer[] reducers = { reduce };
    progressTracker = deployOperations(reducers);
    String jobId = progressTracker.getIdentifier();

    deployOperations(maps, jobId);

  }

  /**
   * Initailize the reducer
   * @param Iterations
   *             The max number of iterations to run
   * @param delta
   *             If the change in centroids falls below delta, stop running
   * @param dm
   *             Distance Measure Class to use.  Found in:
   *             org.apache.mahout.common.distance
   * @param clusterFile
   *             File containing initial cluster positions
   * @return
   * @throws DatasetException
   *             If there are problems initializing the datasets
   * @throws MapReduceException
   *             If there is a MapReduceExcecution because of problems when
   *           setting up the roles.
   */
  private KMeansReducer initializeReducer(int Iterations, String delta,
      String dm, String clusterFile) throws DatasetException,
      MapReduceException {
    KMeansReducer red = new KMeansReducer();

    Properties props = new Properties();
    props.put("Max Iterations", Iterations + "");
    props.put("Convergence Delta", delta);
    props.put("Distance Measure", dm);
    props.put("Initial Clusters", clusterFile);

    red.setProcessingDirectives(props);
    red.setAsExecuteWhenDataAvailable();

    return red;
  }

  /**
   * @param args
   */
  public static void main(String[] args) {
    String module = "IterativeDeployer.main()->";
    if (args.length < 2) {
      System.out.println("Usage: java ds.granules.samples.periodic"
          + "IterativeDeployer <hostname> <port>");
      System.exit(0);
    }
    String hostname = args[0];
    String portnum = args[1];
    int numOfMaps = 1;
    String baseDir = "/tmp/";
    int numOfFiles = 2;
    String fileList = "files.txt";
    int maxIters = 100;
    String delta = "0.5";
    String dm = "org.apache.mahout.common.distance.SquaredEuclideanDistanceMeasure";
    String clusterFile = "initialClusters";

    Properties streamingProperties = new Properties();
    streamingProperties.put("hostname", hostname);
    streamingProperties.put("portnum", portnum);

    String helpInfo = "\n\nThis is the Mahout K-means test program that uses Granules."
        + "The current list of commands include: \n"
        + "submit <baseDir> <num_of_maps> <num_of_files> <fileList> <Max_iterations> <delta> <distance_measure> <cluster_file> \n"
        + "status       : Lists the status of tasks that comprise Job"
        + "h            : List the set of commands \n\n";
    try {
      KMeansDeployer deployer = new KMeansDeployer(streamingProperties);
      StringTokenizer tok;

      InputStreamReader isr = new InputStreamReader(System.in);
      BufferedReader commandLine = new BufferedReader(isr);

      System.out.println(helpInfo);

      while (true) {
        String s = commandLine.readLine();
        if (s == null) {
          System.out.println(module + "String is null!!!");
          break;
        }

        if (s.startsWith("submit")) {
          System.out.println(module + "Processing the [submit] command");

          tok = new StringTokenizer(s, " ");
          int numOfTokens = tok.countTokens();
          tok.nextToken();/* skip the command */

          if (numOfTokens > 1) {
            baseDir = tok.nextToken();
          }

          if (numOfTokens > 2) {
            numOfMaps = Integer.parseInt(tok.nextToken());
          }

          if (numOfTokens > 3) {
            numOfFiles = Integer.parseInt(tok.nextToken());
          }

          if (numOfTokens > 4) {
            fileList = tok.nextToken();
          }

          if (numOfTokens > 5) {
            maxIters = Integer.parseInt(tok.nextToken());
          }

          if (numOfTokens > 6) {
            delta = tok.nextToken();
          }

          if (numOfTokens > 7) {
            dm = tok.nextToken();
          }

          if (numOfTokens > 8) {
            clusterFile = tok.nextToken();
          }

          deployer.prepareInstancesAndDeploy(numOfMaps, baseDir, numOfFiles,
              fileList, maxIters, delta, dm, clusterFile);

          continue;
        }

        if (s.startsWith("status")) {
          deployer.getStatus();
          continue;
        }
      }

    } catch (Exception e) {
      System.out.println(e.getMessage());
      e.printStackTrace();
    }
  }
}
