/**
 * Software License, Version 1.0 Copyright 2010 Colorado State University. All
 * rights reserved. Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following conditions
 * are met: 1) All redistributions of source code must retain the above
 * copyright notice, the list of authors in the original source code, this list
 * of conditions and the disclaimer listed in this license; 2) All
 * redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the disclaimer listed in this license in the
 * documentation and/or other materials provided with the distribution; 3) Any
 * documentation included with all redistributions must include the following
 * acknowledgment: "This product includes software developed by the Department
 * of Computer Science at Colorado State University. For further information
 * contact the department at http://www.cs.colostate.edu/." Alternatively, this
 * acknowledgment may appear in the software itself, and wherever such
 * third-party acknowledgments normally appear. 4) The name Colorado State
 * University or Granules shall not be used to endorse or promote products
 * derived from this software without prior written permission from Colorado
 * State University. For written permission, please contact the Computer Science
 * Department at 1100 Center Avenue, Fort Collins, CO 80523-1873. 5) Products
 * derived from this software may not be called Granules, nor may Colorado State
 * University or Granules appear in their name, without prior written permission
 * of ARTI. Colorado State University provides no reassurances that the source
 * code provided does not infringe the patent or any other intellectual property
 * rights of any other entity. Colorado State University disclaims any liability
 * to any recipient for claims brought by any other entity based on infringement
 * of intellectual property rights or otherwise. LICENSEE UNDERSTANDS THAT
 * SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO WARRANTIES AS TO CAPABILITIES OR
 * ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES NO WARRANTIES AND MAKES NO
 * REPRESENTATION THAT SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT,
 * COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. COLORADO STATE UNIVERSITY MAKES NO
 * WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES",
 * "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE. LICENSEE ASSUMES THE ENTIRE
 * RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS, AND TO
 * THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING SOFTWARE.
 */

package ds.granules.learning.clustering.kmeans;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.FileUtil;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.fs.PathFilter;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.io.Writable;
import org.apache.mahout.clustering.kmeans.Cluster;
import org.apache.mahout.common.IOUtils;
import org.apache.mahout.common.distance.DistanceMeasure;
import org.apache.mahout.math.Vector;
import org.apache.mahout.math.VectorWritable;

import cgl.narada.event.NBEvent;
import cgl.narada.service.ServiceException;
import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.FileAccess;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * An example demonstrating the <code>Map</code> role within a Word Count
 * operation.
 * 
 * @author Shrideep Pallickara
 */
public class KMeansMapper extends MapReduceBase {
  private ResultPayload resultPayload;
  private String datasetIdentifier = "FileDataset-ID-MKMeans";
  private String streamingIdentifier = "StreamingDataset-ID-MKMeans";
  private int WorkerNumber;
  private boolean initialized = false;
  private String initClusters; // filename where initial clusters are located
  double distanceThreshold; // will be entered with list of points/clusters
  DistanceMeasure measure; // will be entered with initialization
  List<Vector> points2Cluster = new ArrayList<Vector>(); // the points it's
                                                         // clustering
  long runTime; // quick and dirty timer...
  int runCount = 0; // loop and gather times
  int maxRun = 20; // how many times to run
  long totalRunTime;

  private Collection<Cluster> clusters;

  public KMeansMapper() {
    setDomain("Generic");
    setVersionInformation(200);
    setOperationIdentifier("MahoutMapper");
    resultPayload = ResultPayload.getInstance();
  }

  private void initializeMapper() {
    clusters = new ArrayList<Cluster>();
    Properties props = this.getProcessingDirectives();
    String dm = props.getProperty("Distance Measure");
    ClassLoader ccl = Thread.currentThread().getContextClassLoader();
    try {
      measure = ccl.loadClass(dm).asSubclass(DistanceMeasure.class)
          .newInstance();
    } catch (InstantiationException e) {
      e.printStackTrace();
    } catch (IllegalAccessException e) {
      e.printStackTrace();
    } catch (ClassNotFoundException e) {
      e.printStackTrace();
    }
    WorkerNumber = Integer.parseInt(props.getProperty("Worker Number"));
    distanceThreshold = Double.parseDouble(props
        .getProperty("Distance Threshold"));
    initClusters = props.getProperty("Initial Clusters");
    loadClusters(initClusters);
    System.out.println("Mapper done initializing!");
    initialized = true;
    runTime = 0;
    totalRunTime = 0;
  }

  /**
   * Load initial set of clusters from file
   * 
   * @param initClusters2
   *          File containing the initial set of clusters
   */
  private void loadClusters(String initClusters2) {
    System.out.println("Loading clusters in mapper from " + initClusters2);
    Configuration conf = new Configuration();
    Path clusterPath = new Path(initClusters2, "*");
    Collection<Path> result = new ArrayList<Path>();

    // filter out the files
    PathFilter clusterFileFilter = new PathFilter() {
      @Override
      public boolean accept(Path path) {
        return path.getName().startsWith("part");
      }
    };

    try {
      // get all filtered file names in result list
      FileSystem fs = clusterPath.getFileSystem(conf);
      FileStatus[] matches = fs.listStatus(FileUtil.stat2Paths(fs.globStatus(
          clusterPath, clusterFileFilter)), clusterFileFilter);

      for (FileStatus match : matches) {
        result.add(fs.makeQualified(match.getPath()));
      }

      // iterate thru the result path list
      for (Path path : result) {
        SequenceFile.Reader reader = null;
        try {
          reader = new SequenceFile.Reader(path.getFileSystem(conf), path, conf);
          Writable key;
          key = reader.getKeyClass().asSubclass(Writable.class).newInstance();
          Cluster value = new Cluster();
          while (reader.next(key, value)) {
            clusters.add(value);
            value = new Cluster();
          }

          reader.close();
        } catch (Exception e) {
          System.out.println("Error: " + e.getMessage());
          e.printStackTrace();
        } finally {
          IOUtils.quietClose(reader);
        }
      }

    } catch (IOException e) {
      throw new IllegalStateException(e);
    }
  }

  public void execute() {
    try {
      DatasetCollection dc = getDatasetCollection();
      if (!initialized) {
        System.out.println("Mapper not yet initialized...");
        FileAccess k = null;
        Dataset d = dc.getDataset(datasetIdentifier);
        if (d.getDatasetType() == Dataset.FILES)
          k = (FileAccess) d;
        else {
          System.out.println(WorkerNumber + ": Dataset issue: "
              + d.getDatasetType() + "initialized.  Returning...");
          return;
        }

        Enumeration fnames = k.getFilenames();

        while (fnames.hasMoreElements()) {
          String f = (String) fnames.nextElement();
          getPoints(f);
        }

        initializeMapper();
        this.changeToExecuteWhenDataAvailable();
        totalRunTime = System.nanoTime();
        HashMap<Integer, GranulesObservations> forReducer = DataforReducer(runIteration());
        writeResults(forReducer);
      }
      // now listen for streaming inputs

      StreamingAccess s = null;
      Dataset t = dc.getDataset(streamingIdentifier);
      if (t.getDatasetType() == Dataset.STREAMS)
        s = (StreamingAccess) t;
      else {
        System.out.println(WorkerNumber + ": Dataset issue: "
            + t.getDatasetType() + "initialized.  Returning...");
        return;
      }

      while (s.isDataAvailable()) {
        NBEvent nbEvent = s.getStreamingData();
        processResultsFromMap(nbEvent);
      }

    } catch (DatasetException e) {
      e.printStackTrace();
    } catch (IOException e) {
      System.out.println("Problems converting bytes to longs");
      e.printStackTrace();
    } catch (MarshallingException e) {
      System.out.println("Problems Marshalling: " + e.getMessage());
      e.printStackTrace();
    } catch (MapReduceException e) {
      System.out.println("Problems writing Results: " + e.getMessage());
      e.printStackTrace();
    } catch (ServiceException e) {
      e.printStackTrace();
    }

  }

  private void getPoints(String f) {
    // add point vectors to the point list. File f contains several points
    try {
      System.out.println("hello world");
      Configuration c = new Configuration();
      System.out.println("Configuration: " + c.toString());
      Path p = new Path(f);
      System.out.println("Path: " + p.toString());
      SequenceFile.Reader reader = new SequenceFile.Reader(p.getFileSystem(c),
          p, c);
      System.out.println("after reader");
      Writable key;
      key = (Writable) reader.getKeyClass().newInstance();
      VectorWritable value = new VectorWritable();
      while (reader.next(key, value)) {
        // get the point info
        points2Cluster.add(value.get());
        value = new VectorWritable();
      }
      // from sequenceTest
      /*
       * Writable key = (Writable)reader.getKeyClass().newInstance(); Writable
       * value = (Writable)reader.getValueClass().newInstance();
       * while(reader.next(key,value)){ System.out.println(key + "\t" + value);
       * } System.out.println(key.getClass().getName() + "\t" +
       * value.getClass().getName());
       */
      reader.close();

    } catch (Exception e) {
      System.out.println("Problem reading in vector file");
      e.printStackTrace();
    }
  }

  private void startRun(NBEvent nbEvent) throws IOException,
      MarshallingException, ServiceException {
    // contains clusters
    if (!initialized)
      initializeMapper();
    // from NBEvent???
    byte[] data = nbEvent.getContentPayload();
    Results results = ResultsFactory.getInstance().getResults(data);
    byte[] resultPayloadBytes = results.getResultPayload();
    ResultPayload rp = ResultPayload.getInstance();
    int length = Integer.parseInt(results.getRespondTo());
    byte[] decompressedBytes = decompressPayload(resultPayloadBytes, length);
    Object[] objArr = rp.getObjectArray(decompressedBytes);

    clusters.clear();
    for (Object c : objArr)
      clusters.add(((GranulesCluster) c).getCluster());
  }

  private byte[] decompressPayload(byte[] compressedPayload, int originalLength)
      throws ServiceException {
    Inflater decompresser = new Inflater();
    decompresser.setInput(compressedPayload);
    byte[] originalPayload = new byte[originalLength];
    try {
      int resultLength = decompresser.inflate(originalPayload);
    } catch (DataFormatException dfe) {
      throw new ServiceException("Worker " + WorkerNumber
          + " Decompression failed => " + dfe.toString());
    }
    return originalPayload;

  }

  private byte[] compressPayload(byte[] payload) {

    Deflater compresser = new Deflater();
    compresser.setInput(payload);
    compresser.finish();
    int payloadLength = payload.length;
    System.out.println("Payload length=" + payloadLength);

    byte[] output = new byte[payloadLength];
    int compressedDataLength = compresser.deflate(output);
    System.out.println("compressedDataLength=" + compressedDataLength);

    if (compressedDataLength == payloadLength) {
      return payload;
    }

    byte[] compressedBytes = new byte[compressedDataLength];
    System.arraycopy(output, 0, compressedBytes, 0, compressedDataLength);
    return compressedBytes;

  }

  private synchronized void processResultsFromMap(NBEvent nbEvent)
      throws IOException, MarshallingException, MapReduceException,
      ServiceException {

    String synopsis = (String) nbEvent.getContentSynopsis();
    if (synopsis.startsWith("Initialize") || synopsis.startsWith("ClusterInfo")) // contains
                                                                                 // clusters,
                                                                                 // maxIterations
                                                                                 // and
                                                                                 // distance
                                                                                 // measure,
    {
      startRun(nbEvent);
      HashMap<Integer, GranulesObservations> forReducer = DataforReducer(runIteration());
      writeResults(forReducer);
    } else // assume termination
    {
      
        //Code for timing information 
        long t = System.nanoTime() - totalRunTime;
        FileWriter fw = new FileWriter(new File("worker" + WorkerNumber + ".timing"), true);
        fw.write(runTime +"\n");
        fw.flush();
        fw.close();
        if(WorkerNumber == 0) { //only if it's the 0th one
            fw = new FileWriter(new File("overall_kmeans.timing"), true);
            fw.write(t + "\n");
            fw.flush();
            fw.close();
        }
       
      runTime = 0;

      if (runCount < maxRun) {// start again!
        initialized = false;
        clusters = null;
        System.gc();
        runCount++;
        initializeMapper();
        System.out.println("starting iteration on " + points2Cluster.size()
            + " points...");
        totalRunTime = System.nanoTime();
        HashMap<Integer, GranulesObservations> forReducer = DataforReducer(runIteration());
        System.out.println("Sending results to reducer");
        writeResults(forReducer);
      }
    }
    byte[] resultBytes = nbEvent.getContentPayload();
    Results results = ResultsFactory.getInstance().getResults(resultBytes);
    byte[] resultPayloadBytes = results.getResultPayload();

  }

  /**
   * Performs an iteration of clustering
   */
  private List<Cluster> runIteration() {
    long startTime = System.nanoTime();
    List<Cluster> results = new ArrayList<Cluster>();
    results.addAll(clusters);
    for (Vector point : points2Cluster) {
      Cluster nearestCluster = null;
      double nearestDistance = Double.MAX_VALUE;
      for (Cluster cluster : results) {
        Vector clusterCenter = cluster.getCenter();
        double length = clusterCenter.getLengthSquared();
        double distance = measure.distance(length, clusterCenter, point);
        if ((distance < nearestDistance) || (nearestCluster == null)) {
          nearestCluster = cluster;
          nearestDistance = distance;
        }
      }
      nearestCluster.observe(point, 1);
    }
    runTime += System.nanoTime() - startTime;
    return results;
  }

  /**
   * Converts data from <code>runIteration()</code> to the format the Reducer expects
   * @param clusters
   *             <code>List</code> of clusters with observations
   * @return
   *             <code>HashMap</code> for the reducer
   */
  private HashMap<Integer, GranulesObservations> DataforReducer(List<Cluster> clusters) {
    HashMap<Integer, GranulesObservations> results = new HashMap<Integer, GranulesObservations>();
    for (Cluster c : clusters) {
      if (c.getObservations().getS1() != null)
        results.put(c.getId(), new GranulesObservations(c.getObservations()));
    }
    return results;
  }

  private void writeResults(HashMap<Integer, GranulesObservations> map)
      throws MarshallingException, MapReduceException {
    boolean completeResult = true;
    ResultPayload resultPayload = ResultPayload.getInstance();

    byte[] resultPayloadBytes = resultPayload.getBytes(map);
    byte[] compressed = compressPayload(resultPayloadBytes);

    Results results = createResults(completeResult, false);
    results.setResultPayload(compressed);
    results.setResultType(ResultPayload.HASHMAP);
    results.setRespondTo(Integer.toString(resultPayloadBytes.length));
    writeResults(results);
  }

}
