/**
 * Software License, Version 1.0 Copyright 2010 Colorado State University. All
 * rights reserved. Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following conditions
 * are met: 1) All redistributions of source code must retain the above
 * copyright notice, the list of authors in the original source code, this list
 * of conditions and the disclaimer listed in this license; 2) All
 * redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the disclaimer listed in this license in the
 * documentation and/or other materials provided with the distribution; 3) Any
 * documentation included with all redistributions must include the following
 * acknowledgment: "This product includes software developed by the Department
 * of Computer Science at Colorado State University. For further information
 * contact the department at http://www.cs.colostate.edu/." Alternatively, this
 * acknowledgment may appear in the software itself, and wherever such
 * third-party acknowledgments normally appear. 4) The name Colorado State
 * University or Granules shall not be used to endorse or promote products
 * derived from this software without prior written permission from Colorado
 * State University. For written permission, please contact the Computer Science
 * Department at 1100 Center Avenue, Fort Collins, CO 80523-1873. 5) Products
 * derived from this software may not be called Granules, nor may Colorado State
 * University or Granules appear in their name, without prior written permission
 * of ARTI. Colorado State University provides no reassurances that the source
 * code provided does not infringe the patent or any other intellectual property
 * rights of any other entity. Colorado State University disclaims any liability
 * to any recipient for claims brought by any other entity based on infringement
 * of intellectual property rights or otherwise. LICENSEE UNDERSTANDS THAT
 * SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO WARRANTIES AS TO CAPABILITIES OR
 * ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES NO WARRANTIES AND MAKES NO
 * REPRESENTATION THAT SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT,
 * COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. COLORADO STATE UNIVERSITY MAKES NO
 * WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES",
 * "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE. LICENSEE ASSUMES THE ENTIRE
 * RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS, AND TO
 * THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING SOFTWARE.
 */
package ds.granules.learning.clustering.kmeans;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.FileUtil;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.fs.PathFilter;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.mahout.clustering.ClusterObservations;
import org.apache.mahout.clustering.kmeans.Cluster;
import org.apache.mahout.common.IOUtils;
import org.apache.mahout.common.distance.DistanceMeasure;
import org.apache.mahout.math.VectorWritable;

import cgl.narada.event.NBEvent;
import cgl.narada.service.ServiceException;
import cgl.narada.service.ServiceProperties;
import cgl.narada.service.compression.CompressionService;
import cgl.narada.service.qos.impl.QosServiceImpl;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * Reduce code for the Mahout K-Means clustering function
 * 
 * @author ericson
 */
public class KMeansReducer extends MapReduceBase {
  private int iterations = 0;
  private int MaxIters;
  private boolean isFaultTolerant = true;
  String workerType = "Reducer ";
  private DistanceMeasure measure;
  private double convergenceDelta;
  boolean initialized;
  HashMap<Integer, Cluster> currentClusters; // hashmap of clusterID, cluster
  boolean converged;
  String clusterFile;
  String synopsis = "ClusterInfo";
  long runTime;

  public KMeansReducer() {
    setDomain("Generic");
    setVersionInformation(200);
    setOperationIdentifier("KMeansMahoutReducer");
    initialized = false;
  }

  private void initialize() {
    System.out.println("in initialize()");
    runTime = 0;
    iterations = 0;
    converged = false;
    Properties p = this.getProcessingDirectives();
    currentClusters = new HashMap<Integer, Cluster>();
    String dm = p.getProperty("Distance Measure");
    ClassLoader ccl = Thread.currentThread().getContextClassLoader();
    try {
      measure = ccl.loadClass(dm).asSubclass(DistanceMeasure.class)
          .newInstance();
    } catch (InstantiationException e) {
      e.printStackTrace();
    } catch (IllegalAccessException e) {
      e.printStackTrace();
    } catch (ClassNotFoundException e) {
      e.printStackTrace();
    }
    MaxIters = Integer.parseInt(p.getProperty("Max Iterations"));
    convergenceDelta = Double.parseDouble(p.getProperty("Convergence Delta"));
    clusterFile = p.getProperty("Initial Clusters");
    loadInitialClusters(clusterFile);
    initialized = true;
  }

  /**
   * Read in initial clusters
   * 
   * @param clusterFile2
   *          File containing cluster information
   */
  private void loadInitialClusters(String clusterFile2) {
    Collection<Cluster> clusters = new ArrayList<Cluster>();
    Configuration conf = new Configuration();
    Path clusterPath = new Path(clusterFile2, "*");
    Collection<Path> result = new ArrayList<Path>();

    // filter out the files
    PathFilter clusterFileFilter = new PathFilter() {
      @Override
      public boolean accept(Path path) {
        return path.getName().startsWith("part");
      }
    };

    try {
      // get all filtered file names in result list
      FileSystem fs = clusterPath.getFileSystem(conf);
      FileStatus[] matches = fs.listStatus(FileUtil.stat2Paths(fs.globStatus(
          clusterPath, clusterFileFilter)), clusterFileFilter);

      for (FileStatus match : matches) {
        result.add(fs.makeQualified(match.getPath()));
      }

      // iterate thru the result path list
      for (Path path : result) {
        // Check if we are restarting partway through: if the input file matches part-r-***, *** is the iteration we stopped on
        String s = path.getName();
        String[] broken = s.split("-");
        try {
           int tmp = Integer.parseInt(broken[broken.length-1]);
           iterations = tmp;
        }
        catch (NumberFormatException e) {
          // Not formatted for a reload, continue on
          System.out.println("Not formatted for a reload, continue on");
        }
        SequenceFile.Reader reader = null;
        try {
          reader = new SequenceFile.Reader(fs, path, conf);
          Class<?> valueClass = reader.getValueClass();
          Writable key;
          try {
            key = reader.getKeyClass().asSubclass(Writable.class).newInstance();
          } catch (InstantiationException e) { // Should not be possible
            throw new IllegalStateException(e);
          } catch (IllegalAccessException e) {
            throw new IllegalStateException(e);
          }
          if (valueClass.equals(Cluster.class)) {
            Cluster value = new Cluster();
            while (reader.next(key, value)) {
              // get the cluster info
              clusters.add(value);
              value = new Cluster();
            }
          }
        } finally {
          IOUtils.quietClose(reader);
        }
      }
    } catch (IOException e) {
      throw new IllegalStateException(e);
    }

    // move to hashmap
    for (Cluster c : clusters) {
      currentClusters.put(c.getId(), c);
    }

  }

  public void execute() {
    try {
      System.out.println("in execute()");
      if (!initialized)
        initialize();
      StreamingAccess mapReduceDataset = getMapReduceDataset();

      while (mapReduceDataset.isDataAvailable()) {
        NBEvent nbEvent = mapReduceDataset.getStreamingData();
        processResultsFromMap(nbEvent);
      }

    } catch (DatasetException e) {
      e.printStackTrace();
    } catch (IOException e) {
      System.out.println("Problems converting bytes to longs");
      e.printStackTrace();
    } catch (MarshallingException e) {
      System.out.println("Problems Marshalling: " + e.getMessage());
      e.printStackTrace();
    } catch (MapReduceException e) {
      System.out.println("Problems writing Results: " + e.getMessage());
      e.printStackTrace();
    } catch (NumberFormatException e) {
      e.printStackTrace();
    } catch (ServiceException e) {
      e.printStackTrace();
    }

  }

  private synchronized void processResultsFromMap(NBEvent nbEvent)
      throws IOException, MarshallingException, MapReduceException,
      NumberFormatException, ServiceException {
    if (!initialized)
      initialize();
    String s = (String) nbEvent.getContentSynopsis();
    byte[] resultBytes = nbEvent.getContentPayload();
    Results results = ResultsFactory.getInstance().getResults(resultBytes);
    byte[] resultPayloadBytes = results.getResultPayload();
    byte[] decompressedBytes = decompressPayload(resultPayloadBytes, Integer
        .parseInt(results.getRespondTo()));
    ResultPayload resultPayload = ResultPayload.getInstance();

    if (results.getResultType() != ResultPayload.HASHMAP) {
      System.out
          .println("Mahout KMeans Reducer: Incorrect ResultType received");
      return;
    }

    /**
     * Remainder of processing assumes that the resultPayload type is of type
     * ResultPayload.INT
     */
    long startTime = System.nanoTime();
    HashMap<Integer, GranulesObservations> incomingMap = (HashMap<Integer, GranulesObservations>) resultPayload
        .getHashMap(decompressedBytes);
     //cummulativeValue += value;

    setReceivedOutputFrom(results.getInstanceIdentifier());

    handleIncoming(incomingMap);
    runTime += System.nanoTime() - startTime;

    if (!isAwaitingOutputs()) {
      resetOutputTracker();
      iterations++;

      List<Cluster> clusters = new ArrayList<Cluster>();
      for (Cluster c : currentClusters.values()) {
        clusters.add(c);
      }
      startTime = System.nanoTime();
      List<Cluster> newClusters = calcNewClusters(clusters);
      converged = testConvergence(newClusters);
      updateMap(newClusters);
      runTime += System.nanoTime() - startTime;
      //simple fault tolerance, write out after every iteration
      if (!converged && iterations < MaxIters) {
        writeResult(newClusters, converged);
        if(isFaultTolerant) {
          writeToFile();
        }
      }
    
      else {
        // start again
        initialized = false;
        // Code to store timing information
        FileWriter fw = new FileWriter(new File("Reducer.timing"),
            true);
        fw.write(runTime + "\n");
        fw.flush();
        fw.close();
        
        byte[] rpb = resultPayload.getBytes("Done");

        Results r = createResults(true, false);
        r.setResultPayload(rpb);
        r.setResultType(ResultPayload.STRING);

        // writeResults(synopsis,results);
        this.writeResults("Termination", r);
        writeToFile();
        currentClusters = null;
        System.gc();
      }
    }

  }

  /**
   * Save out final cluster information
   */
  private void writeToFile() {
    Configuration c = new Configuration();
    Path p = new Path("part-r-" + iterations);
    SequenceFile.Writer writer;
    try {
      writer = new SequenceFile.Writer(p.getFileSystem(c), c, p, Text.class,
          Cluster.class);
      Writable w;
      for (Cluster clus : currentClusters.values()) {
        w = new Text(clus.getIdentifier());
        writer.append(w, clus);
      }

      writer.close();
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  private void updateMap(List<Cluster> clusters) {
    for (Cluster c : clusters) {
      currentClusters.put(c.getId(), c);
    }
  }

  private void writeResult(List<Cluster> data, boolean processingComplete)
      throws MarshallingException, MapReduceException, IOException {

    List<GranulesCluster> sending = new ArrayList<GranulesCluster>();
    for (Cluster c : data)
      sending.add(new GranulesCluster(c));

    ResultPayload resultPayload = ResultPayload.getInstance();

    Object[] intermediate = sending.toArray();
    byte[] resultPayloadBytes = resultPayload.getBytes(intermediate);

    byte[] compressedBytes = compressPayload(resultPayloadBytes);
    Results results = createResults(true, processingComplete);
    results.setResultPayload(compressedBytes);
    results.setResultType(ResultPayload.OBJECT_ARRAY);
    results.setRespondTo(Integer.toString(resultPayloadBytes.length));
    writeResults(synopsis, results);
  }

  private byte[] compressPayload(byte[] payload) {

    Deflater compresser = new Deflater();
    compresser.setInput(payload);
    compresser.finish();
    int payloadLength = payload.length;

    byte[] output = new byte[payloadLength];
    int compressedDataLength = compresser.deflate(output);

    if (compressedDataLength == payloadLength) {
      return payload;
    }

    byte[] compressedBytes = new byte[compressedDataLength];
    System.arraycopy(output, 0, compressedBytes, 0, compressedDataLength);
    return compressedBytes;

  }

  private byte[] decompressPayload(byte[] compressedPayload, int originalLength)
      throws ServiceException {
    Inflater decompresser = new Inflater();
    decompresser.setInput(compressedPayload);
    byte[] originalPayload = new byte[originalLength];
    try {
      int resultLength = decompresser.inflate(originalPayload);
    } catch (DataFormatException dfe) {
      throw new ServiceException("Reducer Decompression failed => "
          + dfe.toString());
    }
    return originalPayload;

  }

  private List<Cluster> calcNewClusters(List<Cluster> oldClusters) {
    int convergeCount = 0;
    List<Cluster> newClusters = new ArrayList<Cluster>();
    for (Cluster c : oldClusters) {
      boolean converged = computeConvergence(c);
      if (converged) {
        // System.out.println("Cluster " + c.getId() + " has converged!");
        convergeCount++;
      }
      c.computeParameters();
      newClusters.add(c);
    }
    System.out.println(convergeCount + " Clusters have converged");
    return newClusters;
  }

  private void handleIncoming(HashMap<Integer, GranulesObservations> data) {
    for (int key : data.keySet()) {
      Cluster t = currentClusters.get(key);
      if (t.equals(null))
        System.out.println("Error: can't find cluster with id: " + key);
      else {
        t.observe(data.get(key).toClusterObservations());
        currentClusters.put(key, t);
      }
    }
  }

  private boolean testConvergence(List<Cluster> clusList) {
    boolean done = true;
    for (Cluster c : clusList) {
      if (!c.isConverged())
        return false;
    }

    return done;
  }

  private boolean computeConvergence(Cluster c) {
    return c.computeConvergence(measure, convergenceDelta);
  }

}
