/**
 * Software License, Version 1.0 Copyright 2010 Colorado State University. All
 * rights reserved. Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following conditions
 * are met: 1) All redistributions of source code must retain the above
 * copyright notice, the list of authors in the original source code, this list
 * of conditions and the disclaimer listed in this license; 2) All
 * redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the disclaimer listed in this license in the
 * documentation and/or other materials provided with the distribution; 3) Any
 * documentation included with all redistributions must include the following
 * acknowledgment: "This product includes software developed by the Department
 * of Computer Science at Colorado State University. For further information
 * contact the department at http://www.cs.colostate.edu/." Alternatively, this
 * acknowledgment may appear in the software itself, and wherever such
 * third-party acknowledgments normally appear. 4) The name Colorado State
 * University or Granules shall not be used to endorse or promote products
 * derived from this software without prior written permission from Colorado
 * State University. For written permission, please contact the Computer Science
 * Department at 1100 Center Avenue, Fort Collins, CO 80523-1873. 5) Products
 * derived from this software may not be called Granules, nor may Colorado State
 * University or Granules appear in their name, without prior written permission
 * of ARTI. Colorado State University provides no reassurances that the source
 * code provided does not infringe the patent or any other intellectual property
 * rights of any other entity. Colorado State University disclaims any liability
 * to any recipient for claims brought by any other entity based on infringement
 * of intellectual property rights or otherwise. LICENSEE UNDERSTANDS THAT
 * SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO WARRANTIES AS TO CAPABILITIES OR
 * ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES NO WARRANTIES AND MAKES NO
 * REPRESENTATION THAT SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT,
 * COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. COLORADO STATE UNIVERSITY MAKES NO
 * WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES",
 * "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE. LICENSEE ASSUMES THE ENTIRE
 * RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS, AND TO
 * THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING SOFTWARE.
 */

package ds.granules.learning.clustering.lda;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.io.Writable;
import org.apache.mahout.clustering.lda.LDAInference;
import org.apache.mahout.clustering.lda.LDAState;
import org.apache.mahout.common.IntPairWritable;
import org.apache.mahout.common.distance.DistanceMeasure;
import org.apache.mahout.math.DenseMatrix;
import org.apache.mahout.math.Vector;
import org.apache.mahout.math.VectorWritable;

import cgl.narada.event.NBEvent;
import cgl.narada.service.ServiceException;
import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.FileAccess;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * An example demonstrating the <code>Map</code> role within a Word Count
 * operation.
 * 
 * @author Shrideep Pallickara
 */
public class LDAMapper extends MapReduceBase {
  private String datasetIdentifier = "FileDataset-ID-MKMeans";
  private String streamingIdentifier = "StreamingDataset-ID-MKMeans";
  private int WorkerNumber;
  private boolean initialized = false;

  double distanceThreshold; // will be entered with list of points/clusters
  DistanceMeasure measure; // will be entered with initialization
  ArrayList<VectorWritable> points2Cluster = new ArrayList<VectorWritable>(); // the
                                                                              // points
                                                                              // it's
                                                                              // clustering
  long runTime; // quick and dirty timer...
  int runCount = 0; // loop and gather times
  int maxRun = 50; // how many times to run
  LDAState state;
  LDAInference infer;
  String statePath;
  long totalRunTime;
  long compressionTime;
  long decompressionTime;

  // private List<SoftCluster> clusters;

  public LDAMapper() {
    setDomain("Generic");
    setVersionInformation(200);
    setOperationIdentifier("MahoutMapper");
  }

  /**
   * Initializes the mapper with inputs from the deployer
   */
  private void initializeMapper() {
    Properties props = this.getProcessingDirectives();
    WorkerNumber = Integer.parseInt(props.getProperty("Worker Number"));
    statePath = props.getProperty("State Path");
    int numTopics = Integer.parseInt(props.getProperty("Num Topics"));
    int numWords = Integer.parseInt(props.getProperty("Num Words"));
    double smoothing = Double.parseDouble(props.getProperty("Smoothing"));
    loadState(statePath, numTopics, numWords, smoothing);
    this.infer = new LDAInference(state);
    System.out.println("Mapper done initializing!");
    initialized = true;
    runTime = 0;
    compressionTime = 0;
    decompressionTime = 0;
    totalRunTime = System.nanoTime();
  }

  /**
   * Loads the initial state information
   * 
   * @param initClusters2
   *          File containing the list of initial clusters
   * @param numTopics
   *          Number of topics to cluster around
   * @param numWords
   *          Number of words (dimensions) expected
   * @param smoothing
   *          Smoothing value
   */
  private void loadState(String initClusters2, int numTopics, int numWords,
      double smoothing) {

    Configuration conf = new Configuration();
    Path statePath = new Path(initClusters2, "*");

    DenseMatrix pWgT = new DenseMatrix(numTopics, numWords);

    double[] logTotals = new double[numTopics];
    double ll = 0.0;
    try {
      FileSystem fs = statePath.getFileSystem(conf);

      IntPairWritable key = new IntPairWritable();
      DoubleWritable value = new DoubleWritable();
      for (FileStatus status : fs.globStatus(new Path(statePath, "part-*"))) {
        Path path = status.getPath();

        SequenceFile.Reader reader = new SequenceFile.Reader(fs, path, conf);
        while (reader.next(key, value)) {
          int topic = key.getFirst();
          int word = key.getSecond();
          if (word == -1) {// TOPIC_SUM_KEY) {
            logTotals[topic] = value.get();

          } else if (topic == -2) {// LOG_LIKELIHOOD_KEY) {
            ll = value.get();
          } else {
            pWgT.setQuick(topic, word, value.get());
          }
        }
        reader.close();
      }
    } catch (Exception e) {
      System.out.println("Error: " + e.getMessage());
      e.printStackTrace();
    }
    this.state = new LDAState(numTopics, numWords, smoothing, pWgT, logTotals,
        ll);
  }

  public void execute() {
    try {
      DatasetCollection dc = getDatasetCollection();
      if (!initialized) {
        System.out.println("Mapper not yet initialized...");
        FileAccess k = null;
        Dataset d = dc.getDataset(datasetIdentifier);
        if (d.getDatasetType() == Dataset.FILES)
          k = (FileAccess) d;
        else {
          System.out.println(WorkerNumber + ": Dataset issue: "
              + d.getDatasetType() + "initialized.  Returning...");
          return;
        }

        Enumeration fnames = k.getFilenames();

        while (fnames.hasMoreElements()) {
          String f = (String) fnames.nextElement();
          getPoints(f);
        }

        initializeMapper();
        this.changeToExecuteWhenDataAvailable();
        HashMap<IntPairWritable, Double> forReducer = runIteration();
        writeResults(forReducer);
      }
      // now listen for streaming inputs

      StreamingAccess s = null;
      Dataset t = dc.getDataset(streamingIdentifier);
      if (t.getDatasetType() == Dataset.STREAMS)
        s = (StreamingAccess) t;
      else {
        System.out.println(WorkerNumber + ": Dataset issue: "
            + t.getDatasetType() + "initialized.  Returning...");
        return;
      }

      while (s.isDataAvailable()) {
        NBEvent nbEvent = s.getStreamingData();
        processResultsFromReducer(nbEvent);
      }

    } catch (DatasetException e) {
      e.printStackTrace();
    } catch (IOException e) {
      System.out.println("Problems converting bytes to longs");
      e.printStackTrace();
    } catch (MarshallingException e) {
      System.out.println("Problems Marshalling: " + e.getMessage());
      e.printStackTrace();
    } catch (MapReduceException e) {
      System.out.println("Problems writing Results: " + e.getMessage());
      e.printStackTrace();
    } catch (NumberFormatException e) {
      e.printStackTrace();
    } catch (ServiceException e) {
      e.printStackTrace();
    }

  }

  /**
   * Loads the points this mapper is responsible for clustering from file
   * 
   * @param f
   *          File which contains points to be clustered
   */
  private void getPoints(String f) {
    try {
      Configuration c = new Configuration();
      Path p = new Path(f);
      SequenceFile.Reader reader = new SequenceFile.Reader(p.getFileSystem(c),
          p, c);
      Writable key;
      key = (Writable) reader.getKeyClass().newInstance();
      VectorWritable value = new VectorWritable();
      while (reader.next(key, value)) {
        // get the point info
        points2Cluster.add(value);
        value = new VectorWritable();
      }

      reader.close();

    } catch (Exception e) {
      System.out.println("Problem reading in vector file");
      e.printStackTrace();
    }
  }

  private void startRun(NBEvent nbEvent) throws IOException,
      MarshallingException, NumberFormatException, ServiceException {
    // contains clusters
    if (!initialized)
      initializeMapper();
    byte[] data = nbEvent.getContentPayload();
    Results results = ResultsFactory.getInstance().getResults(data);
    byte[] resultPayloadBytes = results.getResultPayload();

    byte[] decompressed = decompressPayload(resultPayloadBytes, Integer
        .parseInt(results.getRespondTo()));

    ResultPayload rp = ResultPayload.getInstance();

    HashMap<IntPairWritable, Double> newVals = (HashMap<IntPairWritable, Double>) rp
        .getHashMap(decompressed);
    resetStateandInfer(newVals);

  }

  /**
   * Perform the clustering step
   * @param newVals
   *             Set of new data from the reducer
   */
  private void resetStateandInfer(HashMap<IntPairWritable, Double> newVals) {
    double[] logTotals = new double[state.getNumTopics()];
    DenseMatrix pWgT = new DenseMatrix(state.getNumTopics(), state
        .getNumWords());
    double ll = 0;
    for (IntPairWritable key : newVals.keySet()) {
      int word = key.getSecond();
      if (word == -1)
        logTotals[key.getFirst()] = newVals.get(key);
      else if (word == -2)
        ll = newVals.get(key);
      else
        pWgT.setQuick(key.getFirst(), word, newVals.get(key));

    }
    this.state = new LDAState(state.getNumTopics(), state.getNumWords(), state
        .getTopicSmoothing(), pWgT, logTotals, ll);
    this.infer = new LDAInference(state);
  }

  private synchronized void processResultsFromReducer(NBEvent nbEvent)
      throws IOException, MarshallingException, MapReduceException,
      NumberFormatException, ServiceException {

    String synopsis = (String) nbEvent.getContentSynopsis();
    Results results = ResultsFactory.getInstance().getResults(
        nbEvent.getContentPayload());
    long timeStamp = results.getTimestamp();
    long diff = System.currentTimeMillis() - timeStamp;
    System.out
        .println("Worker nbEvent from Reducer - communications took about "
            + diff + " milliseconds.");
    if (synopsis.startsWith("Initialize") || synopsis.startsWith("ClusterInfo")) // contains
                                                                                 // clusters,
                                                                                 // maxIterations
                                                                                 // and
                                                                                 // distance
                                                                                 // measure,
    {
      long startTime = System.nanoTime();
      startRun(nbEvent);
      HashMap<IntPairWritable, Double> forReducer = runIteration();
      runTime += System.nanoTime() - startTime;
      writeResults(forReducer);
    } else // assume termination
    {
      // Timing Code
      long tmpTime = System.nanoTime() - totalRunTime;
      FileWriter fw = new FileWriter(new File("worker"
          + WorkerNumber + ".timing"), true);
      fw.write(runTime + "\t" + compressionTime + "\t" + decompressionTime
          + "\n");
      fw.flush();
      fw.close();
      if (WorkerNumber == 0) {
        fw = new FileWriter(new File("overall_lda.timing"), true);
        fw.write(tmpTime + "\n");
        fw.flush();
        fw.close();
      }
      runTime = 0;

      if (runCount < maxRun) {// start again
        initialized = false;
        state = null;
        System.gc();
        runCount++;
        initializeMapper();
        System.out.println("starting iteration on " + points2Cluster.size()
            + " points...");
        totalRunTime = System.nanoTime();
        HashMap<IntPairWritable, Double> forReducer = runIteration();
        System.out.println("Sending results to reducer");
        writeResults(forReducer);
      }
    }
  }

  /**
   * Start an iteration of clustering
   * @return
   */
  private HashMap<IntPairWritable, Double> runIteration() {
    System.out.println("in runIteration()");

    HashMap<IntPairWritable, Double> results = new HashMap<IntPairWritable, Double>();

    for (VectorWritable point : points2Cluster) {
      Vector wordCounts = point.get();
      LDAInference.InferredDocument doc = null;
      try {
        doc = infer.infer(wordCounts);
      } catch (Exception e) {
        System.out.println("Couldn't infer document: " + e.getMessage());
        e.printStackTrace();
      }
      double[] logTotals = new double[state.getNumTopics()];
      Arrays.fill(logTotals, Double.NEGATIVE_INFINITY);

      // Output sufficient statistics for each word. == pseudo-log counts.
      double v;
      for (Iterator<Vector.Element> iter = wordCounts.iterateNonZero(); iter
          .hasNext();) {
        Vector.Element e = iter.next();
        int w = e.index();

        for (int k = 0; k < state.getNumTopics(); ++k) {
          v = (doc.phi(k, w) + Math.log(e.get()));

          IntPairWritable kw = new IntPairWritable(k, w);

          // ouput (topic, word)'s logProb contribution
          results.put(kw, v);
          logTotals[k] = logSum(logTotals[k], v);
        }
      }

      // Output the totals for the statistics. This is to make
      // normalizing a lot easier.
      for (int k = 0; k < state.getNumTopics(); ++k) {
        IntPairWritable kw = new IntPairWritable(k, -1);
        v = (logTotals[k]);

        results.put(kw, v);
      }
      IntPairWritable llk = new IntPairWritable(-2, -2);
      // Output log-likelihoods.
      v = (doc.getLogLikelihood());
      if (results.containsKey(llk))
        results.put(llk, v + results.get(llk));
      else
        results.put(llk, v);
    }

    return results;
  }

  // from LDAUtil.java
  private double logSum(double a, double b) {
    /**
     * @return log(exp(a) + exp(b))
     */
    return a == Double.NEGATIVE_INFINITY ? b
        : b == Double.NEGATIVE_INFINITY ? a : a < b ? b
            + Math.log(1 + Math.exp(a - b)) : a + Math.log(1 + Math.exp(b - a));

  }

  /**
   * Send results to the reducer
   * @param forReducer
   *             Cluster updates to be sent to the reducer
   * @throws MarshallingException
   *             Thrown if there are problems marshalling forReducer
   * @throws MapReduceException
   *             Thrown if there are any communications problems
   */
  private void writeResults(HashMap<IntPairWritable, Double> forReducer)
      throws MarshallingException, MapReduceException {
    boolean completeResult = true;
    ResultPayload resultPayload = ResultPayload.getInstance();

    byte[] resultPayloadBytes = resultPayload.getBytes(forReducer);
    byte[] compressedBytes = compressPayload(resultPayloadBytes);

    Results results = createResults(completeResult, false);

    results.setResultPayload(compressedBytes);
    results.setRespondTo(Integer.toString(resultPayloadBytes.length));

    results.setResultType(ResultPayload.HASHMAP);

    writeResults(results);
  }

  // Compression and Decompression code
  private byte[] compressPayload(byte[] payload) {
    long startTime = System.nanoTime();
    Deflater compresser = new Deflater(Deflater.BEST_SPEED);
    compresser.setInput(payload);
    compresser.finish();
    int payloadLength = payload.length;
    System.out.println("Payload length=" + payloadLength);

    byte[] output = new byte[payloadLength];
    int compressedDataLength = compresser.deflate(output);
    System.out.println("compressedDataLength=" + compressedDataLength);

    if (compressedDataLength == payloadLength) {
      return payload;
    }

    byte[] compressedBytes = new byte[compressedDataLength];
    System.arraycopy(output, 0, compressedBytes, 0, compressedDataLength);
    compressionTime += System.nanoTime() - startTime;
    return compressedBytes;

  }

  private byte[] decompressPayload(byte[] compressedPayload, int originalLength)
      throws ServiceException {
    long startTime = System.nanoTime();
    Inflater decompresser = new Inflater();
    decompresser.setInput(compressedPayload);
    byte[] originalPayload = new byte[originalLength];
    try {
      int resultLength = decompresser.inflate(originalPayload);
    } catch (DataFormatException dfe) {
      throw new ServiceException("Reducer Decompression failed => "
          + dfe.toString());
    }
    decompressionTime += System.nanoTime() - startTime;
    return originalPayload;

  }

}
