/**
 * Software License, Version 1.0 Copyright 2010 Colorado State University. All
 * rights reserved. Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following conditions
 * are met: 1) All redistributions of source code must retain the above
 * copyright notice, the list of authors in the original source code, this list
 * of conditions and the disclaimer listed in this license; 2) All
 * redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the disclaimer listed in this license in the
 * documentation and/or other materials provided with the distribution; 3) Any
 * documentation included with all redistributions must include the following
 * acknowledgment: "This product includes software developed by the Department
 * of Computer Science at Colorado State University. For further information
 * contact the department at http://www.cs.colostate.edu/." Alternatively, this
 * acknowledgment may appear in the software itself, and wherever such
 * third-party acknowledgments normally appear. 4) The name Colorado State
 * University or Granules shall not be used to endorse or promote products
 * derived from this software without prior written permission from Colorado
 * State University. For written permission, please contact the Computer Science
 * Department at 1100 Center Avenue, Fort Collins, CO 80523-1873. 5) Products
 * derived from this software may not be called Granules, nor may Colorado State
 * University or Granules appear in their name, without prior written permission
 * of ARTI. Colorado State University provides no reassurances that the source
 * code provided does not infringe the patent or any other intellectual property
 * rights of any other entity. Colorado State University disclaims any liability
 * to any recipient for claims brought by any other entity based on infringement
 * of intellectual property rights or otherwise. LICENSEE UNDERSTANDS THAT
 * SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO WARRANTIES AS TO CAPABILITIES OR
 * ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES NO WARRANTIES AND MAKES NO
 * REPRESENTATION THAT SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT,
 * COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. COLORADO STATE UNIVERSITY MAKES NO
 * WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES",
 * "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE. LICENSEE ASSUMES THE ENTIRE
 * RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS, AND TO
 * THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING SOFTWARE.
 */
package ds.granules.learning.clustering.lda;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.Properties;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.io.Writable;
import org.apache.mahout.clustering.dirichlet.DirichletState;
import org.apache.mahout.common.IntPairWritable;

import cgl.narada.event.NBEvent;
import cgl.narada.service.ServiceException;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * The Reducer for the Latent Dirichlet Allocation (LDA) clustering algorithm
 * with Mahout
 * 
 * @author ericson
 */
public class LDAReducer extends MapReduceBase {
  private int iterations = 0;
  private int MaxIters;
  private int loop = 0;
  String workerType = "Reducer ";
  boolean initialized;
  HashMap<IntPairWritable, Double> currentClusters; // hashmap of clusterID,
                                                    // cluster
  // boolean converged;
  String synopsis = "ClusterInfo";
  long runTime;
  long decompressTime;
  long compressTime;
  private DirichletState state;
  
  private boolean isFaultTolerant = true;

  public LDAReducer() {
    setDomain("Generic");
    setVersionInformation(200);
    setOperationIdentifier("KMeansMahoutReducer");
    initialized = false;
  }

  private void initialize() {
    System.out.println("in initialize()");
    runTime = 0;
    iterations = 0;
    compressTime = 0;
    decompressTime = 0;
    Properties p = this.getProcessingDirectives();
    currentClusters = new HashMap<IntPairWritable, Double>();
    MaxIters = Integer.parseInt(p.getProperty("Max Iterations"));
    initialized = true;
  }

  public void execute() {
    try {
      System.out.println("in execute()");
      if (!initialized)
        initialize();
      StreamingAccess mapReduceDataset = getMapReduceDataset();

      while (mapReduceDataset.isDataAvailable()) {
        NBEvent nbEvent = mapReduceDataset.getStreamingData();
        processResultsFromMap(nbEvent);
      }

    } catch (DatasetException e) {
      e.printStackTrace();
    } catch (IOException e) {
      System.out.println("Problems converting bytes to longs");
      e.printStackTrace();
    } catch (MarshallingException e) {
      System.out.println("Problems Marshalling: " + e.getMessage());
      e.printStackTrace();
    } catch (MapReduceException e) {
      System.out.println("Problems writing Results: " + e.getMessage());
      e.printStackTrace();
    } catch (NumberFormatException e) {
      e.printStackTrace();
    } catch (ServiceException e) {
      e.printStackTrace();
    }

  }

  private synchronized void processResultsFromMap(NBEvent nbEvent)
      throws IOException, MarshallingException, MapReduceException,
      NumberFormatException, ServiceException {

    if (!initialized)
      initialize();
    byte[] resultBytes = nbEvent.getContentPayload();
    Results results = ResultsFactory.getInstance().getResults(resultBytes);
    byte[] resultPayloadBytes = results.getResultPayload();
    byte[] decompressed = decompressPayload(resultPayloadBytes, Integer
        .parseInt(results.getRespondTo()));
    ResultPayload resultPayload = ResultPayload.getInstance();

    if (results.getResultType() != ResultPayload.HASHMAP) {
      System.out
          .println("Mahout KMeans Reducer: Incorrect ResultType received");
      return;
    }

    /**
     * Remainder of processing assumes that the resultPayload type is of type
     * ResultPayload.INT
     */
    long startTime = System.nanoTime();
    HashMap<IntPairWritable, Double> incomingMap = (HashMap<IntPairWritable, Double>) resultPayload
        .getHashMap(decompressed);
    // cummulativeValue += value;

    setReceivedOutputFrom(results.getInstanceIdentifier());

    handleIncoming(incomingMap);
    runTime += System.nanoTime() - startTime;

    if (!isAwaitingOutputs()) {
      resetOutputTracker();
      iterations++;

      if (iterations < MaxIters) {
        writeResult(currentClusters, false);
        if(isFaultTolerant) {
          writeToFile();
        }
      }
      else {
        long tmp = System.nanoTime();
        loop++;
        long fileTime = System.nanoTime() - tmp;
        // start again
        initialized = false;
        
        // Store timing data 
        FileWriter fw = new FileWriter(new File("Reducer.timing"), true);
        fw.write(runTime + "\t" + compressTime + "\t" + decompressTime
          + "\t" + fileTime + "\n");
        fw.flush();
        fw.close();
       
        byte[] rpb = resultPayload.getBytes("Done");

        Results r = createResults(true, false);
        r.setResultPayload(rpb);
        r.setResultType(ResultPayload.STRING);

        this.writeResults("Termination", r);
        state = null;
        writeToFile();
        System.gc();
      }
      currentClusters.clear();
    }

  }

  /**
   * Write cluster data to disk
   */
  private void writeToFile() {
    Configuration c = new Configuration();
    Path p = new Path("part-r-" + iterations);
    SequenceFile.Writer writer;
    Writable w;
    try {
      writer = new SequenceFile.Writer(p.getFileSystem(c), c, p,
          IntPairWritable.class, DoubleWritable.class);
      for (IntPairWritable key : currentClusters.keySet()) {
        w = new DoubleWritable(currentClusters.get(key));
        writer.append(key, w);
      }

      writer.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  /**
   * Write results to the mappers
   * 
   * @param currentClusters2
   *          The set of current clusters
   * @param processingComplete
   *          Have we hit the last iteration
   * @throws MarshallingException
   *           If there is a problem converting the clusters to a
   *           <code>byte[]</code>
   * @throws MapReduceException
   *           If there is a problem communicating with the mappers
   * @throws IOException
   *           If there is any other IO problem
   */
  private void writeResult(HashMap<IntPairWritable, Double> currentClusters2,
      boolean processingComplete) throws MarshallingException,
      MapReduceException, IOException {

    ResultPayload resultPayload = ResultPayload.getInstance();
    byte[] resultPayloadBytes = resultPayload.getBytes(currentClusters2);
    byte[] compressed = compressPayload(resultPayloadBytes);

    Results results = createResults(true, processingComplete);
    results.setResultPayload(compressed);
    results.setRespondTo(Integer.toString(resultPayloadBytes.length));
    results.setResultType(ResultPayload.OBJECT_ARRAY);

    writeResults(synopsis, results);
  }

  private void handleIncoming(HashMap<IntPairWritable, Double> incomingMap) {
    for (IntPairWritable key : incomingMap.keySet()) {
      if (key.getSecond() == -2) // sum likelihoods
      {
        if (currentClusters.containsKey(key))
          currentClusters.put(key, currentClusters.get(key)
              + incomingMap.get(key));
        else
          currentClusters.put(key, incomingMap.get(key));
      } else // log sum sufficient statistics?
      {
        double accum;
        if (currentClusters.containsKey(key))
          accum = currentClusters.get(key);
        else
          accum = Double.NEGATIVE_INFINITY;
        currentClusters.put(key, logSum(accum, incomingMap.get(key)));
      }
    }
  }

  // from LDAUtil.java
  private double logSum(double a, double b) {
    /**
     * @return log(exp(a) + exp(b))
     */
    return a == Double.NEGATIVE_INFINITY ? b
        : b == Double.NEGATIVE_INFINITY ? a : a < b ? b
            + Math.log(1 + Math.exp(a - b)) : a + Math.log(1 + Math.exp(b - a));

  }

  // Compression/decompression code
  private byte[] compressPayload(byte[] payload) {
    long startTime = System.nanoTime();
    Deflater compresser = new Deflater(Deflater.BEST_SPEED);
    compresser.setInput(payload);
    compresser.finish();
    int payloadLength = payload.length;
    System.out.println("Payload length=" + payloadLength);

    byte[] output = new byte[payloadLength];
    int compressedDataLength = compresser.deflate(output);
    System.out.println("compressedDataLength=" + compressedDataLength);

    if (compressedDataLength == payloadLength) {
      return payload;
    }

    byte[] compressedBytes = new byte[compressedDataLength];
    System.arraycopy(output, 0, compressedBytes, 0, compressedDataLength);
    compressTime += System.nanoTime() - startTime;
    return compressedBytes;

  }

  private byte[] decompressPayload(byte[] compressedPayload, int originalLength)
      throws ServiceException {
    long startTime = System.nanoTime();
    Inflater decompresser = new Inflater();
    decompresser.setInput(compressedPayload);
    byte[] originalPayload = new byte[originalLength];
    try {
      int resultLength = decompresser.inflate(originalPayload);
    } catch (DataFormatException dfe) {
      throw new ServiceException("Reducer Decompression failed => "
          + dfe.toString());
    }
    decompressTime += System.nanoTime() - startTime;
    return originalPayload;

  }

}
