/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.naadsm;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;

import cgl.narada.event.NBEvent;
import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.naadsm.resplit.event.OutputVariableValue;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.ProcessingException;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * This is the output variable collector of distributed NAADSM. Once output
 * variable is detected in Controller, it will be sent to the output variable
 * collector. The collector is responsible for collect and organize these
 * variables and write them into disk with some fixed format.
 * 
 * @author simonsui
 * 
 */
public class OutputVariableCollector extends MapReduceBase {
	String outputFileName;
	static String[] variableNames = { "", "mngnUAll", "mngArea", "adqnUAll",
			"adqDistAvg", "latPeriodAvg", "detnUAll", "trnUAll", "desnUAll",
			"deswUAll", "adqnUDir", "adqnUInd", "adqnUAir", "vacnUAll",
			"vacwUAll", "zoneArea" };
	HashMap<String, Double> map;
	int dayNumber;

	public OutputVariableCollector() {
		setDomain("Generic");
		setVersionInformation(200);
		setOperationIdentifier("ABCDEFGHIJKLMN");
		map = new HashMap<String, Double>();
		dayNumber = 1;
	}

	@Override
	public void execute() throws ProcessingException {
		// TODO Auto-generated method stub
		StreamingAccess access = null;
		String datasetIdentifier = "COLLECTOR_DATA";
		DatasetCollection datasetCollection = getDatasetCollection();
		try {
			Dataset dataset = datasetCollection.getDataset(datasetIdentifier);
			if (dataset == null)
				return;
			if (dataset.getDatasetType() != Dataset.STREAMS)
				return;
			access = (StreamingAccess) dataset;
			if (access != null) {
				while (dataset.isAvailable()) {
					NBEvent nbEvent = access.getStreamingData();
					if (nbEvent == null) {
						return;
					}
					if (nbEvent.getContentSynopsis().equals(
							"Async/OutputVariables")) {
						outputVariableRecord(nbEvent);
					} else if (nbEvent.getContentSynopsis().equals(
							"Sync/DownStream")) {
						// Flush to Disk
						if (!map.isEmpty()) {
							File file = new File("Output.csv");
							FileWriter writer = new FileWriter(file, true);
							writer.write("day");
							Object[] names = map.keySet().toArray();
							for (int i = 0; i < names.length; i++) {
								writer.write("," + names[i]);
							}
							writer.write("\n");
							writer.write("" + dayNumber);
							for (int i = 0; i < names.length; i++) {
								writer.write("," + map.get(names[i]));
							}
							writer.write("\n");
							writer.close();
							// Clear map
							map.clear();
						}
						dayNumber++;
					}

				}
			}
		} catch (DatasetException | IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Once got output variables, record them and prepare to write into disk.
	 * 
	 * @param nbEvent
	 */
	private void outputVariableRecord(NBEvent nbEvent) {
		try {
			Results resultsEx = ResultsFactory.getInstance().getResults(
					nbEvent.getContentPayload());
			byte[] reportBytes = resultsEx.getResultPayload();
			OutputVariableValue outputVariableValue = new OutputVariableValue();
			outputVariableValue = outputVariableValue
					.unmarshalling(reportBytes);
			String variableName = new String();
			if (outputVariableValue.getNameLength() > 0) {
				variableName = outputVariableValue.getName();
			} else {
				variableName = variableNames[outputVariableValue.getVarCode()];
			}
			System.out.println("Name:" + variableName);
			if (map.containsKey(variableName)) {
				Double newValue = map.get(variableName)
						+ outputVariableValue.getFValue()
						+ outputVariableValue.getIValue();
				map.put(variableName, newValue);
			} else {
				map.put(variableName,
						new Double(outputVariableValue.getIValue()
								+ outputVariableValue.getFValue()));
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (MarshallingException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

}
