/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.naadsm.dynamicsplit;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.InetAddress;
import java.util.LinkedList;
import cgl.narada.event.NBEvent;
import cgl.narada.event.NBEventException;
import cgl.narada.service.ServiceException;
import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.naadsm.dynamicsplit.event.Checkpoint;
import ds.granules.naadsm.dynamicsplit.event.EndOfDayEvent;
import ds.granules.naadsm.dynamicsplit.event.EndOfSimulationEvent;
import ds.granules.naadsm.dynamicsplit.event.EventType;
import ds.granules.naadsm.dynamicsplit.event.EventTypeFactory;
import ds.granules.naadsm.dynamicsplit.event.Restore;
import ds.granules.naadsm.dynamicsplit.event.SplitEvent;
import ds.granules.naadsm.dynamicsplit.event.StartNewDayEvent;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProcessingException;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * This is the controller for distributed NAADSM. It is responsible for
 * synchronization at the end of simulation days and finalizing data at the end
 * of simulation. Also it controls the split action. It will judge which worker
 * is going to split at the beginning of the next simulation day.
 * 
 * @author Zhiquan (Simon) Sui
 * 
 */
public class Controller extends MapReduceBase {

	private int numOfSplitWorkers;
	private int currentDay;
	private Integer numOfFinishedSplittingWorkers;
	private LinkedList<Boolean> finishedWorker;
	private LinkedList<Boolean> doneWorker;
	private boolean finished;
	private int numberOfWorkers;
	private int numberOfSpareWorkers;
	private long[] executionTime;
	private Results exchangeResults;
	private byte[] buffer;
	long startTime;
	long endTime;
	File record;
	BufferedWriter writer;
	long time1;
	long time2;
	String hostName;

	/**
	 * Get the number of next spare worker
	 * 
	 * @return number of the worker
	 */
	private synchronized int getNextSpare() {
		numberOfWorkers++;
		numberOfSpareWorkers--;
		return numberOfWorkers - 1;
	}

	/**
	 * Constructor. Initialize variables and byte array.
	 */
	public Controller() {
		startTime = System.currentTimeMillis();
		setDomain("Generic");
		setVersionInformation(200);
		setOperationIdentifier("ABCDEFGHIJKLMN");
		finished = false;
		exchangeResults = createResults(false, false);
		buffer = new byte[0];

	}

	@Override
	public void execute() throws ProcessingException {
		StreamingAccess access = null;
		String datasetIdentifier = "CONTROLLER_DATA";
		DatasetCollection datasetCollection = getDatasetCollection();
		try {
			Dataset dataset = datasetCollection.getDataset(datasetIdentifier);
			if (dataset == null) {
				System.out.println("NULL DATASET!");
				return;
			}
			if (dataset.getDatasetType() == Dataset.STREAMS) {
				access = (StreamingAccess) dataset;
			} else {
				System.err.println("Incorrect dataset:"
						+ dataset.getDatasetType());
				return;
			}
			if (access != null) {
				while (dataset.isDataAvailable()) {
					NBEvent nbEvent = access.getStreamingData();
					if (nbEvent == null) {
						System.out
								.println("Receive null message from Stream dataset");
						continue;
					}
					/* Judge the type of the received message. */
					if (nbEvent.getContentSynopsis().equals("Sync/Bootstrap")) {
						bootstrapProcess(nbEvent);
					} else if (nbEvent.getContentSynopsis().equals(
							"Sync/UpStream")) {
						upStreamProcess(nbEvent);
					} else if (nbEvent.getContentSynopsis().toString()
							.equals("Sync/Checkpoint")) {
						checkpointProcess(nbEvent);

					}
					if (finished)
						access.close();

				}
			}
		} catch (DatasetException e) {
			System.out.println("DatasetException");
			e.printStackTrace();
		} catch (ServiceException e) {
			System.out.println("ServiceException");
			e.printStackTrace();
		} catch (NBEventException e) {
			System.out.println("NBEventException");
			e.printStackTrace();
		} catch (IOException e) {
			System.out.println("IOException");
			e.printStackTrace();
		} catch (MapReduceException e) {
			System.out.println("MapReduceException");
			e.printStackTrace();
		} catch (MarshallingException e) {
			System.out.println("MarshallingException");
			e.printStackTrace();
		}

	}

	/**
	 * From a whole message, pick up the END_OF_DAY message
	 * 
	 * @param message
	 *            The whole message
	 * @return END_OF_DAY message
	 */
	private byte[] pickupEndOfDay(byte[] message) {
		int length = message.length;
		byte[] endOfDayArray = new byte[EventType.EventSize[EventType.END_OF_DAY]];
		for (int i = 0; i < EventType.EventSize[EventType.END_OF_DAY]; i++) {
			endOfDayArray[i] = message[length
					- EventType.EventSize[EventType.END_OF_DAY] + i];
		}
		return endOfDayArray;
	}

	/**
	 * Combine two messages into one byte array
	 * 
	 * @param message1
	 *            First message
	 * @param message2
	 *            Second message
	 * @return Combined byte array
	 */
	private byte[] combineByte(byte[] message1, byte[] message2) {
		byte[] result;
		if (message1.length == 0) {
			result = new byte[message2.length
					- EventType.EventSize[EventType.END_OF_DAY]];
			for (int i = 0; i < message2.length
					- EventType.EventSize[EventType.END_OF_DAY]; i++) {
				result[i] = message2[i];
			}
		} else {
			result = new byte[message1.length + message2.length
					- EventType.EventSize[EventType.END_OF_DAY]];
			for (int i = 0; i < message1.length; i++) {
				result[i] = message1[i];
			}
			for (int i = 0; i < message2.length
					- EventType.EventSize[EventType.END_OF_DAY]; i++) {
				result[i + message1.length] = message2[i];
			}
		}
		return result;
	}

	/**
	 * Process function when checkpoint message is arrived.
	 * 
	 * @param nbEvent
	 *            Received message
	 * @throws MapReduceException
	 * @throws IOException
	 * @throws MarshallingException
	 */
	private synchronized void checkpointProcess(NBEvent nbEvent)
			throws MapReduceException, IOException, MarshallingException {
		Results resultsEx = ResultsFactory.getInstance().getResults(
				nbEvent.getContentPayload());
		byte[] reportBytes = resultsEx.getResultPayload();
		Checkpoint checkpoint = new Checkpoint();
		checkpoint = checkpoint.getBoundaries(reportBytes);
		Restore restore = new Restore();
		restore.getBoundaries(checkpoint.addBoundaries());
		int newWorker = this.getNextSpare();
		System.out.println("Number of new worker: " + newWorker);
		restore.worker_ID = newWorker;
		exchangeResults = createResults(false, false);
		exchangeResults.setResultPayload(restore.addBoundaries());
		writeResults("Sync/DownStream/" + newWorker, exchangeResults);
		synchronized (numOfFinishedSplittingWorkers) {
			numOfFinishedSplittingWorkers++;
		}
		/* After all the split work is done, go into the next day */
		if (numOfFinishedSplittingWorkers == numOfSplitWorkers) {
			byte[] bytes;
			StartNewDayEvent startNewDay = new StartNewDayEvent(0,
					currentDay + 1);
			System.out.println();
			System.out.println("All done:" + " Day " + currentDay);
			bytes = startNewDay.marshalling();
			byte[] finalResult = new byte[buffer.length
					+ EventType.EventSize[EventType.START_NEW_DAY] + 1];
			for (int i = 0; i < buffer.length; i++) {
				finalResult[i] = buffer[i];
			}
			for (int i = 0; i < bytes.length; i++) {
				finalResult[buffer.length + i] = bytes[i];
			}
			finalResult[buffer.length
					+ EventType.EventSize[EventType.START_NEW_DAY]] = 0;
			buffer = new byte[0];
			System.out.println("FinalResult.length = " + finalResult.length);
			Results ckptMessage = createResults(false, false);

			ckptMessage.setResultPayload(finalResult);
			writeResults("Sync/DownStream", ckptMessage);
			time2 = System.currentTimeMillis();
			System.out.println("Day: " + currentDay + "  Length: "
					+ finalResult.length + "  Time: " + (int) (time2 - time1));
			time1 = time2;
		}
	}

	/**
	 * If synopsis of received message is UPSTREAM, the events will be proceeded
	 * in this function.
	 * 
	 * @param nbEvent
	 *            Received message containing all the information
	 * @throws MapReduceException
	 * @throws ServiceException
	 * @throws NBEventException
	 * @throws IOException
	 * @throws MarshallingException
	 */
	private synchronized void upStreamProcess(NBEvent nbEvent)
			throws MapReduceException, ServiceException, NBEventException,
			IOException, MarshallingException {
		int numberOfWorkers = this.numberOfWorkers;
		Results resultsEx = ResultsFactory.getInstance().getResults(
				nbEvent.getContentPayload());
		byte[] reportBytes = resultsEx.getResultPayload();

		buffer = combineByte(buffer, reportBytes);
		EndOfDayEvent endOfDayEvent = (EndOfDayEvent) EventTypeFactory
				.createEventType(pickupEndOfDay(reportBytes));
		currentDay = endOfDayEvent.getDayNumber();
		int num = endOfDayEvent.getWorkerID();
		long timeTemp = System.currentTimeMillis();
		long elapsedTime = timeTemp - time1;
		executionTime[num] = elapsedTime;
		if (endOfDayEvent.isFinished()) {
			for (int i = 0; i < numberOfWorkers; i++) {
				System.out.println(executionTime[i] + " ");
			}
			System.out.println();
			finishedWorker.set(endOfDayEvent.getWorkerID(), true);
			doneWorker.set(endOfDayEvent.getWorkerID(), true);
			boolean flag = true;
			for (int i = 0; i < numberOfWorkers; i++) {
				if (!finishedWorker.get(i)) {
					flag = false;
					break;
				}
			}
			/* Check END_OF_SIMULATION condition */
			if (flag) {
				System.out.println("All finished:" + " Day "
						+ endOfDayEvent.getDayNumber());
				endTime = System.currentTimeMillis();
				System.out.println("Elapsed Time =" + (endTime - startTime));

				EndOfSimulationEvent endOfSimulationEvent = new EndOfSimulationEvent();
				byte[] bytes = endOfSimulationEvent.marshalling();
				byte[] finalResult = new byte[buffer.length
						+ EventType.EventSize[EventType.END_OF_SIMULATION] + 1];

				for (int i = 0; i < buffer.length; i++) {
					finalResult[i] = buffer[i];
				}
				for (int i = 0; i < bytes.length; i++) {
					finalResult[buffer.length + i] = bytes[i];
				}
				finalResult[buffer.length
						+ EventType.EventSize[EventType.END_OF_SIMULATION]] = 1;
				exchangeResults = createResults(true, true);
				exchangeResults.setResultPayload(finalResult);
				writeResults("Sync/DownStream", exchangeResults);

				Results results = createResults(true, true);
				String res = "END";
				results.setResultPayload(res.getBytes());
				writeResults("Sync/Result", results);
				writer.write("" + endOfDayEvent.getDayNumber());
				writer.newLine();
				writer.write("" + finalResult.length);
				writer.newLine();
				time2 = System.currentTimeMillis();
				writer.write("" + (int) (time2 - time1));
				System.out.println("Day: " + endOfDayEvent.getDayNumber()
						+ "  Length: " + finalResult.length + "  Time: "
						+ (int) (time2 - time1));
				time1 = time2;
				writer.newLine();
				System.out.println("Writer Closed!");
				writer.close();
				System.out
						.println("Invoked Termination Condition Reached on Controller.");
				return;
			}
		} else {
			doneWorker.set(endOfDayEvent.getWorkerID(), true);
		}

		boolean flag = true;
		for (int i = 0; i < numberOfWorkers; i++) {
			if (!doneWorker.get(i)) {
				flag = false;
				break;
			}
		}
		System.out.print(endOfDayEvent.getWorkerID() + " ");
		if (flag) {
			File file = new File(
					"/s/chopin/e/proj/granules/simonsui/Granules/trunk/bin/execution_time.txt");
			FileWriter fw;
			if (!file.exists()) {
				file.createNewFile();
				fw = new FileWriter(file, false);
			} else
				fw = new FileWriter(file, true);
			System.out.println("");
			System.out.println("====================");
			for (int i = 0; i < executionTime.length; i++) {
				fw.write(executionTime[i] + " ");
				System.out.print(executionTime[i] / 1000 + " ");
			}
			fw.write("\n");
			fw.close();
			System.out.println("");
			System.out.println("====================");
			for (int i = 0; i < numberOfWorkers; i++) {
				doneWorker.set(i, false);
				finishedWorker.set(i, false);
			}
			// See which one should be splitted
			numOfSplitWorkers = 0;
			if (numberOfSpareWorkers > 0) {
				long minTime = Long.MAX_VALUE;
				long maxTime = Long.MIN_VALUE;
				long secondMaxTime = Long.MIN_VALUE;
				for (int i = 0; i < numberOfWorkers; i++) {
					if (minTime > executionTime[i]) {
						minTime = executionTime[i];
					}
					if (maxTime < executionTime[i]) {
						secondMaxTime = maxTime;
						maxTime = executionTime[i];
					} else if (secondMaxTime < executionTime[i]) {
						secondMaxTime = executionTime[i];
					}
				}
				int workerLeft = numberOfSpareWorkers;
				int workerActive = numberOfWorkers;
				double dWorkerLeft = 1.0 * workerLeft;
				double dWorkerActive = 1.0 * workerActive;
				double rate = 1.0
						* ((dWorkerActive - 1.0) / (dWorkerActive + dWorkerLeft))
						* ((dWorkerLeft + dWorkerActive) / (dWorkerActive
								+ dWorkerLeft - 2)) + 1.0; // Threshold function
				int lowThreshold = (int) (8000 * rate); // 8000 is constant of 8
														// seconds
				int highThreshold = (int) (15000 * rate); // 15000 is constant
															// of 15 seconds
				for (int i = 0; i < workerActive; i++) {
					if (workerLeft == 0) {
						break;
					}
					if (executionTime[i] < lowThreshold)
						continue;
					else if (workerActive == 1
							|| executionTime[i] > highThreshold
							|| executionTime[i] >= 1.5 * secondMaxTime) {
						numOfSplitWorkers++;
						SplitEvent splitEvent = new SplitEvent();
						splitEvent.worker_ID = i;
						splitEvent.howToSplit = 1;
						splitEvent.coordinate = 0;
						exchangeResults = createResults(false, false);
						exchangeResults.setResultPayload(splitEvent
								.marshalling());
						writeResults("Sync/DownStream/" + i, exchangeResults);
						System.out.println("SPLIT :" + i);
						System.out.println("rate:" + rate + " low:"
								+ lowThreshold + " high:" + highThreshold);
						workerLeft--;
						System.out.println("LEFT WORKER = " + workerLeft);
					}
				}
			}

			numOfFinishedSplittingWorkers = 0;
			/* If no worker needs to be split, just go into the next day */
			if (numOfSplitWorkers == 0) {
				numOfFinishedSplittingWorkers = 0;
				byte[] bytes;
				StartNewDayEvent startNewDay = new StartNewDayEvent(0,
						endOfDayEvent.getDayNumber() + 1);
				System.out.println();
				System.out.println("All done:" + " Day "
						+ endOfDayEvent.getDayNumber());
				bytes = startNewDay.marshalling();
				byte[] finalResult = new byte[buffer.length
						+ EventType.EventSize[EventType.START_NEW_DAY] + 1];
				for (int i = 0; i < buffer.length; i++) {
					finalResult[i] = buffer[i];
				}
				for (int i = 0; i < bytes.length; i++) {
					finalResult[buffer.length + i] = bytes[i];
				}
				finalResult[buffer.length
						+ EventType.EventSize[EventType.START_NEW_DAY]] = 0;
				buffer = new byte[0];
				System.out
						.println("FinalResult.length = " + finalResult.length);
				exchangeResults = createResults(false, false);

				exchangeResults.setResultPayload(finalResult);
				writeResults("Sync/DownStream", exchangeResults);
				time2 = System.currentTimeMillis();
				System.out.println("Day: " + endOfDayEvent.getDayNumber()
						+ "  Length: " + finalResult.length + "  Time: "
						+ (int) (time2 - time1));
				time1 = time2;
			}
		}
	}

	/**
	 * Once receive a BOOTSTRAP message, start the first day.
	 * 
	 * @param nbEvent
	 *            Received message containing all the information.
	 * @throws MapReduceException
	 * @throws ServiceException
	 * @throws NBEventException
	 */
	private void bootstrapProcess(NBEvent nbEvent) throws MapReduceException,
			ServiceException, NBEventException {
		hostName = "";
		try {
			InetAddress address = InetAddress.getLocalHost();
			hostName = address.getHostName();
		} catch (Exception e) {
		}
		numberOfWorkers = 1;
		numberOfSpareWorkers = Integer.parseInt(this.getProcessingDirectives()
				.get("Number of Workers").toString()) - 1;
		executionTime = new long[numberOfSpareWorkers + numberOfWorkers];
		finishedWorker = new LinkedList<Boolean>();
		doneWorker = new LinkedList<Boolean>();
		for (int i = 0; i < numberOfWorkers + numberOfSpareWorkers; i++) {
			finishedWorker.add(new Boolean(false));
			doneWorker.add(new Boolean(false));
		}
		StartNewDayEvent startEvent = new StartNewDayEvent(0, 1);
		byte[] startMessage = new byte[EventType.EventSize[EventType.START_NEW_DAY] + 1];
		byte[] start = startEvent.marshalling();
		for (int i = 0; i < EventType.EventSize[EventType.START_NEW_DAY]; i++) {
			startMessage[i] = start[i];
		}
		startMessage[EventType.EventSize[EventType.START_NEW_DAY]] = 0;
		exchangeResults.setResultPayload(startMessage);
		writeResults("Sync/DownStream", exchangeResults);
		time1 = System.currentTimeMillis();

	}

}
