/**
 * Software License, Version 1.0
 * 
 * Copyright 2008 The Trustees of Indiana University.  All rights reserved.
 * 
 *
 *Redistribution and use in source and binary forms, with or without 
 *modification, are permitted provided that the following conditions are met:
 *
 *1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 *2) All redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 *3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Community Grids Lab. For 
 * further information contact the Community Grids Lab at 
 * http://communitygrids.iu.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 *4) The name Indiana University or Community Grids Lab or Granules, 
 * shall not be used to endorse or promote products derived from this software 
 * without prior written permission from Indiana University.  For written 
 * permission, please contact the Advanced Research and Technology Institute 
 * ("ARTI") at 351 West 10th Street, Indianapolis, Indiana 46202.
 *5) Products derived from this software may not be called Granules, 
 * nor may Indiana University or Community Grids Lab or Granules appear
 * in their name, without prior written permission of ARTI.
 * 
 *
 * Indiana University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Indiana University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.naadsm.ft;

import java.io.IOException;
import java.util.Properties;

import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.DatasetFactory;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.CommunicationsException;
import ds.granules.exception.DeploymentException;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.InstanceDeployer;
import ds.granules.operation.ProgressTracker;

/**
 * Deploys fault tolerance components.
 */
public class Deployer extends InstanceDeployer {

    public static final String datasetIdentifier = "CONTROLLER_DATA";
    private ProgressTracker progressTracker;

    /**
     * Initializes the deployer.
     * 
     * @param streamingProperties
     *            The properties for streaming
     * @throws CommunicationsException
     *             If there are network problems during communications
     * @throws IOException
     *             If there are IO problems
     * @throws MarshallingException
     *             If there are marshalling errors
     * @throws DeploymentException
     *             If there are deployment related problems
     */
    public Deployer(Properties streamingProperties)
            throws CommunicationsException, IOException, MarshallingException,
            DeploymentException {
        initialize(streamingProperties);
    }

    /**
     * This method deploys <code>GalileoMap</code> instances over the set of
     * available resources.
     * 
     * @param numOfOperations
     *             Number of Speculators to start
     * @throws DatasetException
     *             If there are problems initializing the datasets
     * @throws CommunicationsException
     *             If there are network problems during communications
     * @throws DeploymentException
     *             If there are deployment related problems
     * @throws MarshallingException
     *             If there are marshalling problems.
     */
    public void prepareInstancesAndDeploy(int numOfOperations)
    throws DatasetException, CommunicationsException,
           DeploymentException, MarshallingException {

        DatasetFactory datasetFactory = DatasetFactory.getInstance();
        StreamingAccess streamingAccess =
            datasetFactory.createStreamingDataset(datasetIdentifier,
                datasetIdentifier);

        streamingAccess.addInputStream("Sync/UpStream",
                StreamingAccess.STREAMS);
        streamingAccess.addInputStream("Sync/DownStream",
                StreamingAccess.STREAMS);
        streamingAccess.addInputStream("Sync/WorkerInfo",
                StreamingAccess.STREAMS);
        streamingAccess.addInputStream("Sync/ControllerSuspendResponse",
                StreamingAccess.STREAMS);
        streamingAccess.addInputStream("Sync/WorkerSuspendResponse",
                StreamingAccess.STREAMS);
        streamingAccess.addInputStream("Async/Checkpoint",
                StreamingAccess.STREAMS);

        streamingAccess.addInputStream("FT/Heartbeat",
                StreamingAccess.STREAMS);

        Speculator[] speculators = new Speculator[numOfOperations];
        for (int i = 0; i < numOfOperations; ++i) {
            speculators[i] = new Speculator();
            speculators[i].setInitializationNeeded(true);
            speculators[i].setAsExecuteWhenDataAvailable();

            DatasetCollection datasets = speculators[i].getDatasetCollection();
            datasets.addDataset(streamingAccess);
        }
        progressTracker = deployOperations(speculators);
        System.out.println("Deployed " + numOfOperations +
            " Speculator instances.");
    }

    public void deployHeartbeats(int numOfOperations)
    throws DatasetException, CommunicationsException,
           DeploymentException, MarshallingException {

        DatasetFactory datasetFactory = DatasetFactory.getInstance();
        StreamingAccess streamingAccess =
            datasetFactory.createStreamingDataset(datasetIdentifier,
                datasetIdentifier);

        streamingAccess.addOutputStream("FT/Heartbeat",
                StreamingAccess.STREAMS);

        Heartbeat[] heartbeats = new Heartbeat[numOfOperations];
        for (int i = 0; i < numOfOperations; ++i) {
            heartbeats[i] = new Heartbeat();
            heartbeats[i].setInitializationNeeded(true);
            heartbeats[i].setAsExecutePeriodically(
                    Speculator.HEARTBEAT_INTERVAL_MS);

            DatasetCollection datasets = heartbeats[i].getDatasetCollection();
            datasets.addDataset(streamingAccess);
        }
        progressTracker = deployOperations(heartbeats);
        System.out.println("Deployed " + numOfOperations +
            " Heartbeat instances.");
    }

    /**
     * This method attempts to terminate the job that was previously submitted.
     * 
     * @throws DeploymentException
     *             If there are deployment related problems
     * @throws CommunicationsException
     *             If there are network problems during communications
     * @throws MarshallingException
     *             If there are marshalling problems.
     */
    public void terminateJob() throws DeploymentException,
            CommunicationsException, MarshallingException {
        String jobIdentifier = progressTracker.getIdentifier();
        killJob(jobIdentifier);
    }

    /**
     * @param args
     */
    public static void main(String[] args) {
        if (args.length != 3) {
            System.out.println("Usage: ds.granules.naadsm.ft.Deployer" +
                " <broker hostname> <port> <num-machines>");

            System.exit(1);
        }

        String hostname = args[0];
        String portnum = args[1];
        int heartbeats = Integer.parseInt(args[2]);

        Properties streamingProperties = new Properties();
        streamingProperties.put("hostname", hostname);
        streamingProperties.put("portnum", portnum);

        Deployer deployer;
        try {
            deployer = new Deployer(streamingProperties);
            deployer.prepareInstancesAndDeploy(1);
            deployer.deployHeartbeats(heartbeats);
        } catch (Exception e) {
            System.out.println("(!) Deployment Error");
            e.printStackTrace();
        }
    }
}
