/**
 * Software License, Version 1.0
 * 
 * Copyright 2008 The Trustees of Indiana University.  All rights reserved.
 * 
 *
 *Redistribution and use in source and binary forms, with or without 
 *modification, are permitted provided that the following conditions are met:
 *
 *1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 *2) All redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 *3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Community Grids Lab. For 
 * further information contact the Community Grids Lab at 
 * http://communitygrids.iu.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 *4) The name Indiana University or Community Grids Lab or Granules, 
 * shall not be used to endorse or promote products derived from this software 
 * without prior written permission from Indiana University.  For written 
 * permission, please contact the Advanced Research and Technology Institute 
 * ("ARTI") at 351 West 10th Street, Indianapolis, Indiana 46202.
 *5) Products derived from this software may not be called Granules, 
 * nor may Indiana University or Community Grids Lab or Granules appear
 * in their name, without prior written permission of ARTI.
 * 
 *
 * Indiana University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Indiana University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.naadsm.ft;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Timer;
import java.util.TimerTask;

import cgl.narada.event.NBEvent;

import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;

import ds.granules.naadsm.dynamicmerge.WorkerList;
import ds.granules.naadsm.dynamicmerge.event.Checkpoint;
import ds.granules.naadsm.dynamicmerge.event.EndOfDayEvent;
import ds.granules.naadsm.dynamicmerge.event.EndOfSimulationEvent;
import ds.granules.naadsm.dynamicmerge.event.EventType;
import ds.granules.naadsm.dynamicmerge.event.EventTypeFactory;
import ds.granules.naadsm.dynamicmerge.event.Restore;
import ds.granules.naadsm.dynamicmerge.event.StartNewDayEvent;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProcessingException;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

public class Speculator extends MapReduceBase {

    public static final String HEARTBEAT_SYNOPSIS = "FT/Heartbeat";
    public static final int HEARTBEAT_INTERVAL_MS = 5000;
    public static final int HOST_TIMEOUT_MS = HEARTBEAT_INTERVAL_MS * 2 + 1000;

    private int currentDay = 0;
    private boolean recovering = false;

    private Map<String, HostInfo> hostMap =
        new HashMap<String, HostInfo>();
    private Map<Integer, WorkerInfo> workerMap =
        new HashMap<Integer, WorkerInfo>();

    private List<Snapshot> snapshots = new ArrayList<Snapshot>();
    private List<RecoveryProcess> recoveries = new ArrayList<RecoveryProcess>();

    private Timer hostTimeoutTimer;

    private class HostTimeout extends TimerTask {
        public void run() {
            timeoutCheck();
        }
    }

    public Speculator() {
        setDomain("Generic");
        setVersionInformation(200);
        setOperationIdentifier("ABCDEFGHIJKLMN");
    }

    @Override
    public void initialize(Properties processingDirectives) {
        hostTimeoutTimer = new Timer();
        hostTimeoutTimer.schedule(new HostTimeout(), HOST_TIMEOUT_MS,
                HOST_TIMEOUT_MS);

        setInitializationComplete();
    }

    @Override
    public void execute() throws ProcessingException {
        DatasetCollection datasetCollection = getDatasetCollection();
        StreamingAccess access = null;

        try {
            Dataset dataset =
                datasetCollection.getDataset(Deployer.datasetIdentifier);

            if (dataset.getDatasetType() == Dataset.STREAMS) {
                access = (StreamingAccess) dataset;
            } else {
                System.out.println("Incorrect datatset: "
                    + dataset.getDatasetType() + " initialized. Returning ...");
                return;
            }

            while (dataset.isDataAvailable()) {
                NBEvent nbEvent = access.getStreamingData();
                processStreamPacket(nbEvent);
            }

        } catch (Exception e) {
            System.out.println("Caught exception: " + e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * Handles incoming events.  Depending on the current system state, events
     * are handled differently or different events are expected.  The two
     * primary states the Speculator can be in are: recovery and general.
     */
    public void processStreamPacket(NBEvent nbEvent)
    throws IOException, MarshallingException {
        String synopsis = (String) nbEvent.getContentSynopsis();

        ResultsFactory resultsFactory = ResultsFactory.getInstance();
        Results results =
            resultsFactory.getResults(nbEvent.getContentPayload());
        byte[] resultBytes = results.getResultPayload();

        /* Heartbeat */
        if (synopsis.equals(HEARTBEAT_SYNOPSIS)) {
            String deploymentStream = new String(resultBytes);
            handleHeartbeat(deploymentStream);

            return;
        }

        if (recovering) {
            processRecoveryEvent(synopsis, resultBytes);
        } else {
            processGeneralEvent(synopsis, resultBytes);
        }
    }

    /**
     * Handles events that occur during normal system execution.
     */
    private void processGeneralEvent(String synopsis, byte[] payload)
    throws IOException {
        /* Message from Worker */
        if (synopsis.equals("Sync/UpStream")) {
            int endSize = EventType.EventSize[EventType.END_OF_DAY];
            if (endSize > payload.length) {
                /* size mismatch */
                return;
            }

            byte[] endBytes = new byte[endSize];
            System.arraycopy(payload, payload.length - endSize,
                    endBytes, 0, endSize);

            EndOfDayEvent endEvent =
                (EndOfDayEvent) EventTypeFactory.createEventType(endBytes);

            handleEndOfDay(endEvent);
            return;
        }

        /* Message from Controller to Workers */
        if (synopsis.equals("Sync/DownStream")) {
            Snapshot currentSnapshot = getCurrentSnapshot();
            if (currentSnapshot == null || currentSnapshot.isComplete()) {
                return;
            }

            int newDaySize = EventType.EventSize[EventType.START_NEW_DAY];

            if (newDaySize > payload.length) {
                return;
            }

            byte[] newDayBytes = new byte[newDaySize];
            System.arraycopy(payload, payload.length - (newDaySize + 1),
                    newDayBytes, 0, newDaySize);

            StartNewDayEvent newDayEvent =
                (StartNewDayEvent) EventTypeFactory.createEventType(newDayBytes);

            if (newDayEvent.getDayNumber() == currentSnapshot.getDay() + 1) {
                currentSnapshot.setGlobalState(payload);
            }
            return;
        }

        /* Worker list from Controller */
        if (synopsis.equals("Sync/WorkerInfo")) {
            System.out.println(" -> Received updated Worker List");
            ByteArrayInputStream byteIn = new ByteArrayInputStream(payload);
            ObjectInputStream objIn = new ObjectInputStream(byteIn);
            WorkerList workers = null;
            try {
                workers = (WorkerList) objIn.readObject();
            } catch (Exception e) {
                System.out.println("(!) Could not recover Worker List.");
                e.printStackTrace();
            }
            objIn.close();
            getCurrentSnapshot().setWorkerList(workers);
        }

        /* Checkpoint data */
        if (synopsis.equals("Async/Checkpoint")) {
            Checkpoint checkpoint =
                (Checkpoint) EventTypeFactory.createEventType(payload);

            handleCheckpoint(checkpoint);
            return;
        }
    }

    /**
     * Handles events that occur during a recovery operation.
     */
    private void processRecoveryEvent(String synopsis, byte[] payload) {
        /* Controller has suspended operation */
        if (synopsis.equals("Sync/ControllerSuspendResponse")) {
            System.out.println(" -> Controller reports suspended state.");
            suspendWorkers();
            return;
        }

        /* A worker has suspended its operation */
        if (synopsis.equals("Sync/WorkerSuspendResponse")) {
            System.out.println(" -> Worker reports suspended state.");

            RecoveryProcess recovery = getCurrentRecoveryProcess();
            recovery.incrementSuspendedWorkers();

            if (recovery.allWorkersSuspended()) {
                System.out.println("(i) All workers have suspended.");
                distributeStateInformation();
                resumeController();
                resumeWorkers();
            }

            return;
        }
    }

    /**
     * Requests checkpoints from the active Workers.
     * This is accomplished by publishing data to the Controller, which then
     * appends the request to the next batch of binary information being sent to
     * the Workers.
     */
    private void requestCheckpoints() {
        Snapshot snap = new Snapshot(currentDay);
        snapshots.add(snap);

        System.out.println(" <- Requesting checkpoints...");
        publishEvent("Async/CheckpointRequest");
    }

    /**
     * Restores the system to the last complete checkpoint Snapshot received.
     */
    private void recoverSystem() {
        Snapshot lastComplete = getLastCompleteSnapshot();
        if (lastComplete == null) {
            System.out.println("(!) No complete checkpoints available.  " +
                    "Failure recovery aborted.");
            return;
        }

        recoverSystem(lastComplete);
    }

    /**
     * Restores the system to the requested Snapshot of state information.
     *
     * @param shot Snapshot to roll back to.
     */
    private void recoverSystem(Snapshot shot) {
        System.out.println("(i) System entering recovery state.");
        recovering = true;
        RecoveryProcess recovery = new RecoveryProcess(shot);
        recoveries.add(recovery);
        suspendController();
    }

    /**
     * Instructs the Controller to suspend operation.
     */
    private void suspendController() {
        System.out.println(" <- Suspending Controller.");
        WorkerList workerList =
            getCurrentRecoveryProcess().getSnapshot().getWorkerList();

        byte[] workerListPayload = null;
        try {
            ByteArrayOutputStream byteOut =
                new ByteArrayOutputStream();
            ObjectOutput objOut =
                new ObjectOutputStream(byteOut);
            objOut.writeObject(workerList);
            objOut.close();

            workerListPayload = byteOut.toByteArray();
        } catch (IOException e) {
            System.out.println("(!) Could not prepare recovery Worker list.");
            e.printStackTrace();
        }

        publishEvent("Async/SuspendRequest", workerListPayload);
    }

    /**
     * Resumes a suspended Controller.
     */
    private void resumeController() {
        System.out.println(" <- Resuming Controller operation.");
        publishEvent("Async/ResumeRequest");
    }

    /**
     * Instructs all active Worker instances to suspend their operation.
     */
    private void suspendWorkers() {
        System.out.println(" <- Suspending Workers.");

        EndOfSimulationEvent endEvent = new EndOfSimulationEvent(0, false);
        byte[] endPayload = endEvent.marshalling();
        byte[] buffer = new byte[endPayload.length + 1];
        System.arraycopy(endPayload, 0, buffer, 0, endPayload.length);
        buffer[buffer.length - 1] = 1;

        publishEvent("Sync/DownStream", buffer);
    }

    /**
     * Distributes Worker state bundles during a recovery operation.
     */
    private void distributeStateInformation() {
        RecoveryProcess recovery = getCurrentRecoveryProcess();
        Snapshot snapshot = recovery.getSnapshot();
        System.out.println("(i) Restoring system state to snapshot: " +
                snapshot.getCreationTime());

        for (Checkpoint checkpoint : snapshot.getCheckpoints()) {
            int id = checkpoint.getWorkerID();
            System.out.println(" <- Sending checkpoint to Worker " + id);

            Restore restore = new Restore();
            restore.setOverwrite(true);
            restore.setWorkerID(id);
            restore.setContent(checkpoint.getContent());

            publishEvent("Sync/DownStream/" + id, restore.marshalling());
        }
    }

    /**
     * Resumes suspended Workers.  This is achieved by bootstrapping the workers
     * with global state information that kickstarts a new simulation day.
     */
    private void resumeWorkers() {
        System.out.println(" <- Resuming Worker operation.");
        publishEvent("Sync/DownStream", getCurrentSnapshot().getGlobalState());
    }

    /**
     * Handle the END_OF_DAY message that is send out from the Workers to the
     * controller.  The Speculator listens in on this stream to acquire
     * information about the current simulation state.
     */
    private void handleEndOfDay(EndOfDayEvent endEvent) {
        int day = endEvent.getDayNumber();
        if (currentDay < day) {
            System.out.println("(i) Started new day: " + day);
            currentDay = day;

            if (day > 0 && day % 5 == 0)
                requestCheckpoints();

//            if (day == 20) {
//                recoverSystem();
//            }
                //suspendWorkers();
        }

//        HostInfo host = hostMap.get(ftMsg.deploymentStream);
//        if (host == null) {
//            System.out.println("(!) Unknown resource host: "
//                    + ftMsg.deploymentStream);
//            return;
//        }
//
//        WorkerInfo worker = host.getWorker(ftMsg.workerNumber);
//        if (worker == null) {
//            worker = new WorkerInfo(ftMsg.workerNumber);
//            host.addWorker(worker);
//            workerMap.put(ftMsg.workerNumber, worker);
//        }
//
//        if (currentDay < ftMsg.currentDay) {
//            currentDay = ftMsg.currentDay;
//            System.out.println("New day started: " + currentDay);
//            requestCheckpoints();
//        }
//        worker.setCurrentDay(ftMsg.currentDay);
//
//        if (ftMsg.bytes != null) {
//            worker.setCheckpoint(ftMsg.bytes);
//            System.out.println("Installed new checkpoint for worker "
//                    + ftMsg.workerNumber);
//        }
//        worker.setUpdateTime();
//        System.out.println("(i) Worker " + ftMsg.workerNumber +
//                "is on day " + ftMsg.currentDay);
    }

    /**
     * Process an incoming Worker checkpoint.
     */
    private void handleCheckpoint(Checkpoint checkpoint) {
        System.out.println(" -> Got a checkpoint from worker "
                + checkpoint.getWorkerID());
        Snapshot currentSnapshot = getCurrentSnapshot();
        currentSnapshot.addCheckpoint(checkpoint);
        if (currentSnapshot.isComplete()) {
            System.out.println("Complete!");
        }
    }

    /**
     * Process an incoming Heartbeat message.
     */
    private void handleHeartbeat(String deploymentStream) {
        System.out.println(" -> Heartbeat: " + deploymentStream);

        synchronized (hostMap) {
            HostInfo host = hostMap.get(deploymentStream);
            if (host == null) {
                HostInfo newHost = new HostInfo(deploymentStream);
                hostMap.put(deploymentStream, newHost);

                System.out.println("(i) New resource host discovered: " +
                        deploymentStream);
                host = newHost;
            }

            host.setHeartbeatTime();
        }
    }

    /**
     * Determine whether any hosts have failed to report a Heartbeat.
     */
    private void timeoutCheck() {
        synchronized (hostMap) {
            System.out.println("Checking for timed out hosts...");

            long now = System.nanoTime();
            for (HostInfo host : hostMap.values()) {
                long elapsed = now - host.getLastHeartbeatTime();
                if (elapsed >= HOST_TIMEOUT_MS * 1E6) {
                    System.out.println("(!) Host at stream: "
                            + host.getDeploymentStream() + " has died!");
                    hostMap.remove(host.getDeploymentStream());
                }
            }
        }
    }

    private Snapshot getCurrentSnapshot() {
        if (snapshots.size() == 0) {
            return null;
        }

        return snapshots.get(snapshots.size() - 1);
    }

    private RecoveryProcess getCurrentRecoveryProcess() {
        return recoveries.get(recoveries.size() - 1);
    }

    private Snapshot getLastCompleteSnapshot() {
        for (int i = snapshots.size() - 1; i >= 0; --i) {
            Snapshot shot = snapshots.get(i);
            if (shot.isComplete()) {
                return shot;
            }
        }
        return null;
    }

    private void publishEvent(String synopsis) {
        publishEvent(synopsis, new byte[1]);
    }

    /**
     * Simple event publishing wrapper.
     */
    private void publishEvent(String synopsis, byte[] payload) {
        Results results;
        results = createResults(true, true);
        results.setResultPayload(payload);

        try {
            writeResults(synopsis, results);
        } catch (MapReduceException e) {
            System.out.println("(!) Could not publish event.");
            e.printStackTrace();
        }
    }
}
