/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.naadsm.portionaldynamicmerge;

import java.io.IOException;
import java.util.Properties;

import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.DatasetFactory;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.CommunicationsException;
import ds.granules.exception.DeploymentException;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.InstanceDeployer;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProgressTracker;

/**
 * Deployer is to deploy the controller and the workers. It initializes the
 * parameters.
 * 
 * @author Zhiquan (Simon) Sui
 */

public class Deployer extends InstanceDeployer {
	private Properties props;
	private int finishedJob;

	/**
	 * Constructor. Initialization.
	 * 
	 * @param properties
	 *            Contains parameters
	 * @throws CommunicationsException
	 * @throws IOException
	 * @throws MarshallingException
	 * @throws DeploymentException
	 */
	public Deployer(Properties properties) throws CommunicationsException,
			IOException, MarshallingException, DeploymentException {
		finishedJob = 0;
		props = properties;
		initialize(properties);
	}

	/**
	 * Initialize one worker
	 * 
	 * @param workerNum
	 *            Number of the current worker.
	 * @param number
	 *            Number of total workers.
	 * @return The created worker
	 * @throws DatasetException
	 * @throws MapReduceException
	 */
	public Worker initializeWorker(int workerNum, int number)
			throws DatasetException, MapReduceException {
		Worker worker = new Worker();
		/* Put all the integer parameters into properties in string. */
		Properties props = (Properties) this.props.clone();
		String key = "Worker Number";
		String value = new String();
		value = "" + workerNum;
		props.put(key, value);

		worker.setProcessingDirectives(props);

		DatasetFactory datasetFactory = DatasetFactory.getInstance();
		StreamingAccess streamAccess = datasetFactory.createStreamingDataset(
				"WORKER_DATA" + value, "WORKER DATA");
		/* Register synopsis of the worker */
		String downstreamSynopsis2 = "Sync/DownStream";
		streamAccess.addInputStream(downstreamSynopsis2,
				StreamingAccess.STRING_SYNOPSIS);
		String downstreamSynopsis3 = "Sync/DownStream/" + workerNum;
		streamAccess.addInputStream(downstreamSynopsis3,
				StreamingAccess.STRING_SYNOPSIS);
		String chkptRequestSynopsis = "Sync/CheckpointRequest";
		streamAccess.addInputStream(chkptRequestSynopsis,
				StreamingAccess.STRING_SYNOPSIS);
		DatasetCollection datasetCollection = worker.getDatasetCollection();
		datasetCollection.addDataset(streamAccess);
		worker.setDatasetCollection(datasetCollection);
		worker.setAsExecuteWhenDataAvailable();
		return worker;
	}

	/**
	 * Initialize the controller
	 * 
	 * @param numOfWorkers
	 *            Total number of workers
	 * @return The controller
	 * @throws DatasetException
	 * @throws MapReduceException
	 */
	public Controller initializeController(int numOfWorkers)
			throws DatasetException, MapReduceException {
		Controller controller = new Controller();
		Properties props = (Properties) this.props.clone();
		String key = "Number of Workers";
		String value = "" + numOfWorkers;
		props.put(key, value);
		controller.setProcessingDirectives(props);

		DatasetFactory datasetFactory = DatasetFactory.getInstance();
		StreamingAccess streamAccess = datasetFactory.createStreamingDataset(
				"CONTROLLER_DATA", "CONTROLLER DATA");
		/* Register the synopsis of controller */
		String bootstrapSynopsis = "Sync/Bootstrap";
		String upstreamSynopsis = "Sync/UpStream";
		streamAccess.addInputStream(bootstrapSynopsis,
				StreamingAccess.STRING_SYNOPSIS);
		streamAccess.addInputStream(upstreamSynopsis,
				StreamingAccess.STRING_SYNOPSIS);
		String chkptSynopsis = "Sync/Checkpoint";
		streamAccess.addInputStream(chkptSynopsis,
				StreamingAccess.STRING_SYNOPSIS);
		String asynChkptSynopsis = "Async/CheckpointRequest";
		streamAccess.addInputStream(asynChkptSynopsis,
				StreamingAccess.STRING_SYNOPSIS);
		streamAccess.addInputStream("Async/SuspendRequest",
				StreamingAccess.STRING_SYNOPSIS);
		streamAccess.addInputStream("Async/ResumeRequest",
				StreamingAccess.STRING_SYNOPSIS);
		DatasetCollection datasetCollection = controller.getDatasetCollection();
		datasetCollection.addDataset(streamAccess);
		controller.setDatasetCollection(datasetCollection);
		controller.setAsExecuteWhenDataAvailable();
		return controller;
	}

	/**
	 * Initialize the controller and all the workers
	 * 
	 * @param number
	 *            Number of total workers.
	 */
	public void prepareInstancesAndDeploy(int number) {
		int numOfWorkers = number;
		MapReduceBase[] maps = new MapReduceBase[1 + numOfWorkers];

		try {
			for (int i = 0; i < numOfWorkers; i++) {
				maps[i] = initializeWorker(i, number);
			}
			maps[numOfWorkers] = initializeController(numOfWorkers);
			deployOperations(maps);
		} catch (CommunicationsException e) {
			e.printStackTrace();
		} catch (DeploymentException e) {
			e.printStackTrace();
		} catch (MarshallingException e) {
			e.printStackTrace();
		} catch (DatasetException e) {
			e.printStackTrace();
		} catch (MapReduceException e) {
			e.printStackTrace();
		}

	}

	@Override
	public void completedSuccessfully(String jobIdentifier,
			ProgressTracker progressTracker) {
		finishedJob++;
		if (finishedJob == 1) {
			System.exit(0);
		}
	}

	/**
	 * @param args
	 *            args[0] is host name of narada brokering. args[1] is port
	 *            number of narada brokering. args[2] is number of workers.
	 *            args[3] is the command line of the workers.
	 * 
	 */
	public static void main(String[] args) {
		String hostname = args[0];
		String portnum = args[1];
		String num = args[2];
		String cmd = args[3];
		Properties streamProperties = new Properties();
		streamProperties.put("hostname", hostname);
		streamProperties.put("portnum", portnum);
		streamProperties.put("cmdLine", cmd);
		if (System.getProperty("mapfile") != null) {
			streamProperties.put("mapfile", System.getProperty("mapfile"));
			System.out.println(System.getProperty("mapfile"));
		}
		int number = Integer.parseInt(num);
		try {
			Deployer deployer = new Deployer(streamProperties);
			deployer.prepareInstancesAndDeploy(number);
		} catch (CommunicationsException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (MarshallingException e) {
			e.printStackTrace();
		} catch (DeploymentException e) {
			e.printStackTrace();
		}
	}

}
