/**
 * Software License, Version 1.0
 * 
 * Copyright 2008 The Trustees of Indiana University.  All rights reserved.
 * 
 *
 *Redistribution and use in source and binary forms, with or without 
 *modification, are permitted provided that the following conditions are met:
 *
 *1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 *2) All redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 *3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Community Grids Lab. For 
 * further information contact the Community Grids Lab at 
 * http://communitygrids.iu.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 *4) The name Indiana University or Community Grids Lab or Granules, 
 * shall not be used to endorse or promote products derived from this software 
 * without prior written permission from Indiana University.  For written 
 * permission, please contact the Advanced Research and Technology Institute 
 * ("ARTI") at 351 West 10th Street, Indianapolis, Indiana 46202.
 *5) Products derived from this software may not be called Granules, 
 * nor may Indiana University or Community Grids Lab or Granules appear
 * in their name, without prior written permission of ARTI.
 * 
 *
 * Indiana University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Indiana University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.naadsm.portionaldynamicmerge;

import java.io.IOException;
import java.io.Serializable;

import org.encog.ml.data.MLData;
import org.encog.ml.data.basic.BasicMLData;

import ds.granules.naadsm.ml.Predictor;

/**
 * This class maintains information about a worker, its attributes and
 * corresponding functions.
 * 
 * @author simonsui
 * 
 */
public class WorkerNode implements Serializable {
	int workerNumber;
	boolean justMerged;
	boolean active;
	boolean finished;
	boolean done;
	double executionTime;
	double judgementTime;
	double predictedTime;
	long startTime;

	public WorkerNode(int num) {
		workerNumber = num;
		active = false;
		finished = false;
		done = false;
		justMerged = false;
	}

	/**
	 * Get the worker number.
	 * 
	 * @return
	 */
	public int getWorkerNumber() {
		return workerNumber;
	}

	/**
	 * Check whether the worker has finished the whole simulation.
	 * 
	 * @return
	 */
	public boolean isFinished() {
		return finished;
	}

	/**
	 * Check whether the worker has finished the simulation of current day.
	 * 
	 * @return
	 */
	public boolean isDone() {
		return done;
	}

	/**
	 * Set this worker as finished.
	 */
	public void finish() {
		finished = true;
	}

	/**
	 * Set this worker as finished in current day.
	 */
	public void done() {
		done = true;
	}

	/**
	 * Set this worker as active.
	 * 
	 * @return
	 */
	public boolean activate() {
		if (active)
			return false;
		else {
			active = true;
			return true;
		}
	}

	/**
	 * Set this worker as inactive.
	 * 
	 * @return
	 */
	public boolean deactivate() {
		if (!active)
			return false;
		else {
			active = false;
			executionTime = 0;
			return true;
		}
	}

	/**
	 * Start a new day for the current worker.
	 */
	public void start() {
		if (active) {
			startTime = System.currentTimeMillis();
			finished = false;
			done = false;
		}

	}

	/**
	 * Finish the current day for the current worker.
	 */
	public void stop() {
		if (active) {
			executionTime = System.currentTimeMillis() - startTime;
			judgementTime = executionTime;
			done = true;
		}
	}

	/**
	 * Finish the current day for the current worker and set the predicted time
	 * as criteria
	 * 
	 * @param predictedTime
	 */
	public void stop(double predictedTime) {
		if (active) {
			executionTime = System.currentTimeMillis() - startTime;
			this.predictedTime = predictedTime;
			judgementTime = predictedTime;
			done = true;
		}
	}

	/**
	 * Set the execution time for the current worker.
	 * 
	 * @param time
	 */
	public void setExecutionTime(long time) {
		executionTime = time;
	}

	/**
	 * Check whether this worker is active.
	 * 
	 * @return
	 */
	public boolean isActive() {
		return active;
	}

	/**
	 * Get the time that is as the criteria of load balancing operations
	 * 
	 * @return
	 */
	public double getJudgementTime() {
		return judgementTime;
	}

	/**
	 * Get the execution time of the current worker.
	 * 
	 * @return
	 */
	public double getExecutionTime() {
		return executionTime;
	}

	/**
	 * Get the predicted time of the current worker.
	 * 
	 * @return
	 */
	public double getPredictedTime() {
		return predictedTime;
	}

	/**
	 * Finish the current day for the current worker and use the predictor to
	 * predict the execution time of the next simulation day.
	 * 
	 * @param predictor
	 * @param array
	 */
	public void stop(Predictor predictor, double array[]) {
		if (active) {
			executionTime = System.currentTimeMillis() - startTime;
			double[] newArray = new double[array.length + 1];
			for (int i = 0; i < array.length; i++) {
				newArray[i] = array[i];
			}
			newArray[array.length] = executionTime;
			MLData data = new BasicMLData(newArray);
			try {
				predictedTime = predictor.predict(data);
			} catch (IOException e) {
				e.printStackTrace();
			}
			judgementTime = predictedTime;
			done = true;
			System.out.println("Execution Time:" + executionTime
					+ " PredictedTime:" + predictedTime);
		}
	}
}
