/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.naadsm.resplit;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.LinkedList;

import cgl.narada.event.NBEvent;
import cgl.narada.event.NBEventException;
import cgl.narada.service.ServiceException;
import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.naadsm.resplit.event.Checkpoint;
import ds.granules.naadsm.resplit.event.EndOfDayEvent;
import ds.granules.naadsm.resplit.event.EndOfSimulationEvent;
import ds.granules.naadsm.resplit.event.EventType;
import ds.granules.naadsm.resplit.event.EventTypeFactory;
import ds.granules.naadsm.resplit.event.MultiRestore;
import ds.granules.naadsm.resplit.event.MultiSplitByDensity;
import ds.granules.naadsm.resplit.event.OutputVariableValue;
import ds.granules.naadsm.resplit.event.Restore;
import ds.granules.naadsm.resplit.event.StartNewDayEvent;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProcessingException;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * This is the controller for distributed NAADSM. It is responsible for
 * synchronization at the end of simulation days and finalizing data at the end
 * of simulation. Also it controls the split action. It will judge which worker
 * is going to split at the beginning of the next simulation day.
 * 
 * @author Zhiquan (Simon) Sui
 * 
 */
public class Controller extends MapReduceBase {

	private Results exchangeResults;
	private int currentDay;
	private boolean initialSplit;
	private boolean finished;
	private int finishCheckpoint;
	private int numberOfWorkers;
	private int numberOfActiveWorkers;
	private Checkpoint[] checkpoints;
	private byte[] buffer;
	private boolean[] workerDoneToday; // whether worker is done for the current
										// day
	private boolean[] workerFinished;
	private double[] executionTime;
	private double[] ratio;
	private int[] numMessagesReceived;
	private long startTime;
	private long endTime;
	private long currentTime;

	/**
	 * Print out the byte array of an event.
	 * @param event
	 */
	public static void printEventMessage(EventType event) {
		byte[] bytes = event.marshalling();
		for (int i = 0; i < bytes.length; i++) {
			System.out.print(bytes[i] + " ");
		}
		System.out.println();
	}
	/**
	 * Get the end of day event from a list of events.
	 * @param eventList
	 * @return
	 * @throws MapReduceException
	 */

	public static byte[] pickupEndOfDay(LinkedList<EventType> eventList) {
		EventType event;
		for (int i = 0; i < eventList.size(); i++) {
			event = eventList.get(i);
			if (event.getType() == EventType.END_OF_DAY)
				return event.marshalling();
		}
		return null;
	}

	/**
	 * From a whole message, pick up the END_OF_DAY message
	 * 
	 * @param message
	 *            The whole message
	 * @return END_OF_DAY message
	 */
	public static byte[] pickupEndOfDay(byte[] message) {
		int length = message.length;
		byte[] endOfDayArray = new byte[EventType.EventSize[EventType.END_OF_DAY]];
		for (int i = 0; i < EventType.EventSize[EventType.END_OF_DAY]; i++) {
			endOfDayArray[i] = message[length
					- EventType.EventSize[EventType.END_OF_DAY] + i];
		}
		return endOfDayArray;
	}

	/**
	 * Get the output variable values from a list of events.
	 * @param eventList
	 * @return
	 * @throws MapReduceException
	 */
	public double[] pickupOutputVariableValue(
			LinkedList<EventType> eventList) throws MapReduceException {
		double[] output = new double[OutputVariableValue.VAR_NUM + 1];
		for (int i = 0; i < eventList.size(); i++) {
			EventType event = eventList.get(i);
			if (event.getType() == EventType.OUTPUT_VARIABLE_VALUE) {
				OutputVariableValue outputVariableValue = (OutputVariableValue) event;
				exchangeResults = createResults(false, false);
				exchangeResults
						.setResultPayload(outputVariableValue.marshalling());
				writeResults("Async/OutputVariables", exchangeResults);
				if (outputVariableValue.getVarCode() > 0)
					output[outputVariableValue.getVarCode()] = outputVariableValue
							.getIValue() + outputVariableValue.getFValue();
			}
		}
		return output;
	}

	/**
	 * Get output variable values from a byte array.
	 * @param message
	 * @return
	 */
	public double[] pickupOutputVariableValue(byte[] message) {
		byte[] outputVariableValueArray = new byte[EventType.EventSize[EventType.OUTPUT_VARIABLE_VALUE]];
		double[] output = new double[OutputVariableValue.VAR_NUM + 1];
		int offset = message.length - EventType.EventSize[EventType.END_OF_DAY]
				- EventType.EventSize[EventType.OUTPUT_VARIABLE_VALUE];
		while (true) {
			try {
				for (int i = 0; i < EventType.EventSize[EventType.OUTPUT_VARIABLE_VALUE]; i++) {
					outputVariableValueArray[i] = message[offset + i];
				}
				EventType eventType = EventTypeFactory
						.createEventType(outputVariableValueArray);
				if (eventType.getType() != EventType.OUTPUT_VARIABLE_VALUE) {
					break;
				} else {
					OutputVariableValue outputVariableValue = (OutputVariableValue) eventType;
					exchangeResults = createResults(false, false);
					exchangeResults
							.setResultPayload(outputVariableValue.marshalling());
					writeResults("Async/OutputVariables", exchangeResults);
					if (outputVariableValue.getVarCode() == 1
							|| outputVariableValue.getVarCode() == 3
							|| (outputVariableValue.getVarCode() >= 6 && outputVariableValue
									.getVarCode() <= 14)) {
						output[outputVariableValue.getVarCode()] = outputVariableValue
								.getIValue();
					} else {
						output[outputVariableValue.getVarCode()] = outputVariableValue
								.getFValue();
					}
					offset = offset
							- EventType.EventSize[EventType.OUTPUT_VARIABLE_VALUE];
				}
			} catch (Exception e) {
				break;
			}
		}
		return output;

	}

	/**
	 * Combine two messages into one byte array
	 * 
	 * @param message1
	 *            First message
	 * @param message2
	 *            Second message
	 * @return Combined byte array
	 */
	public static byte[] combineByte(byte[] message1, byte[] message2) {
//		byte[] result;
		if (message1 == null)
			return message2;
		if (message2 == null)
			return message1;
		byte[] newArray = new byte[message1.length+message2.length];
		System.arraycopy(message1, 0, newArray, 0, message1.length);
		System.arraycopy(message2, 0, newArray, message1.length, message2.length);
//		ByteBuffer byteBuffer = ByteBuffer.allocate(message1.length
//				+ message2.length);
//		byteBuffer.put(message1);
//		byteBuffer.put(message2);
//		if (message1.length == 0) {
//			return message2;
//		} else {
//			result = new byte[message1.length + message2.length];
//			for (int i = 0; i < message1.length; i++) {
//				result[i] = message1[i];
//			}
//			for (int i = 0; i < message2.length; i++) {
//				result[i + message1.length] = message2[i];
//			}
//		}
		return newArray;
	}

	/**
	 * Process function when checkpoint message is arrived.
	 * 
	 * @param nbEvent
	 *            Received message
	 * @throws MapReduceException
	 * @throws IOException
	 * @throws MarshallingException
	 */
	private synchronized void checkpointRequestProcess(NBEvent nbEvent)
			throws MapReduceException, IOException, MarshallingException {

	}

	/**
	 * Process function when checkpoint message is arrived.
	 * 
	 * @param nbEvent
	 *            Received message
	 * @throws MapReduceException
	 * @throws IOException
	 * @throws MarshallingException
	 */
	private synchronized void checkpointProcess(NBEvent nbEvent) {
		try {
			Results resultsEx = ResultsFactory.getInstance().getResults(
					nbEvent.getContentPayload());
			byte[] reportBytes = resultsEx.getResultPayload();
			Checkpoint checkpoint = new Checkpoint();
			checkpoint = checkpoint.unmarshalling(reportBytes);
			if (initialSplit) {
				initialSplit = false;
				for (int i = 1; i < numberOfWorkers; i++) {
					MultiRestore multiRestore = new MultiRestore(i, true,
							numberOfWorkers, i + 1, checkpoint.getLength(),
							checkpoint.getContent());
					exchangeResults = createResults(false, false);
					exchangeResults
							.setResultPayload(multiRestore.marshalling());
					writeResults("Sync/DownStream/" + i, exchangeResults);
				}
				getIntoNextDay(currentDay + 1);
			} else {
				// Restore restore = new Restore();
				// restore.setWorkerID(0);
				// restore.setOverwrite(false);
				// restore.setContent(checkpoint.getContent());
				// exchangeResults = createResults(false, false);
				// exchangeResults.setResultPayload(restore.marshalling());
				// writeResults("Sync/DownStream/0", exchangeResults);
				// getIntoNextDay(currentDay+1);
				// finishCheckpoint++;
				// checkpoints[checkpoint.getWorkerID()] = checkpoint;
				// if(finishCheckpoint == 2){
				// finishCheckpoint = 0;
				// MultiRestore restore0 = new MultiRestore(0, true, 1, 1,
				// checkpoints[0].getLength(),checkpoints[0].getContent());
				// MultiRestore restore1 = new MultiRestore(0, false, 1, 1,
				// checkpoints[1].getLength(),checkpoints[1].getContent());
				// byte[] bytes0 = combineByte(restore0.marshalling(),
				// restore1.marshalling());
				// exchangeResults = createResults(false, false);
				// exchangeResults
				// .setResultPayload(bytes0);
				// writeResults("Sync/DownStream/0", exchangeResults);
				// MultiRestore restore2 = new MultiRestore(1, true, 1, 0, 0,
				// new byte[0]);
				// exchangeResults = createResults(false, false);
				// exchangeResults
				// .setResultPayload(restore2.marshalling());
				// writeResults("Sync/DownStream/1", exchangeResults);
				// }

				finishCheckpoint++;
				checkpoints[checkpoint.getWorkerID()] = checkpoint;
				if (finishCheckpoint == numberOfActiveWorkers) {
					finishCheckpoint = 0;
					System.out.println("Time Begin: " + System.currentTimeMillis());
					LinkedList<byte[]> restoreMessages = createRestoreMessage(
							ratio, checkpoints);
					System.out.println("Time Create Messages: " + System.currentTimeMillis());
					for (int i = 0; i < restoreMessages.size(); i++) {
												
						exchangeResults = createResults(false, false);
						exchangeResults
								.setResultPayload(restoreMessages.get(i));
						writeResults("Sync/DownStream/" + i, exchangeResults);
						System.out.println("Time send out message "+i+" : " + System.currentTimeMillis());
					}
					System.out.println("Time End: " + System.currentTimeMillis());
					getIntoNextDay(currentDay + 1);
				}
			}

		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	/**
	 * Get the last few bytes of a byte array
	 * @param bytes
	 * @param offset
	 * @return
	 */
	public static synchronized byte[] getTail(byte[] bytes, int offset) {
		byte[] tail = new byte[bytes.length - offset];
		for (int i = 0; i < tail.length; i++) {
			tail[i] = bytes[i + offset];
		}
		return tail;
	}

	/**
	 * Reset the status of all workers
	 */
	private synchronized void resetWorkerStatus() {
		for (int i = 0; i < numberOfWorkers; i++) {
			workerDoneToday[i] = false;
			workerFinished[i] = false;
			// executionTime[i] = 0.0;
		}
	}

	/**
	 * Get list of events from a byte array.
	 * @param bytes
	 * @return
	 */
	public synchronized static LinkedList<EventType> parseEvents(byte[] bytes) {
		byte[] tail = getTail(bytes, 0);
		LinkedList<EventType> eventList = new LinkedList<EventType>();
		EventType event;
		int offset;
		while (tail.length > 0) {
			event = EventTypeFactory.createEventType(tail);
			eventList.add(event);
			if (event.getType() == EventType.OUTPUT_VARIABLE_VALUE) {
				OutputVariableValue outputVariableValue = (OutputVariableValue) event;
				offset = EventType.EventSize[EventType.OUTPUT_VARIABLE_VALUE]
						+ outputVariableValue.getNameLength();
			} else if (event.getType() == EventType.CHECKPOINT) {
				Checkpoint checkpoint = (Checkpoint) event;
				offset = EventType.EventSize[EventType.CHECKPOINT]
						+ checkpoint.getLength();
			} else if (event.getType() == EventType.RESTORE) {
				Restore restore = (Restore) event;
				offset = EventType.EventSize[EventType.RESTORE]
						+ restore.getLength();
			} else if (event.getType() == EventType.MULTI_SPLIT_BY_DENSITY) {
				MultiSplitByDensity multiSplitByDensity = (MultiSplitByDensity) event;
				offset = EventType.EventSize[EventType.MULTI_SPLIT_BY_DENSITY]
						+ multiSplitByDensity.numberOfPieces * 4 - 4;
			} else if (event.getType() == EventType.MULTI_RESTORE) {
				MultiRestore multiRestore = (MultiRestore) event;
				offset = EventType.EventSize[EventType.MULTI_RESTORE]
						+ multiRestore.getLength();
			} else {
				offset = EventType.EventSize[event.getType()];
			}
			tail = getTail(tail, offset);
		}
		return eventList;
	}

	/**
	 * Calculate the execution time of each simulation day
	 * @param day
	 * @param length
	 * @throws IOException
	 */
	private void dailyTiming(int day, int length) throws IOException {
		long newTime = System.currentTimeMillis();
		// pretty print time
		System.out.println("Day: " + day + "  Length: " + length + "  Time: "
				+ (int) (newTime - currentTime));
		// write time into files
		File file = new File(
				"/s/chopin/e/proj/granules/simonsui/Granules/trunk/bin/execution_time.txt");
		FileWriter fw;
		if (!file.exists()) {
			file.createNewFile();
			fw = new FileWriter(file, true);
		} else
			fw = new FileWriter(file, true);
		fw.write(day + "  " + length + "  " + (int) (newTime - currentTime)
				+ "\n");
		fw.close();

		File file1 = new File(
				"/s/chopin/e/proj/granules/simonsui/Granules/trunk/bin/workers.txt");
		FileWriter fw1;
		if (!file1.exists()) {
			file1.createNewFile();
			fw1 = new FileWriter(file1, true);
		} else
			fw1 = new FileWriter(file1, true);
		System.out.println("************** Working Time *************");
		for (int i = 0; i < numberOfWorkers; i++) {
			System.out.print(executionTime[i] / 1000 + " ");
			fw1.write(executionTime[i] + " ");
		}
		fw1.write("\n");
		System.out.println();
		System.out.println("************** ************ *************");
		fw1.close();
		currentTime = newTime;

	}

	/**
	 * The controller finishes all the worker from one day and get into the next simulation day
	 * @param day
	 */
	private synchronized void getIntoNextDay(int day) {
		try {
			byte[] bytes;
			StartNewDayEvent startNewDay = new StartNewDayEvent(0, day);
			// currentTime = System.currentTimeMillis();
			bytes = startNewDay.marshalling();
			byte[] finalResult = combineByte(buffer, bytes);
			buffer = new byte[0];
			exchangeResults = createResults(false, false);
			exchangeResults.setResultPayload(finalResult);
			dailyTiming(day - 1, finalResult.length);
			writeResults("Sync/DownStream", exchangeResults);
			System.out.println("All done:" + " Day " + (day - 1));
			resetWorkerStatus();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * If synopsis of received message is UPSTREAM, the events will be proceeded
	 * in this function.
	 * 
	 * @param nbEvent
	 *            Received message containing all the information
	 * @throws MapReduceException
	 * @throws ServiceException
	 * @throws NBEventException
	 * @throws IOException
	 * @throws MarshallingException
	 */
	private synchronized void upStreamProcess(NBEvent nbEvent)
			throws MapReduceException, ServiceException, NBEventException,
			IOException, MarshallingException {
		// Get results from nbEvent
		Results resultsEx = ResultsFactory.getInstance().getResults(
				nbEvent.getContentPayload());
		byte[] reportBytes = resultsEx.getResultPayload();
		LinkedList<EventType> eventList = parseEvents(reportBytes);
		// Get all output variables
		double[] outputVariables = pickupOutputVariableValue(eventList);
		// Get END_OF_DAY message
		byte[] temp = pickupEndOfDay(eventList);
		EndOfDayEvent endOfDayEvent = (EndOfDayEvent) EventTypeFactory
				.createEventType(temp);
		int day = endOfDayEvent.getDayNumber();
		if (day > currentDay + 2 || day < currentDay) {
			System.out.println("Day mismatch!");
			return;
		}
		currentDay = day;
		// System.out.println("~~~~~~~~~~~~~~Day:" + currentDay
		// + "~~~~~~~~~~~~~~~");
		try {
			// Pack all the events except END_OF_DAY event and
			// OUTPUT_VARIABLE_VALUE event
			for (int i = 0; i < eventList.size(); i++) {
				if (eventList.get(i).getType() != EventType.END_OF_DAY
						&& eventList.get(i).getType() != EventType.OUTPUT_VARIABLE_VALUE) {
					buffer = combineByte(buffer, eventList.get(i).marshalling());
				}
			}
			double[] outputVariablesFrom0 = new double[OutputVariableValue.VAR_NUM];
			for (int i = 0; i < OutputVariableValue.VAR_NUM; i++) {
				outputVariablesFrom0[i] = outputVariables[i + 1];
			}

		} catch (Exception e) {
			e.printStackTrace();
		}
		int num = endOfDayEvent.getWorkerID();
		// Record the execution time of current worker
		executionTime[num] = System.currentTimeMillis() - currentTime;
		// System.out.println("~~~~~~~~~~~~~~Worker:" + num +
		// "~~~~~~~~~~~~~~~");
		// Check whether the message is duplicated
		if (workerDoneToday[num]) {
			System.out.println("Error: Receive END_OF_DAY message from worker "
					+ num + " twice!");
			System.out.println("Previously I've got "
					+ numMessagesReceived[num] + " messages!");
			System.out.println("This time I've got " + eventList.size()
					+ " messages!");
		}
		numMessagesReceived[num] = eventList.size();
		// Mark the worker done
		workerDoneToday[num] = true;
		workerFinished[num] = endOfDayEvent.isFinished();
		int doneToday = 0;
		int finishedToday = 0;
		for (int i = 0; i < workerDoneToday.length; i++) {
			if (workerDoneToday[i])
				doneToday++;
		}
		for (int i = 0; i < workerFinished.length; i++) {
			if (workerFinished[i])
				finishedToday++;
		}
		// If all the workers are done today, go into the next day, load
		// balancing operations or simulation ends
		if (doneToday == numberOfActiveWorkers) {
			// If END_OF_SIMULATION
			if (finishedToday == numberOfActiveWorkers) {
				// End timing
				endTime = System.currentTimeMillis();
				System.out.println("Elapsed Time =" + (endTime - startTime));
				// Send END_OF_SIMULATION to workers
				EndOfSimulationEvent endOfSimulationEvent = new EndOfSimulationEvent(
						num);
				byte[] bytes = endOfSimulationEvent.marshalling();
				exchangeResults = createResults(true, true);
				exchangeResults.setResultPayload(bytes);
				writeResults("Sync/DownStream", exchangeResults);
				// Send end message to deployer
				Results results = createResults(true, true);
				String res = "END";
				results.setResultPayload(res.getBytes());
				writeResults("Sync/Result", results);
				// Daily Timing and pretty print
				dailyTiming(endOfDayEvent.getDayNumber(), bytes.length);
				System.out.println("End of simulation");
				System.exit(0);
			}
			// If not END_OF_SIMULATION
			else {
				// Condition of load balancing operations
				double averageTime = 0.0;
				double maxTime = 0.0;
				for (int i = 0; i < executionTime.length; i++) {
					averageTime += executionTime[i];
					if (executionTime[i] > maxTime)
						maxTime = executionTime[i];
				}
				averageTime = averageTime / numberOfActiveWorkers;
				if (currentDay == 1) {
					// Initial split
					initialSplit = true;
					numberOfActiveWorkers = numberOfWorkers;
					float[] portions = new float[numberOfWorkers - 1];
					for (int i = 1; i < numberOfWorkers; i++) {
						portions[i - 1] = (float) (1.0 * i / numberOfWorkers);
					}
					MultiSplitByDensity multiSplitByDensity = new MultiSplitByDensity(
							0, numberOfWorkers, portions, 1);
					exchangeResults = createResults(false, false);
					exchangeResults.setResultPayload(multiSplitByDensity
							.marshalling());
					writeResults("Sync/DownStream/0", exchangeResults);

					// } else if (currentDay == 180) {
					// RequestForFTCheckpoint request = new
					// RequestForFTCheckpoint(
					// 1, (byte) 7, false);
					// exchangeResults = createResults(false, false);
					// exchangeResults.setResultPayload(request.marshalling());
					// writeResults("Sync/DownStream/1", exchangeResults);
					// MultiSplitByDensity splitMessage0 = new
					// MultiSplitByDensity(0, 1, new float[0], 0);
					// MultiSplitByDensity splitMessage1 = new
					// MultiSplitByDensity(1, 1, new float[0], 0);
					// exchangeResults = createResults(false, false);
					// exchangeResults.setResultPayload(splitMessage0.marshalling());
					// writeResults("Sync/DownStream/0", exchangeResults);
					// exchangeResults = createResults(false, false);
					// exchangeResults.setResultPayload(splitMessage1.marshalling());
					// writeResults("Sync/DownStream/1", exchangeResults);
				} else if (maxTime - averageTime > 3000 && currentDay > 3) {
					System.out.println("Average Time = " + averageTime
							+ " Max Time = " + maxTime);
					// Record the time if received checkpoint message
					ratio = convertTimeToRatio(executionTime);
					LinkedList<MultiSplitByDensity> splitMessages = createResplitMessages(ratio);
					for (int i = 0; i < splitMessages.size(); i++) {
						exchangeResults = createResults(false, false);
						exchangeResults.setResultPayload(splitMessages.get(i)
								.marshalling());
						writeResults("Sync/DownStream/" + i, exchangeResults);
					}
				} else {
					// Or get into the next day
					getIntoNextDay(endOfDayEvent.getDayNumber() + 1);
					// resetWorkerStatus();
				}
			}
		}

	}

	/**
	 * Once receive a BOOTSTRAP message, start the first day.
	 * 
	 * @param nbEvent
	 *            Received message containing all the information.
	 * @throws MapReduceException
	 * @throws ServiceException
	 * @throws NBEventException
	 */
	private void bootstrapProcess(NBEvent nbEvent) throws MapReduceException,
			ServiceException, NBEventException {
		System.out.println("Get into bootstrap");
		// Start timing
		startTime = System.currentTimeMillis();
		currentTime = startTime;
		// Send START_NEW_DAY message to worker
		numberOfWorkers = Integer.parseInt(this.getProcessingDirectives()
				.get("Number of Workers").toString());
		workerDoneToday = new boolean[numberOfWorkers];
		workerFinished = new boolean[numberOfWorkers];
		executionTime = new double[numberOfWorkers];
		numMessagesReceived = new int[numberOfWorkers];
		numberOfActiveWorkers = 1;
		finishCheckpoint = 0;
		checkpoints = new Checkpoint[numberOfWorkers];
		resetWorkerStatus();
		StartNewDayEvent startNewDay = new StartNewDayEvent(0, 1);
		byte[] start = startNewDay.marshalling();
		exchangeResults = createResults(false, false);
		exchangeResults.setResultPayload(start);
		writeResults("Sync/DownStream", exchangeResults);
	}

	/**
	 * Constructor. Initialize variables and byte array.
	 */
	public Controller() {
		setDomain("Generic");
		setVersionInformation(200);
		setOperationIdentifier("ABCDEFGHIJKLMN");
		exchangeResults = createResults(false, false);
		finished = false;
		currentDay = 0;
		initialSplit = false;
		buffer = new byte[0];
	}

	private synchronized void syncExecution() {
		StreamingAccess access = null;
		String datasetIdentifier = "CONTROLLER_DATA";
		DatasetCollection datasetCollection = getDatasetCollection();
		try {
			Dataset dataset = datasetCollection.getDataset(datasetIdentifier);
			if (dataset == null) {
				System.out.println("NULL DATASET!");
				return;
			}
			if (dataset.getDatasetType() == Dataset.STREAMS) {
				access = (StreamingAccess) dataset;
			} else {
				System.err.println("Incorrect dataset:"
						+ dataset.getDatasetType());
				return;
			}
			if (access != null) {
				while (dataset.isDataAvailable()) {
					NBEvent nbEvent = access.getStreamingData();
					if (nbEvent == null) {
						System.out
								.println("Receive null message from Stream dataset");
						continue;
					}
					// System.out.println(nbEvent.getContentSynopsis());
					if (nbEvent.getContentSynopsis().equals("Sync/Bootstrap")) {
						bootstrapProcess(nbEvent);
					} else if (nbEvent.getContentSynopsis().equals(
							"Sync/UpStream")) {
						upStreamProcess(nbEvent);
					} else if (nbEvent.getContentSynopsis().equals(
							"Sync/Checkpoint")) {
						checkpointProcess(nbEvent);

					}
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	public void execute() throws ProcessingException {
		syncExecution();
	}

	public double[] convertTimeToRatio(double[] executionTime) {
		double[] ratio = new double[executionTime.length];
		double totalTime = 0.0;
		for (int i = 0; i < executionTime.length; i++) {
			totalTime = totalTime + executionTime[i];
		}
		double averageTime = totalTime / executionTime.length;
		int j = 0;
		double occupiedExecutionTime = 0.0;
		for (int i = 0; i < executionTime.length; i++) {
			double remainTime = averageTime;
			if (i == 0)
				occupiedExecutionTime = executionTime[j];
			while (true) {
				if (remainTime - occupiedExecutionTime <= 1e-8) {
					occupiedExecutionTime = occupiedExecutionTime - remainTime;
					ratio[i] = j + (executionTime[j] - occupiedExecutionTime)
							/ executionTime[j];
					break;
				} else {
					remainTime = remainTime - occupiedExecutionTime;
					j++;
					occupiedExecutionTime = executionTime[j];
				}
			}
		}
		return ratio;
	}

	/**
	 * From the ratio of work load, get the resplit messages.
	 * @param ratio
	 * @return
	 */
	public LinkedList<MultiSplitByDensity> createResplitMessages(double[] ratio) {
		LinkedList<MultiSplitByDensity> splitList = new LinkedList<MultiSplitByDensity>();
		int index = 1;
		int i = 0;
		LinkedList<Double> proportion = new LinkedList<Double>();
		while (i < ratio.length) {
			if (ratio[i] < index) {
				proportion.add(new Double(ratio[i] - index + 1));
				i++;
			} else {
				if (!proportion.isEmpty()) {
					float[] prop = new float[proportion.size()];
					for (int j = 0; j < prop.length; j++) {
						prop[j] = (float) (double) proportion.get(j);
					}
					MultiSplitByDensity split = new MultiSplitByDensity(
							index - 1, proportion.size() + 1, prop, 0);
					proportion.clear();
					splitList.add(split);
					index++;
				} else {
					while (ratio[i] >= index) {
						MultiSplitByDensity split = new MultiSplitByDensity(
								index - 1, 1, new float[] {}, 0);
						splitList.add(split);
						proportion.clear();
						index++;
						if (ratio[i] == index - 1) {
							i++;
							break;
						}
					}
				}
			}
		}
		return splitList;
	}

	/**
	 * From the ratio of work load and checkpoint messages, create the restore messages.
	 * @param ratio
	 * @param checkpoints
	 * @return
	 */
	
	public LinkedList<byte[]> createRestoreMessage(double[] ratio,
			Checkpoint[] checkpoints) {
		LinkedList<byte[]> restoreList = new LinkedList<byte[]>();
		for (int i = 0; i < ratio.length; i++) {
			if (i == 0) {
				byte[] message = new byte[0];
				for (int j = 0; j < ratio[i]; j++) {
					MultiRestore multiRestore = new MultiRestore(i, j == 0,
							checkPieceNumber(ratio, j), 1,
							checkpoints[j].getLength(),
							checkpoints[j].getContent());
					// printEventMessage(multiRestore);
					message = combineByte(message, multiRestore.marshalling());
				}
				restoreList.add(message);
			} else {
				if (Math.ceil(ratio[i - 1]) >= ratio[i]) {
					int pieceNum = checkPieceNumber(ratio,
							(int) Math.floor(ratio[i - 1]));
					int activeNum = checkActivePiece(ratio, i);
					MultiRestore multiRestore = new MultiRestore(i, true,
							pieceNum, activeNum,
							checkpoints[(int) Math.floor(ratio[i - 1])]
									.getLength(),
							checkpoints[(int) Math.floor(ratio[i - 1])]
									.getContent());
					// printEventMessage(multiRestore);
					restoreList.add(multiRestore.marshalling());
				} else {
					byte[] message = new byte[0];
					if ((Math.ceil(ratio[i - 1]) > Math.floor(ratio[i - 1]))) {
						int pieceNum = checkPieceNumber(ratio,
								(int) Math.floor(ratio[i - 1]));
						MultiRestore multiRestore = new MultiRestore(i, true,
								pieceNum, pieceNum,
								checkpoints[(int) Math.floor(ratio[i - 1])]
										.getLength(),
								checkpoints[(int) Math.floor(ratio[i - 1])]
										.getContent());
						// printEventMessage(multiRestore);
						message = combineByte(message,
								multiRestore.marshalling());
					}
					for (int j = (int) Math.ceil(ratio[i - 1]); j < ratio[i]; j++) {
						boolean first = Math.ceil(ratio[i - 1]) == Math
								.floor(ratio[i - 1])
								&& j == Math.ceil(ratio[i - 1]);
						MultiRestore multiRestore = new MultiRestore(i, first,
								checkPieceNumber(ratio, j), 1,
								checkpoints[j].getLength(),
								checkpoints[j].getContent());
						// printEventMessage(multiRestore);
						message = combineByte(message,
								multiRestore.marshalling());
					}
					restoreList.add(message);
				}
			}
		}
		return restoreList;
	}

	/**
	 * Calculate the number that one worker will be split into 
	 * @param ratio
	 * @param num
	 * @return
	 */
	public int checkPieceNumber(double[] ratio, int num) {
		int pieceNumber = 1;
		for (int i = 0; i < ratio.length; i++) {
			if (Math.floor(ratio[i]) == num && Math.ceil(ratio[i]) == num + 1)
				pieceNumber++;
		}
		return pieceNumber;
	}

	/**
	 * Calculate the number of active pieces of one worker.
	 * @param ratio
	 * @param num
	 * @return
	 */
	public int checkActivePiece(double[] ratio, int num) {
		int activePiece = 1;
		for (int i = num - 1; i >= 0; i--) {
			if (Math.ceil(ratio[i]) >= ratio[num])
				activePiece++;
			else
				break;
		}
		return activePiece;
	}
}
