/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.naadsm.staticsplit;

/**
 * This is the controller for distributed NAADSM. It is responsible for 
 * synchronization at the end of simulation days and finalizing data at the
 * end of simulation.
 * 
 * @author Zhiquan (Simon) Sui
 */
import java.io.IOException;

import cgl.narada.event.NBEvent;
import cgl.narada.event.NBEventException;
import cgl.narada.service.ServiceException;
import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.naadsm.staticsplit.event.EndOfDayEvent;
import ds.granules.naadsm.staticsplit.event.EndOfSimulationEvent;
import ds.granules.naadsm.staticsplit.event.EventType;
import ds.granules.naadsm.staticsplit.event.EventTypeFactory;
import ds.granules.naadsm.staticsplit.event.StartNewDayEvent;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProcessingException;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

public class Controller extends MapReduceBase {

	private boolean[] finishedWorker;
	private boolean[] doneWorker;
	private boolean finished;
	private int numberOfWorkers;
	private Results exchangeResults;
	private byte[] buffer;
	long startTime;
	long endTime;

	/**
	 * Constructor, allocate memory for local buffer and initialize signals
	 * variables.
	 */
	public Controller() {
		startTime = System.currentTimeMillis();
		setDomain("Generic");
		setVersionInformation(200);
		setOperationIdentifier("ABCDEFGHIJKLMN");
		finished = false;
		exchangeResults = createResults(false, false);
		buffer = new byte[0];
	}

	/**
	 * Implement execute() function in MapReduceBase class. Once a message with
	 * registered synopsis is sent out, execute() will be activated.
	 */
	@Override
	public void execute() throws ProcessingException {
		StreamingAccess access = null;
		String datasetIdentifier = "CONTROLLER_DATA";
		DatasetCollection datasetCollection = getDatasetCollection();
		try {
			// Check data set
			Dataset dataset = datasetCollection.getDataset(datasetIdentifier);
			if (dataset == null) {
				System.err.println("NULL DATASET");
				return;
			}
			if (dataset.getDatasetType() == Dataset.STREAMS) {
				access = (StreamingAccess) dataset;
			} else {
				System.err.println("Incorrect dataset:"
						+ dataset.getDatasetType());
				return;
			}
			// Check stream access
			if (access != null) {
				while (dataset.isDataAvailable()) {
					NBEvent nbEvent = access.getStreamingData();
					if (nbEvent == null) {
						System.err
								.println("Receive null message from Stream dataset");
						continue;
					}
					// Judge synopsis
					if (nbEvent.getContentSynopsis().equals("Sync/Bootstrap")) {
						bootstrapProcess(nbEvent);
					} else if (nbEvent.getContentSynopsis().equals(
							"Sync/UpStream")) {
						upStreamProcess(nbEvent);
					}
					if (finished)
						access.close();

				}
			}
		} catch (DatasetException e) {
			System.out.println("DatasetException");
			e.printStackTrace();
		} catch (ServiceException e) {
			System.out.println("ServiceException");
			e.printStackTrace();
		} catch (NBEventException e) {
			System.out.println("NBEventException");
			e.printStackTrace();
		} catch (IOException e) {
			System.out.println("IOException");
			e.printStackTrace();
		} catch (MapReduceException e) {
			System.out.println("MapReduceException");
			e.printStackTrace();
		} catch (MarshallingException e) {
			System.out.println("MarshallingException");
			e.printStackTrace();
		}

	}

	/**
	 * From a whole message, pick up the END_OF_DAY message
	 * 
	 * @param message
	 *            The whole message
	 * @return END_OF_DAY message
	 */
	private byte[] pickupEndOfDay(byte[] message) {
		int length = message.length;
		byte[] endOfDayArray = new byte[EventType.EventSize[EventType.END_OF_DAY]];
		for (int i = 0; i < EventType.EventSize[EventType.END_OF_DAY]; i++) {
			endOfDayArray[i] = message[length
					- EventType.EventSize[EventType.END_OF_DAY] + i];
		}
		return endOfDayArray;
	}

	/**
	 * Combine two messages into one byte array
	 * 
	 * @param message1
	 *            First message
	 * @param message2
	 *            Second message
	 * @return Combined byte array
	 */
	private byte[] combineByte(byte[] message1, byte[] message2) {
		byte[] result;
		if (message1.length == 0) {
			result = new byte[message2.length
					- EventType.EventSize[EventType.END_OF_DAY]];
			for (int i = 0; i < message2.length
					- EventType.EventSize[EventType.END_OF_DAY]; i++) {
				result[i] = message2[i];
			}
		} else {
			result = new byte[message1.length + message2.length
					- EventType.EventSize[EventType.END_OF_DAY]];
			for (int i = 0; i < message1.length; i++) {
				result[i] = message1[i];
			}
			for (int i = 0; i < message2.length
					- EventType.EventSize[EventType.END_OF_DAY]; i++) {
				result[i + message1.length] = message2[i];
			}
		}
		return result;
	}

	/**
	 * If synopsis of received message is UPSTREAM, the events will be proceeded
	 * in this function.
	 * 
	 * @param nbEvent
	 *            Received message containing all the information
	 * @throws MapReduceException
	 * @throws ServiceException
	 * @throws NBEventException
	 * @throws IOException
	 * @throws MarshallingException
	 */
	private synchronized void upStreamProcess(NBEvent nbEvent)
			throws MapReduceException, ServiceException, NBEventException,
			IOException, MarshallingException {
		/* Initialization */
		int numberOfWorkers = Integer.parseInt(this.getProcessingDirectives()
				.get("Number of Workers").toString());
		Results resultsEx = ResultsFactory.getInstance().getResults(
				nbEvent.getContentPayload());
		byte[] reportBytes = resultsEx.getResultPayload();
		buffer = combineByte(buffer, reportBytes);
		/* Get END_OF_DAY message from the whole message */
		EndOfDayEvent endOfDayEvent = (EndOfDayEvent) EventTypeFactory
				.createEventType(pickupEndOfDay(reportBytes));
		/* Check end of simulation condition */
		if (endOfDayEvent.isFinished()) {
			finishedWorker[endOfDayEvent.getWorkerID()] = true;
			doneWorker[endOfDayEvent.getWorkerID()] = true;
			boolean flag = true;
			for (int i = 0; i < numberOfWorkers; i++) {
				if (!finishedWorker[i]) {
					flag = false;
					break;
				}
			}
			if (flag) {
				/* If end of simulation condition is met, do finalization */
				System.out.println("All finished:" + " Day "
						+ endOfDayEvent.getDayNumber());
				endTime = System.currentTimeMillis();
				System.out.println("Elapsed Time =" + (endTime - startTime));
				EndOfSimulationEvent endOfSimulationEvent = new EndOfSimulationEvent();
				byte[] bytes = endOfSimulationEvent.marshalling();
				byte[] finalResult = new byte[buffer.length
						+ EventType.EventSize[EventType.END_OF_SIMULATION] + 1];
				for (int i = 0; i < buffer.length; i++) {
					finalResult[i] = buffer[i];
				}
				for (int i = 0; i < bytes.length; i++) {
					finalResult[buffer.length + i] = bytes[i];
				}
				finalResult[buffer.length
						+ EventType.EventSize[EventType.END_OF_SIMULATION]] = 1;
				Results exchangeResults = createResults(true, true);
				exchangeResults.setResultPayload(finalResult);
				writeResults("Sync/DownStream", exchangeResults);
				Results results = createResults(true, true);
				String res = "END";
				results.setResultPayload(res.getBytes());
				writeResults("Sync/Result", results);
				System.out
						.println("Invoked Termination Condition Reached on Controller.");
				return;
			}
		} else {
			doneWorker[endOfDayEvent.getWorkerID()] = true;
		}
		/* Check whether all the workers finish the current day */
		boolean flag = true;
		for (int i = 0; i < numberOfWorkers; i++) {
			if (!doneWorker[i]) {
				flag = false;
				break;
			}
		}
		/*
		 * If all the workers finish the current day, broadcast START_NEW_DAY
		 * message for the next day.
		 */
		if (flag) {
			for (int i = 0; i < numberOfWorkers; i++) {
				doneWorker[i] = false;
				finishedWorker[i] = false;
			}
			byte[] bytes;
			StartNewDayEvent startNewDay = new StartNewDayEvent(0,
					endOfDayEvent.getDayNumber() + 1);
			System.out.println("All done:" + " Day "
					+ endOfDayEvent.getDayNumber());
			bytes = startNewDay.marshalling();
			byte[] finalResult = new byte[buffer.length
					+ EventType.EventSize[EventType.START_NEW_DAY] + 1];
			for (int i = 0; i < buffer.length; i++) {
				finalResult[i] = buffer[i];
			}
			for (int i = 0; i < bytes.length; i++) {
				finalResult[buffer.length + i] = bytes[i];
			}
			finalResult[buffer.length
					+ EventType.EventSize[EventType.START_NEW_DAY]] = 0;
			buffer = new byte[0];
			Results exchangeResults = createResults(false, false);

			exchangeResults.setResultPayload(finalResult);
			writeResults("Sync/DownStream", exchangeResults);
		}
	}

	/**
	 * Once receive a BOOTSTRAP message, start the first day.
	 * @param nbEvent
	 *            Received message containing all the information.
	 * @throws MapReduceException
	 * @throws ServiceException
	 * @throws NBEventException
	 */
	private void bootstrapProcess(NBEvent nbEvent) throws MapReduceException,
			ServiceException, NBEventException {
		/* Initialize everything */
		numberOfWorkers = Integer.parseInt(this.getProcessingDirectives()
				.get("Number of Workers").toString());
		finishedWorker = new boolean[numberOfWorkers];
		doneWorker = new boolean[numberOfWorkers];
		for (int i = 0; i < numberOfWorkers; i++) {
			finishedWorker[i] = false;
			doneWorker[i] = false;
		}
		StartNewDayEvent startEvent = new StartNewDayEvent(0, 1);
		byte[] startMessage = new byte[EventType.EventSize[EventType.START_NEW_DAY] + 1];
		byte[] start = startEvent.marshalling();
		for (int i = 0; i < EventType.EventSize[EventType.START_NEW_DAY]; i++) {
			startMessage[i] = start[i];
		}
		startMessage[EventType.EventSize[EventType.START_NEW_DAY]] = 0;
		exchangeResults.setResultPayload(startMessage);
		writeResults("Sync/DownStream", exchangeResults);

	}

}
