/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.naadsm.staticsplit;

/**
 * Deployer is to deploy the controller and the workers. It initializes the parameters.
 * 
 * @author Zhiquan (Simon) Sui 
 */
import java.io.IOException;
import java.util.Properties;

import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.DatasetFactory;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.CommunicationsException;
import ds.granules.exception.DeploymentException;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.InstanceDeployer;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProgressTracker;

public class Deployer extends InstanceDeployer {
	private Properties props;
	private int finishedJob;
	/**
	 * Constructor. Initialization.
	 * @param properties
	 *         Contains parameters
	 * @throws CommunicationsException
	 * @throws IOException
	 * @throws MarshallingException
	 * @throws DeploymentException
	 */
	public Deployer(Properties properties) throws CommunicationsException,
			IOException, MarshallingException, DeploymentException {
		finishedJob = 0;
		props = properties;
		initialize(properties);
	}

	/**
	 * Initialize one worker
	 * @param workerNum
	 *         Number of the current worker.
	 * @param numH
	 *         Number of total workers in log based scale in horizontal.
	 * @param numV
	 *         Number of total workers in log based scale in vertical.
	 * @param up
	 *         Upper bower of the region.
	 * @param down
	 *         Lower bower of the region.
	 * @param left
	 *         Left bower of the region.
	 * @param right
	 *         Right bower of the region.
	 * @return
	 *         The created worker
	 * @throws DatasetException
	 * @throws MapReduceException
	 */
	public Worker initializeWorker(int workerNum, int numH, int numV, int up,
			int down, int left, int right) throws DatasetException,
			MapReduceException {
		Worker worker = new Worker();
		/* Put all the integer parameters into properties in string. */
		Properties props = (Properties) this.props.clone();
		String key = "Worker Number";
		String value = new String();
		value = "" + workerNum;
		props.put(key, value);
		String key_H = "Number H";
		String value_H = "" + numH;
		props.put(key_H, value_H);
		String key_V = "Number V";
		String value_V = "" + numV;
		props.put(key_V, value_V);
		String upKey = "Upper Bound";
		String upValue = "" + up;
		props.put(upKey, upValue);
		String downKey = "Lower Bound";
		String downValue = "" + down;
		props.put(downKey, downValue);
		String leftKey = "Left Bound";
		String leftValue = "" + left;
		props.put(leftKey, leftValue);
		String rightKey = "Right Bound";
		String rightValue = "" + right;
		props.put(rightKey, rightValue);
		worker.setProcessingDirectives(props);

		DatasetFactory datasetFactory = DatasetFactory.getInstance();
		StreamingAccess streamAccess = datasetFactory.createStreamingDataset(
				"WORKER_DATA" + value, "WORKER DATA");

		/* Register synopsis of the worker */
		String downstreamSynopsis = "Sync/DownStream";
		streamAccess.addInputStream(downstreamSynopsis,
				StreamingAccess.STRING_SYNOPSIS);

		DatasetCollection datasetCollection = worker.getDatasetCollection();
		datasetCollection.addDataset(streamAccess);
		worker.setDatasetCollection(datasetCollection);
		worker.setAsExecuteWhenDataAvailable();
		return worker;
	}

	/**
	 * Initialize the controller
	 * @param numOfWorkers
	 *         Total number of workers
	 * @return
	 *         The controller
	 * @throws DatasetException
	 * @throws MapReduceException
	 */
	public Controller initializeController(int numOfWorkers)
			throws DatasetException, MapReduceException {
		Controller controller = new Controller();
		Properties props = (Properties) this.props.clone();
		String key = "Number of Workers";
		String value = "" + numOfWorkers;
		props.put(key, value);
		controller.setProcessingDirectives(props);

		DatasetFactory datasetFactory = DatasetFactory.getInstance();
		StreamingAccess streamAccess = datasetFactory.createStreamingDataset(
				"CONTROLLER_DATA", "CONTROLLER DATA");

		/* Register the synopsis of controller */
		String bootstrapSynopsis = "Sync/Bootstrap";
		String upstreamSynopsis = "Sync/UpStream";
		streamAccess.addInputStream(bootstrapSynopsis,
				StreamingAccess.STRING_SYNOPSIS);
		streamAccess.addInputStream(upstreamSynopsis,
				StreamingAccess.STRING_SYNOPSIS);

		DatasetCollection datasetCollection = controller.getDatasetCollection();
		datasetCollection.addDataset(streamAccess);
		controller.setDatasetCollection(datasetCollection);
		controller.setAsExecuteWhenDataAvailable();
		return controller;
	}

	/**
	 * Initialize the controller and all the workers
	 * @param numH
	 *         Number of total workers in log based scale in horizontal.
	 * @param numV
	 *         Number of total workers in log based scale in vertical.
	 * @param up
	 *         Upper bower of the region
	 * @param down
	 *         Lower bower of the region
	 * @param left
	 *         Left bower of the region
	 * @param right
	 *         Right bower of the region
	 */
	public void prepareInstancesAndDeploy(int numH, int numV, int up, int down,
			int left, int right) {
		int numOfWorkers = (int) (Math.pow(2, numH) * Math.pow(2, numV));
		MapReduceBase[] maps = new MapReduceBase[1 + numOfWorkers];

		try {
			for (int i = 0; i < numOfWorkers; i++) {
				maps[i] = initializeWorker(i, numH, numV, up, down, left, right);
			}
			maps[numOfWorkers] = initializeController(numOfWorkers);
			deployOperations(maps);
		} catch (CommunicationsException e) {
			e.printStackTrace();
		} catch (DeploymentException e) {
			e.printStackTrace();
		} catch (MarshallingException e) {
			e.printStackTrace();
		} catch (DatasetException e) {
			e.printStackTrace();
		} catch (MapReduceException e) {
			e.printStackTrace();
		}

	}

	@Override
	public void completedSuccessfully(String jobIdentifier,
			ProgressTracker progressTracker) {
		finishedJob++;
		if (finishedJob == 1) {
			System.exit(0);
		}
	}

	/**
	 * @param args
	 *            args[0] is host name of narada brokering. args[1] is port
	 *            number of narada brokering. args[2] and args[3] are log based
	 *            number of workers in horizontal and vertical directions.
	 *            args[4] through args[7] are upper, lower, left and right
	 *            boundaries of the area. args[8] is the command line of the
	 *            workers.
	 * 
	 */
	public static void main(String[] args) {
		String hostname = args[0];
		String portnum = args[1];
		String numHS = args[2];
		String numVS = args[3];
		String upS = args[4];
		String downS = args[5];
		String leftS = args[6];
		String rightS = args[7];
		String cmd = args[8];
		Properties streamProperties = new Properties();
		streamProperties.put("hostname", hostname);
		streamProperties.put("portnum", portnum);
		streamProperties.put("cmdLine", cmd);
		int numH = Integer.parseInt(numHS);
		int numV = Integer.parseInt(numVS);
		int up = Integer.parseInt(upS);
		int down = Integer.parseInt(downS);
		int left = Integer.parseInt(leftS);
		int right = Integer.parseInt(rightS);

		try {
			Deployer deployer = new Deployer(streamProperties);
			deployer.prepareInstancesAndDeploy(numH, numV, up, down, left,
					right);
		} catch (CommunicationsException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (MarshallingException e) {
			e.printStackTrace();
		} catch (DeploymentException e) {
			e.printStackTrace();
		}
	}

}
