/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.naadsm.staticsplit;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import cgl.narada.event.NBEvent;
import cgl.narada.event.NBEventException;
import cgl.narada.service.ServiceException;
import ds.granules.bridge.ControlMessage;
import ds.granules.bridge.JavaByteMessage;
import ds.granules.bridge.JavaGenericComputation;
import ds.granules.bridge.JavaMessageHandler;
import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.naadsm.staticsplit.event.EventType;
import ds.granules.naadsm.staticsplit.event.EventTypeFactory;
import ds.granules.naadsm.staticsplit.event.SplitEvent;
import ds.granules.naadsm.staticsplit.event.StartNewDayEvent;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProcessingException;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;
/**
 * This is the worker for distributed NAADSM. It is responsible for 
 * communicate the NAADSM project with controllers.
 * 
 * @author Zhiquan (Simon) Sui
 *
 */
public class Worker extends MapReduceBase implements JavaMessageHandler {
	JavaGenericComputation jc;
	private boolean finished;
	private byte[] buffer;
	FileWriter writer;
	BufferedWriter out;
	boolean started;
	String sysCommand;
	private Results exchangeResults;
	private boolean silent;
	File output;

	/**
	 * Combine two messages into one byte array.
	 * @param message1
	 * @param message2
	 * @return combined byte array
	 */
	private byte[] combineByte(byte[] message1, byte[] message2) {
		byte[] result;
		if (message1.length == 0) {
			result = new byte[message2.length];
			for (int i = 0; i < message2.length; i++) {
				result[i] = message2[i];
			}
		} else {
			result = new byte[message1.length + message2.length];
			for (int i = 0; i < message1.length; i++) {
				result[i] = message1[i];
			}
			for (int i = 0; i < message2.length; i++) {
				result[i + message1.length] = message2[i];
			}
		}
		return result;
	}

	/**
	 * Constructor. Initialize the variables and the byte array. 
	 */
	public Worker() {
		setDomain("Generic");
		setVersionInformation(200);
		setOperationIdentifier("ABCDEFGHIJKLMN");
		// default command line
		sysCommand = "/s/chopin/e/proj/granules/simonsui/dis-naadsm/naadsm-3.2.7/bin/naadsm -h /s/chopin/e/proj/granules/simonsui/dis-naadsm/660000_population_scaled_from_kansas_9_infected.xml /s/chopin/e/proj/granules/simonsui/dis-naadsm/Kansas_params_SlowerDetection_HigherInfectionRate.xml";
		started = false;
		finished = false;
		silent = true; // true means non-debugging model, false means debugging-model
		jc = new JavaGenericComputation("Worker");
		buffer = new byte[0];
		exchangeResults = createResults(false, false);
	}

	@Override
	public void execute() throws ProcessingException {
		try {
			// The first time the worker is called.
			if (!started) {
				started = true; // never come into this part again.
				try {
					// Start NAADSM program
					if (this.getProcessingDirectives().getProperty("cmdLine")!="")
						sysCommand = this.getProcessingDirectives().getProperty("cmdLine");
					jc.startExecution(sysCommand, ControlMessage.UPIPES, this);
					// Set parameter
					int num = Integer.parseInt(this.getProcessingDirectives()
							.getProperty("Worker Number"));
					// Initialize output file
					output = new File("./output" + num + ".log");
					writer = new FileWriter(output);
					out = new BufferedWriter(writer);
					// How many pieces in each horizontal strip log based 2
					int log2H = Integer.parseInt(this.getProcessingDirectives()
							.getProperty("Number H"));
					// How many pieces in each vertical strip in log based 2
					int log2V = Integer.parseInt(this.getProcessingDirectives()
							.getProperty("Number V"));
					int v = (int) Math.pow(2, log2V);
					int h = (int) Math.pow(2, log2H);
					float leftBound = Integer.parseInt(this.getProcessingDirectives()
							.getProperty("Left Bound"));;
					float rightBound = Integer.parseInt(this.getProcessingDirectives()
							.getProperty("Right Bound"));;
					float lowerBound = Integer.parseInt(this.getProcessingDirectives()
							.getProperty("Lower Bound"));;
					float upperBound = Integer.parseInt(this.getProcessingDirectives()
							.getProperty("Upper Bound"));;
					// Split the region statically
					for (int i = 0; i < log2H; i++) {
						float where = (leftBound + rightBound) / 2;
						int pivot = (int) Math.pow(2, log2H - i);
						if ((num % pivot) < pivot / 2) {
							SplitEvent splitEvent = new SplitEvent(num, 1,
									where);
							rightBound = where;
							byte[] send = splitEvent.marshalling();
							buffer = combineByte(buffer, send);
						} else {
							SplitEvent splitEvent = new SplitEvent(num, 2,
									where);
							leftBound = where;
							byte[] send = splitEvent.marshalling();
							buffer = combineByte(buffer, send);
						}
					}
					for (int i = 0; i < log2V; i++) {
						float where = (lowerBound + upperBound) / 2;
						int pivot = (int) Math.pow(2, log2V - i);
						if (((num / h) % pivot) < pivot / 2) {
							SplitEvent splitEvent = new SplitEvent(num, 3,
									where);
							System.out.println("Number of worker: "+num);
							System.out.println("Split way: 3");
							System.out.println("Where: " + where);
							lowerBound = where;
							byte[] send = splitEvent.marshalling();
							buffer = combineByte(buffer, send);
						} else {
							SplitEvent splitEvent = new SplitEvent(num, 4,
									where);
							System.out.println("Number of worker: "+num);
							System.out.println("Split way: 4");
							System.out.println("Where: " + where);
							upperBound = where;
							byte[] send = splitEvent.marshalling();
							buffer = combineByte(buffer, send);
						}
					}
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
			StreamingAccess access = null;
			String datasetIdentifier = "WORKER_DATA"
					+ this.getProcessingDirectives().getProperty(
							"Worker Number");
			DatasetCollection datasetCollection = getDatasetCollection();
			if (datasetCollection == null)
				return;
			try {
				Dataset dataset = datasetCollection
						.getDataset(datasetIdentifier);
				if (dataset == null) {
					System.err
							.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!DATASET IS NULL!!!!!!!!!!!!!!!");
					return;
				}
				if (dataset.getDatasetType() == Dataset.STREAMS) {
					access = (StreamingAccess) dataset;
				} else {
					System.err.println("Incorrect dataset:"
							+ dataset.getDatasetType());
					return;
				}
				if (access != null) {
					while (dataset.isDataAvailable()) {
						NBEvent nbEvent = access.getStreamingData();
						if (nbEvent == null) {
							continue;
						}
						// If synopsis is DownStream, go into the process function.
						if (nbEvent.getContentSynopsis().equals(
								"Sync/DownStream")) {
							downStreamProcess(nbEvent);
							if (finished)
								;
						}
					}
				} else {
					System.err.println("The dataset is null!");
				}
			} catch (DatasetException e) {
				System.out.println("DatasetException");
				e.printStackTrace();
			} catch (ServiceException e) {
				System.out.println("ServiceException");
				e.printStackTrace();
			} catch (NBEventException e) {
				System.out.println("NBEventException");
				e.printStackTrace();
			} catch (IOException e) {
				System.out.println("IOException");
				e.printStackTrace();
			} catch (MapReduceException e) {
				System.out.println("MapReduceException");
				e.printStackTrace();
			} catch (MarshallingException e) {
				System.out.println("MarshallingException");
				e.printStackTrace();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Process messages with synopsis of DownStream 
	 * @param nbEvent
	 *         Message containing all information
	 * @throws ServiceException
	 * @throws NBEventException
	 * @throws IOException
	 * @throws MapReduceException
	 * @throws MarshallingException
	 * @throws NullPointerException
	 */
	private void downStreamProcess(NBEvent nbEvent) throws ServiceException,
			NBEventException, IOException, MapReduceException,
			MarshallingException, NullPointerException {
		// Set parameter worker number
		int num = Integer.parseInt(this.getProcessingDirectives().getProperty(
				"Worker Number"));
		Results resultsEx = ResultsFactory.getInstance().getResults(
				nbEvent.getContentPayload());
		// Check message
		if (resultsEx == null) {
			return;
		}
		byte[] send = resultsEx.getResultPayload();

		if (send == null) {
			return;
		}
		/* Check whether it is the first time sending message. If so, combine
		 * split messages into sending message.
		*/
		if (buffer != null) {
			byte[] extraSend = new byte[send.length + buffer.length];
			for (int i = 0; i < buffer.length; i++) {
				extraSend[i] = buffer[i];
			}
			for (int i = 0; i < send.length; i++) {
				extraSend[i + buffer.length] = send[i];
			}
			buffer = null;
			send = extraSend;
		}
		/*
		 * If the last digit is 0, the simulation is not finished. START_NEW_DAY 
		 * message is sent to NAADSM. Or END_SIMULATION message is sent to NAADSM.
		*/ 
		if (send[send.length - 1] == 0) {
			JavaByteMessage jbm = new JavaByteMessage(ControlMessage.DATA);
			byte[] start = new byte[EventType.EventSize[EventType.START_NEW_DAY]];
			for (int i = 0; i < EventType.EventSize[EventType.START_NEW_DAY]; i++) {
				start[i] = send[send.length - 1
						- EventType.EventSize[EventType.START_NEW_DAY] + i];
			}
			StartNewDayEvent startEvent = (StartNewDayEvent) EventTypeFactory
					.createEventType(start);
			startEvent.setWorker(num);
			start = startEvent.marshalling();
			byte[] bytes = new byte[send.length - 1];
			for (int i = 0; i < send.length - 1; i++) {
				bytes[i] = send[i];
			}
			for (int i = 0; i < EventType.EventSize[EventType.START_NEW_DAY]; i++) {
				bytes[bytes.length
						- EventType.EventSize[EventType.START_NEW_DAY] + i] = start[i];
			}
			jbm.setInput(bytes);
			byte[] totalMessage = jbm.toByteArray();
			if (!silent) {
				out.write("----------------ALL Send to NAADSM:-------------------\n");
				for (int i = 0; i < totalMessage.length; i++)
					out.write(Integer.toHexString(totalMessage[i]) + " ");
				out.newLine();
				System.out
						.println("----------------ALL Send to NAADSM:-------------------");
				for (int i = 0; i < totalMessage.length; i++)
					System.out
							.print(Integer.toHexString(totalMessage[i]) + " ");
				System.out.println();
				System.out
						.println("---------------------------------------------------");
			}
			jc.sendMessage(jbm);

		} else if (send[send.length - 1] == 1) {
			JavaByteMessage jbm = new JavaByteMessage(ControlMessage.DATA);
			byte[] bytes = new byte[send.length - 1];
			for (int i = 0; i < send.length - 1; i++) {
				bytes[i] = send[i];
			}
			jbm.setInput(bytes);
			byte[] totalMessage = jbm.toByteArray();
			if (!silent) {
				out.write("----------------ALL Send to NAADSM:-------------------\n");
				for (int i = 0; i < totalMessage.length; i++)
					out.write(Integer.toHexString(totalMessage[i]) + " ");
				out.newLine();
				System.out
						.println("----------------ALL Send to NAADSM:-------------------");
				for (int i = 0; i < totalMessage.length; i++)
					System.out
							.print(Integer.toHexString(totalMessage[i]) + " ");
				System.out.println();
				System.out
						.println("---------------------------------------------------");
			}

			jc.sendMessage(jbm);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see ds.granules.bridge.JavaMessageHandler#handleMessage(byte[])
	 * Once got message from NAADSM project, this function will be activated.
	 */
	@Override
	public void handleMessage(final byte[] msg) {
		final byte[] par = msg;
		if (!silent) {
			try {
				out.write("----------------Got From NAADSM:-------------------\n");
				for (int i = 0; i < par.length; i++)
					out.write(Integer.toHexString(par[i]) + " ");
				out.newLine();
				System.out
						.println("----------------Got From NAADSM:-------------------");
				for (int i = 0; i < par.length; i++)
					System.out.print(Integer.toHexString(par[i]) + " ");
				System.out.println();
				System.out
						.println("---------------------------------------------------");
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		synchronized (exchangeResults) {
			JavaByteMessage jbm = new JavaByteMessage(par);
			Results exchangeResults = createResults(false, false);
			exchangeResults.setResultPayload(jbm.getInput());
			try {
				writeResults("Sync/UpStream", exchangeResults);
			} catch (MapReduceException e) {
				e.printStackTrace();
			}
		}

	}
}
