/**
 * Software License, Version 1.0
 * 
 * Copyright 2008 The Trustees of Indiana University.  All rights reserved.
 * 
 *
 *Redistribution and use in source and binary forms, with or without 
 *modification, are permitted provided that the following conditions are met:
 *
 *1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 *2) All redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 *3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Community Grids Lab. For 
 * further information contact the Community Grids Lab at 
 * http://communitygrids.iu.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 *4) The name Indiana University or Community Grids Lab or Granules, 
 * shall not be used to endorse or promote products derived from this software 
 * without prior written permission from Indiana University.  For written 
 * permission, please contact the Advanced Research and Technology Institute 
 * ("ARTI") at 351 West 10th Street, Indianapolis, Indiana 46202.
 *5) Products derived from this software may not be called Granules, 
 * nor may Indiana University or Community Grids Lab or Granules appear
 * in their name, without prior written permission of ARTI.
 * 
 *
 * Indiana University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Indiana University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.operation;

/**
 * This interfaces encapsulates the execution profile of the operation. An
 * operation can be one which <li>Executes exactly once and terminates <li>
 * Executes periodically at the specified intervals <li>Executes a fixed number
 * of times <li>Executes till a boolean flag has been set to true
 * 
 * @author Shrideep Pallickara
 * 
 */
public interface ExecutionProfile {

    /**
     * This indicates if the operation needs to be executed exactly once
     * 
     * @return <code>true</code> if this is to be executed once;
     *         <code>false</code> otherwise.
     */
    public boolean isExecuteOnce();


    /**
     * Indicates whether the operation needs to be executed periodically.
     * 
     * @return <code>true</code> if this is to be executed periodically;
     *         <code>false</code> otherwise.
     */
    public boolean isExecutePeriodically();


    /**
     * Indicates whether the operation needs to be executed a fixed number of
     * times
     * 
     * @return <code>true</code> if the operation needs to be executed a fixed
     *         number of times;<code>false</code> otherwise.
     */
    public boolean isExcuteAFixedNumberOfTimes();


    /**
     * Retrieve the execution interval. A value of 0 indicates that the
     * operation needs not be executed periodically. The upper-bound for this
     * value depends on the policies at individual nodes
     * 
     * @return The execution interval.
     */
    public long getExecutionInterval();


    /**
     * Retrieves the number of times the operation needs to be executed.
     * 
     * @return The number of times, which the operation needs to be executed.
     */
    public int getNumberOfTimesToExecute();


    /**
     * This indicates that the operation should be allowed to continually run
     * until the operation deems necessary. When the operation is ready to
     * terminate, the operation needs to assert the terminationConditionReached
     * flag. This is particularly useful for operations that need to run until a
     * certain condition has been satisfied.
     * 
     * @return <code>true</code> if this is stay alive; <code>false</code>
     *         otherwise.
     */
    public boolean isStayAlive();


    /**
     * This indicates that the operation should be run when new data is made
     * available on its input datasets. This is especially useful for streaming
     * datasets where data is continually arriving, each of which needs to be
     * processed.
     * 
     * @return <code>true</code> if this is execute when data is available;
     *         <code>false</code> otherwise.
     */
    public boolean isExecuteWhenDataAvailable();
    // public boolean isExecuteUntilCompletionNotification();
}
