/**
 * Software License, Version 1.0
 * 
 * Copyright 2008 The Trustees of Indiana University.  All rights reserved.
 * 
 *
 *Redistribution and use in source and binary forms, with or without 
 *modification, are permitted provided that the following conditions are met:
 *
 *1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 *2) All redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 *3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Community Grids Lab. For 
 * further information contact the Community Grids Lab at 
 * http://communitygrids.iu.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 *4) The name Indiana University or Community Grids Lab or Granules, 
 * shall not be used to endorse or promote products derived from this software 
 * without prior written permission from Indiana University.  For written 
 * permission, please contact the Advanced Research and Technology Institute 
 * ("ARTI") at 351 West 10th Street, Indianapolis, Indiana 46202.
 *5) Products derived from this software may not be called Granules, 
 * nor may Indiana University or Community Grids Lab or Granules appear
 * in their name, without prior written permission of ARTI.
 * 
 *
 * Indiana University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Indiana University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.operation;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * Implements the ExecutionProfile. Includes support for configuring the various
 * fields in the execution profile.
 * 
 * @author Shrideep Pallickara
 * 
 */
public class ExecutionProfileSetup implements ExecutionProfile {
    private String operationBinding;
    private long executionInterval = 0;
    private int numberOfTimesToExecute = 0;
    private boolean executeOnce = true;
    private boolean executePeriodically = false;
    private boolean executeAFixedNumberOfTimes = false;
    private boolean executeWhenDataAvailable = false;
    private boolean stayAlive = false;


    public ExecutionProfileSetup(String operationBinding) {
        this.operationBinding = operationBinding;
    }


    public ExecutionProfileSetup(byte[] marshalledBytes) throws IOException {
        ByteArrayInputStream baInputStream =
            new ByteArrayInputStream(marshalledBytes);
        DataInputStream din = new DataInputStream(baInputStream);

        int bindingLength = din.readInt();
        byte[] operationBindingBytes = new byte[bindingLength];

        din.readFully(operationBindingBytes);
        operationBinding = new String(operationBindingBytes);

        executeOnce = din.readBoolean();
        stayAlive = din.readBoolean();
        executeWhenDataAvailable = din.readBoolean();

        executeAFixedNumberOfTimes = din.readBoolean();
        if (executeAFixedNumberOfTimes) {
            numberOfTimesToExecute = din.readInt();
        }

        executePeriodically = din.readBoolean();
        if (executePeriodically) {
            executionInterval = din.readLong();
        }
    }


    public byte[] getBytes() throws IOException {
        byte[] marshalledBytes = null;
        ByteArrayOutputStream baOutputStream = new ByteArrayOutputStream();
        DataOutputStream dout = new DataOutputStream(baOutputStream);

        int bindingLength = operationBinding.length();
        dout.writeInt(bindingLength);
        dout.write(operationBinding.getBytes());

        dout.writeBoolean(executeOnce);
        dout.writeBoolean(stayAlive);
        dout.writeBoolean(executeWhenDataAvailable);

        dout.writeBoolean(executeAFixedNumberOfTimes);
        if (executeAFixedNumberOfTimes) {
            dout.writeInt(numberOfTimesToExecute);
        }

        dout.writeBoolean(executePeriodically);
        if (executePeriodically) {
            dout.writeLong(executionInterval);
        }

        dout.flush();
        marshalledBytes = baOutputStream.toByteArray();
        return marshalledBytes;
    }


    public void setAsExecuteOnce(String binding) {
        if (!passesBindingCheck(binding)) {
            return;
        }
        changeToExecuteOnce();
    }


    public void changeToExecuteOnce() {
        executeOnce = true;
        executePeriodically = false;
        executeAFixedNumberOfTimes = true;
        numberOfTimesToExecute = 0;
        executionInterval = 0;
        stayAlive = false;
    }


    public void setAsExecutePeriodically(String binding, long executionInterval) {
        if (!passesBindingCheck(binding)) {
            return;
        }
        changeToExecutePeriodically(executionInterval);
    }


    public void changeToExecutePeriodically(long executionInterval) {
        executeOnce = false;
        executePeriodically = true;
        this.executionInterval = executionInterval;
    }


    public void setAsExecuteFixedNumberOfTimes(String binding, int numOfTimes) {
        if (!passesBindingCheck(binding)) {
            return;
        }
        changeToExecuteFixedNumberOfTimes(numOfTimes);
    }


    public void changeToExecuteFixedNumberOfTimes(int numOfTimes) {
        executeOnce = false;
        executeAFixedNumberOfTimes = true;
        numberOfTimesToExecute = numOfTimes;
    }


    public void setAsStayAlive(String binding) {
        if (!passesBindingCheck(binding)) {
            return;
        }
        changeToStayAlive();
    }


    public void changeToStayAlive() {
        executeOnce = false;
        executeAFixedNumberOfTimes = false;
    }


    public void setAsExecuteWhenDataAvailable(String binding) {

        if (!passesBindingCheck(binding)) {
            return;
        }
        changeToExecuteWhenDataAvailable();
    }


    public void changeToExecuteWhenDataAvailable() {
        executeWhenDataAvailable = true;
        stayAlive = true;
        executeOnce = false;
        executeAFixedNumberOfTimes = false;
        executePeriodically = false;
        executionInterval = 0;
    }


    @Override
    public long getExecutionInterval() {
        return executionInterval;
    }


    @Override
    public int getNumberOfTimesToExecute() {
        return numberOfTimesToExecute;
    }


    @Override
    public boolean isExcuteAFixedNumberOfTimes() {
        return executeAFixedNumberOfTimes;
    }


    @Override
    public boolean isExecuteOnce() {
        return executeOnce;
    }


    @Override
    public boolean isExecutePeriodically() {
        return executePeriodically;
    }


    @Override
    public boolean isExecuteWhenDataAvailable() {
        return executeWhenDataAvailable;
    }


    @Override
    public boolean isStayAlive() {
        return stayAlive;
    }


    @Override
    public String toString() {
        String stringRepresentation =
            "executeOnce=" + executeOnce + ", stayAlive=" + stayAlive
                + ", executeWhenDataAvailable= " + executeWhenDataAvailable
                + ", executePeriodically=" + executePeriodically
                + ", executionInterval=" + executionInterval
                + ", executeFixedNumberOfTimes=" + executeAFixedNumberOfTimes
                + ", numberOfTimesToExecute=" + numberOfTimesToExecute;
        return stringRepresentation;
    }


    /**
     * Check to see if the method passes a binding check
     * 
     * @param binding
     * @return
     */
    private boolean passesBindingCheck(String binding) {
        boolean passesCheck = true;
        if (binding != operationBinding) {
            System.out.println("Mismatch in ExecutionProfile binding");
            passesCheck = false;
        }

        return passesCheck;
    }
}
