/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.samples.bridge;

/***
 * This is the mapper for the fibonacci python application.
 * It can use TCP and unnamed pipes to bridge communications.
 * This can be modified in <code> JavaGenericComputation </code>
 * 
 * @author K Ericson
 */

import java.io.IOException;
import java.util.Properties;

import ds.granules.bridge.JavaByteMessage;
import ds.granules.bridge.ControlMessage;
import ds.granules.bridge.JavaGenericComputation;
import ds.granules.bridge.JavaMessageHandler;
import cgl.narada.event.NBEvent;
import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProcessingException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;

/**
 * Simple example of using python through Granules Bridges This example is
 * designed to handle a simple Fibonacci script. The JavaMessageHandler code may
 * be modified to handle other functions simply
 * 
 * @author ericson
 * 
 */
public class FibMap extends MapReduceBase implements JavaMessageHandler {

    JavaGenericComputation jc;
    String executeCommand;
    int executionType;


    public FibMap() throws IOException {
        setDomain("Generic");
        setVersionInformation(200);
        setOperationIdentifier("ABCDEFGHIJKLMN");
        jc = null;
    }


    /**
     * Simple function to actually initialize the program
     */
    public synchronized void fibInit() {
        System.out.println("Initializing");
        Properties props = getProcessingDirectives();
        if (!props.containsKey("Program")) {
            System.out.println("Can't get program!");
            System.exit(-1);
        }
        if (!props.containsKey("Execute Command")) {
            System.out.println("Can't get execute command!");
            System.exit(-1);
        }
        if (!props.containsKey("Communication Type")) {
            System.out.println("Can't get communication type!");
            System.exit(-1);
        }
        String pyProg = props.getProperty("Program");
        executeCommand = props.getProperty("Execute Command");
        executionType =
            Integer.parseInt(props.getProperty("Communication Type"));
        try {
            jc = new JavaGenericComputation(pyProg);
            jc.startExecution(executeCommand, executionType, this);

            JavaByteMessage jm = new JavaByteMessage(ControlMessage.INITIALIZE);

            jm.setStreamIdentifier(this.getOperationIdentifier()); // identifying
                                                                   // ID
            jm.setDescription("initializing"); // useful for debugging -
                                               // description of the purpose of
                                               // this message

            jc.sendMessage(jm);
        } catch (IOException e) {
            System.out.println("Could not start execution");
            e.printStackTrace();
        }
    }


    @Override
    public void execute() throws ProcessingException {
        if (jc == null) {
            System.out.println("initializing...");
            fibInit();
        }

        String datasetIdentifier = "FibStreamDatasetID";

        System.out.println("Executing...");
        DatasetCollection datasetCollection = getDatasetCollection();
        StreamingAccess access = null;
        try {
            Dataset dataset = datasetCollection.getDataset(datasetIdentifier);

            if (dataset.getDatasetType() == Dataset.STREAMS) {
                access = (StreamingAccess) dataset;
            } else {
                System.out.println("Incorrect datatset: "
                    + dataset.getDatasetType() + " initialized. Returning ...");
                return;
            }

            if (access != null) {

                while (dataset.isDataAvailable()) {
                    NBEvent nbEvent = access.getStreamingData();
                    processStreamPacket(nbEvent);
                }

            } else {
                System.out.println("\nThe INPUT Dataset is NULL!\n");
            }

        } catch (DatasetException e) {
            e.printStackTrace();

        } catch (MapReduceException e) {
            System.out.println("Problems writing Results: " + e.getMessage());
            e.printStackTrace();
        } catch (MarshallingException e) {
            System.out.println("Problems marshalling input: " + e.getMessage());
            e.printStackTrace();
        }
    }


    private void processStreamPacket(NBEvent nbEvent)
        throws MapReduceException, MarshallingException {
        // first check to see if the program hasn't been initialized yet:
        ResultPayload resultPayload;
        resultPayload = ResultPayload.getInstance();
        if (jc == null) {
            fibInit();
        }

        System.out.println("Processing stream fragment: " + nbEvent);
        if (nbEvent == null) {
            return;
        }
        System.out.println("Received Stream Packet-> "
            + nbEvent.getContentSynopsis());
        byte[] payload = nbEvent.getContentPayload();

        // assuming very simple communication at the moment - simply a long is
        // expected.
        // a positive number means that fibonacci will be run on it, while a
        // negative number means the program should exit

        long toFib = resultPayload.getLong(payload);

        if (toFib >= 0) {
            System.out.println("Long to Fib: " + toFib);
            // send out a new message here
            JavaByteMessage j = new JavaByteMessage(ControlMessage.START_STOP);
            j.setStreamIdentifier(this.getOperationIdentifier());
            j.setDescription("beginning a run");
            byte[] varType = resultPayload.getBytes(ControlMessage.LONG);
            byte[] lBytes = resultPayload.getBytes(toFib);
            byte[] input = new byte[varType.length + lBytes.length];
            System.arraycopy(varType, 0, input, 0, varType.length);
            System.arraycopy(lBytes, 0, input, varType.length, lBytes.length);

            j.setInput(input);
            System.out.println("Message:\n" + j.toString());
            jc.sendMessage(j);
        } else {
            System.out.println("quitting...");
            // send a quit message
            JavaByteMessage j = new JavaByteMessage(ControlMessage.START_STOP);
            j.setStreamIdentifier(this.getOperationIdentifier());
            j.setDescription("stopping");
            jc.sendMessage(j);

        }

    }


    @Override
    public void handleMessage(byte[] msg) {
        // fudging things to make asynchronous
        final byte[] par = msg;
        new Thread(new Runnable() {
            public void run() {
                messageSwitch(par);
            }
        }).start();
    }


    private void messageSwitch(byte[] msg) {
        ResultPayload resultPayload;
        resultPayload = ResultPayload.getInstance();
        JavaByteMessage jm = new JavaByteMessage(msg);
        int type = jm.getCONTROL();
        switch (type) {
        case ControlMessage.RESULTS:

            byte[] res = jm.getOutput();
            // expecting a long, so ignoring first 4 bytes
            byte[] longbyte = new byte[8];
            byte[] tpe = new byte[4];
            System.arraycopy(res, 0, tpe, 0, 4);
            System.arraycopy(res, 4, longbyte, 0, 8);
            try {
                long result = resultPayload.getLong(longbyte);
                Results results = createResults(true, true);
                results.setResultPayload(resultPayload.getBytes(result));
                String resultsTopic = "WorkerStream/Result";
                writeResults(resultsTopic, results);
            } catch (MarshallingException e) {
                System.out.println("Error getting result: " + e.getMessage());
                e.printStackTrace();
            } catch (MapReduceException e) {
                System.out.println("Error with MapReduce: " + e.getMessage());
                e.printStackTrace();
            }
            break;
        case ControlMessage.STATUS:
            System.out.println("Program Status:");
            byte[] r2 = jm.getOutput();
            // first 4 bytes are type - expecting STRING (-2), so ignoring
            byte[] lengthbyte = new byte[4];
            System.arraycopy(r2, 4, lengthbyte, 0, lengthbyte.length);
            int size;
            try {
                size = resultPayload.getInt(lengthbyte);
                System.out.println("size: " + size);
                byte[] stringbyte = new byte[size];
                System.arraycopy(r2, 8, stringbyte, 0, size);
                String stat = new String(stringbyte);
                System.out.println(stat);
            } catch (MarshallingException e) {
                e.printStackTrace();
            }
            break;
        default:
            // otherwise, unhandled message type was sent
            System.out.println("Unhandled message type: " + type);
            break;
        }
    }
}
