package ds.granules.samples.checksum;

import java.io.IOException;
import java.util.Properties;

import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.DatasetFactory;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.CommunicationsException;
import ds.granules.exception.DeploymentException;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.InstanceDeployer;
import ds.granules.operation.ProgressTracker;

public class ChecksumDeployer extends InstanceDeployer {

    private ProgressTracker progressTracker;

    public ChecksumDeployer(Properties streamingProperties)
    throws CommunicationsException, IOException, DatasetException,
            MarshallingException, DeploymentException {

        initialize(streamingProperties);
    }

    /**
     * This method deploys ChecksumMap instances over the set of
     * available resources.
     * 
     * @param numOfOperations
     *            The total number of instances that need to be initialized.
     */
    public void prepareInstancesAndDeploy(int numOfOperations)
            throws DatasetException, CommunicationsException,
            DeploymentException, MarshallingException {

        ChecksumMap[] operations = new ChecksumMap[numOfOperations];
        for (int i = 0; i < numOfOperations; i++) {
            ChecksumMap map = new ChecksumMap();

            DatasetFactory datasetFactory = DatasetFactory.getInstance();
            StreamingAccess streamingAccess = datasetFactory
                .createStreamingDataset("ChecksumDataset", "Checksum Data");

            String streamSynopsis = "WorkerStreams/Checksum";
            streamingAccess.addInputStream(streamSynopsis,
                    StreamingAccess.STRING_SYNOPSIS);

            DatasetCollection datasetCollection = map.getDatasetCollection();
            datasetCollection.addDataset(streamingAccess);
            map.setDatasetCollection(datasetCollection);
            map.setAsExecuteWhenDataAvailable();

            operations[i] = map;
        }

        progressTracker = deployOperations(operations);
    }

    /**
     * This method attempts to terminate the job that was previously submitted.
     */
    public void terminateJob() throws DeploymentException,
            CommunicationsException, MarshallingException {
        String jobIdentifier = progressTracker.getIdentifier();
        killJob(jobIdentifier);
    }

    public static void main(String[] args)
    throws Exception {
        if (args.length < 3) {
            String name = "ds.granules.samples.checksum.ChecksumDeployer";
            String usage = "Usage: " + name + " <hostname> <port> <instances>";

            System.out.println(usage);
            System.exit(1);
        }

        String hostname = args[0];
        String portnum = args[1];
        int numOfMaps = Integer.parseInt(args[2]);

        Properties streamingProperties = new Properties();
        streamingProperties.put("hostname", hostname);
        streamingProperties.put("portnum", portnum);

        ChecksumDeployer deployer = new ChecksumDeployer(streamingProperties);

        System.out.println("Deploying " + numOfMaps + " instances.");
        deployer.prepareInstancesAndDeploy(numOfMaps);

    }
}
