/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.samples.flip;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Properties;
import java.util.StringTokenizer;

import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.DatasetFactory;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.CommunicationsException;
import ds.granules.exception.DeploymentException;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.InstanceDeployer;
import ds.granules.operation.ProgressTracker;

public class FlipDeployer extends InstanceDeployer {
	private ProgressTracker progressTracker;

	/**
	 * Initializes the deployer.
	 * 
	 * @param streamingProperties
	 *            The properties for streaming
	 * @throws CommunicationsException
	 *             If there are network problems during communications
	 * @throws IOException
	 *             If there are IO problems
	 * @throws MarshallingException
	 *             If there are marshalling errors
	 * @throws DeploymentException
	 *             If there are deployment related problems
	 */
	public FlipDeployer(Properties streamingProperties)
			throws CommunicationsException, IOException, MarshallingException,
			DeploymentException {
		initialize(streamingProperties);
	}

	/**
	 * Initialize a specific <code>FlipMap</code> instance.
	 * 
	 * @param postFix
	 *            The instance ID for the instance in question.
	 * @return The initialized <code>FlipMap</code> instance.
	 * @throws DatasetException
	 *             If there are problems initializing the dataset.
	 */
	public FlipMap initializeFlipMap(int postFix) throws DatasetException {
		FlipMap operation = new FlipMap();
		Properties props = new Properties();
		String value = "(" + postFix + ")";
		String property = "Worker Number";
		props.put(property, value);
		operation.setProcessingDirectives(props);

		DatasetFactory datasetFactory = DatasetFactory.getInstance();
		StreamingAccess streamingAccess = datasetFactory
				.createStreamingDataset("FlipStreamDatasetID", "Testcase");

		String streamSynopsis = "WorkerStreams/Flip";

		streamingAccess.addInputStream(streamSynopsis,
				StreamingAccess.STRING_SYNOPSIS);

		DatasetCollection datasetCollection = operation.getDatasetCollection();

		datasetCollection.addDataset(streamingAccess);
		operation.setDatasetCollection(datasetCollection);
		operation.setAsExecuteWhenDataAvailable();

		return operation;
	}

	/**
	 * This method deploys <code>FlipMap</code> instances over the set of
	 * available resources.
	 * 
	 * @param numOfOperations
	 *            The total number of instances that need to be initialized.
	 * @throws DatasetException
	 *             If there are problems initializing the datasets
	 * @throws CommunicationsException
	 *             If there are network problems during communications
	 * @throws DeploymentException
	 *             If there are deployment related problems
	 * @throws MarshallingException
	 *             If there are marshalling problems.
	 */
	public void prepareInstancesAndDeploy(int numOfOperations)
			throws DatasetException, CommunicationsException,
			DeploymentException, MarshallingException {
		FlipMap[] operations = new FlipMap[numOfOperations];
		for (int i = 0; i < numOfOperations; i++) {
			operations[i] = initializeFlipMap(i);
		}
		progressTracker = deployOperations(operations);
	}

	/**
	 * This method attempts to terminate the job that was previously submitted.
	 * 
	 * @throws DeploymentException
	 *             If there are deployment related problems
	 * @throws CommunicationsException
	 *             If there are network problems during communications
	 * @throws MarshallingException
	 *             If there are marshalling problems.
	 */
	public void terminateJob() throws DeploymentException,
			CommunicationsException, MarshallingException {
		String jobIdentifier = progressTracker.getIdentifier();
		killJob(jobIdentifier);
	}

	
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		String module = "FlipDeployer.main()->";
		if (args.length < 2) {
			System.out.println("Usage: java ds.granules.samples.flip."
					+ "FlipDeployer <hostname> <port>");
			System.exit(0);
		}
		String hostname = args[0];
		String portnum = args[1];
		int numOfMaps = 2;

		Properties streamingProperties = new Properties();
		streamingProperties.put("hostname", hostname);
		streamingProperties.put("portnum", portnum);

		String helpInfo = "\n\nThis is a simple example that demonstrates processing streams "
				+ "using Granules. The current list of commands include: \n"
				+ "submit <num_of_maps>  \n"
				+ "status        : Lists the status of the all Jobs\n"
				+ "kill          : Kill all tasks within the Job\n"
				+ "h             : List the set of commands \n\n";
		try {
			FlipDeployer deployer = new FlipDeployer(streamingProperties);
			StringTokenizer tok;

			InputStreamReader isr = new InputStreamReader(System.in);
			BufferedReader commandLine = new BufferedReader(isr);

			System.out.println(helpInfo);

			while (true) {
				String s = commandLine.readLine();
				if (s == null) {
					System.out.println(module + "String is null!!!");
					break;
				}

				if (s.startsWith("submit")) {
					System.out.println(module
							+ "Processing the [submit] command");

					tok = new StringTokenizer(s, " ");
					int numOfTokens = tok.countTokens();
					tok.nextToken();/* skip the command */

					if (numOfTokens > 1) {
						numOfMaps = Integer.parseInt(tok.nextToken());
					}

					deployer.prepareInstancesAndDeploy(numOfMaps);
					continue;
				}

				if (s.startsWith("status")) {
					System.out.println(deployer.getStatus());
					continue;
				}

				if (s.equals("h")) {
					System.out.println(helpInfo);
					continue;
				}

				if (s.equals("q")) {
					System.out.println(module
							+ "Exiting the sample instance deployer");
					break;
				}

				if (s.startsWith("kill")) {
					System.out.println(module
							+ "Attempting to kill the Job");
					deployer.terminateJob();
					break;
				}
				
				if (s.equals("p")) {
					System.out.println(deployer.getStatus());
					continue;
				}

			}

		} catch (Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}

	}

}
