/**
 * Software License, Version 1.0
 * 
 * Copyright 2008 The Trustees of Indiana University.  All rights reserved.
 * 
 *
 *Redistribution and use in source and binary forms, with or without 
 *modification, are permitted provided that the following conditions are met:
 *
 *1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 *2) All redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 *3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Community Grids Lab. For 
 * further information contact the Community Grids Lab at 
 * http://communitygrids.iu.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 *4) The name Indiana University or Community Grids Lab or Granules, 
 * shall not be used to endorse or promote products derived from this software 
 * without prior written permission from Indiana University.  For written 
 * permission, please contact the Advanced Research and Technology Institute 
 * ("ARTI") at 351 West 10th Street, Indianapolis, Indiana 46202.
 *5) Products derived from this software may not be called Granules, 
 * nor may Indiana University or Community Grids Lab or Granules appear
 * in their name, without prior written permission of ARTI.
 * 
 *
 * Indiana University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Indiana University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.samples.iterative;

import java.io.IOException;

import cgl.narada.event.NBEvent;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * An example demonstrating the <code>Map</code> role within a Word Count
 * operation.
 * 
 * @author Shrideep Pallickara
 * 
 */
public class StageOne extends MapReduceBase {
    private ResultPayload resultPayload;
    private int initialValueToBeWritten = 1000;
    private int cummulativeValue = 0;
    private int terminationValue = 1000000;
    private boolean initialized = false;
    private int iteration = 0;


    public StageOne() {
        setDomain("Generic");
        setVersionInformation(200);
        setOperationIdentifier("Iterative-Stage-One");
        resultPayload = ResultPayload.getInstance();
    }


    public void execute() {

        try {
            if (!initialized) {
                writeResult(initialValueToBeWritten, false);
                initialized = true;

                changeToExecuteWhenDataAvailable();
                return;
            }

            StreamingAccess mapReduceDataset = getMapReduceDataset();

            while (mapReduceDataset.isDataAvailable()) {
                NBEvent nbEvent = mapReduceDataset.getStreamingData();
                processResultsFromMap(nbEvent);
            }

        } catch (DatasetException e) {
            e.printStackTrace();
        } catch (IOException e) {
            System.out.println("Problems converting bytes to longs");
            e.printStackTrace();
        } catch (MarshallingException e) {
            System.out.println("Problems Marshalling: " + e.getMessage());
            e.printStackTrace();
        } catch (MapReduceException e) {
            System.out.println("Problems writing Results: " + e.getMessage());
            e.printStackTrace();
        }

    }


    private void writeResult(int value, boolean processingComplete)
        throws MarshallingException, MapReduceException {
        boolean completeResult = true;
        ResultPayload resultPayload = ResultPayload.getInstance();
        byte[] resultPayloadBytes = resultPayload.getBytes(value);

        Results results = createResults(completeResult, processingComplete);
        results.setResultPayload(resultPayloadBytes);
        results.setResultType(ResultPayload.INT);

        writeResults(results);
    }


    private void writeTermination(boolean processingComplete)
        throws MarshallingException, MapReduceException {
        boolean completeResult = true;
        String terminateString = "TERMINATE";
        ResultPayload resultPayload = ResultPayload.getInstance();
        byte[] resultPayloadBytes = resultPayload.getBytes(terminateString);

        Results results = createResults(completeResult, processingComplete);
        results.setResultPayload(resultPayloadBytes);
        results.setResultType(ResultPayload.STRING);

        writeResults(results);
    }


    private synchronized void processResultsFromMap(NBEvent nbEvent)
        throws IOException, MarshallingException, MapReduceException {

        String synopsis = (String) nbEvent.getContentSynopsis();
        byte[] resultBytes = nbEvent.getContentPayload();
        Results results = ResultsFactory.getInstance().getResults(resultBytes);
        byte[] resultPayloadBytes = results.getResultPayload();

        ResultPayload resultPayload = ResultPayload.getInstance();

        if (results.getResultType() != ResultPayload.INT) {
            System.out.println("StageOne: Incorrect ResultType received");
            return;
        }
        /** Processing assumes that the resultType is ResultPayload.INT */

        int value = resultPayload.getInt(resultPayloadBytes);
        cummulativeValue = value;

        setReceivedOutputFrom(results.getInstanceIdentifier());

        System.out.println("StageOne: Received result (" + cummulativeValue
            + ") on [" + synopsis + "] from MAP ("
            + results.getInstanceIdentifier() + ")");

        if (!isAwaitingOutputs()) {
            resetOutputTracker();
            iteration++;
            if (cummulativeValue >= terminationValue) {
                // writeResult(cummulativeValue, true);
                System.out.println("StageOne: Termination condition with the "
                    + " cummulativeValue=" + cummulativeValue + " > "
                    + terminationValue + " reached at iteration (" + iteration
                    + ")");
                setTeminationConditionReached();
                writeTermination(true);
            }

            if (cummulativeValue < terminationValue) {
                writeResult(cummulativeValue, false);
            }
        }

    }

}
