package ds.granules.samples.kmeans;

import java.io.IOException;

import cgl.narada.event.NBEvent;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * Kmeans Clustering - StageOne. Determines whether to perform further
 * iterations by calculating the differance between the current cluster centers
 * and the new cluster centers. More information of the MapReduce algorithm for
 * Kmeans clustering can be found here.
 * http://www.cs.indiana.edu/~jekanaya/cglmr.html#Kmeans_Clustering
 * 
 * @author Jaliya Ekanayake
 * @author Shrideep Pallickara
 * 
 */
public class KMStageOne extends MapReduceBase {

    private ResultPayload resultPayload;

    /**
     * TODO This parameter should not be hard coded like this. Need a mechanism
     * to transfer these parameters initially.
     */
    private String initCCFileName =
        "/globalhome/jaliya/projects/cglmr/kmeans_data/new_data/init_clusters.txt";

    private KMVectorData initCData = null;

    private boolean initialized = false;

    private int iteration = 0;


    public KMStageOne() {
        setDomain("Generic");
        setVersionInformation(200);
        setOperationIdentifier("Kmeans-Stage-One");
        resultPayload = ResultPayload.getInstance();
    }


    public void execute() {

        try {
            if (!initialized) {
                initCData = new KMVectorData(initCCFileName);
                writeResult(ObjectMarshaller.getBytes(initCData), false);
                initialized = true;
                changeToExecuteWhenDataAvailable();
                return;
            }

            StreamingAccess mapReduceDataset = getMapReduceDataset();

            while (mapReduceDataset.isDataAvailable()) {
                NBEvent nbEvent = mapReduceDataset.getStreamingData();
                processResultsFromMap(nbEvent);
            }

        } catch (DatasetException e) {
            e.printStackTrace();
        } catch (IOException e) {
            System.out.println("Problems converting bytes to longs");
            e.printStackTrace();
        } catch (MarshallingException e) {
            System.out.println("Problems Marshalling: " + e.getMessage());
            e.printStackTrace();
        } catch (MapReduceException e) {
            System.out.println("Problems writing Results: " + e.getMessage());
            e.printStackTrace();
        }

    }


    private void writeResult(byte[] resultPayloadBytes,
                             boolean processingComplete)
        throws MarshallingException, MapReduceException {
        boolean completeResult = true;

        Results results = createResults(completeResult, processingComplete);
        results.setResultPayload(resultPayloadBytes);
        results.setResultType(ResultPayload.INT);

        writeResults(results);
    }


    private void writeTermination(boolean processingComplete)
        throws MarshallingException, MapReduceException {
        boolean completeResult = true;
        String terminateString = "TERMINATE";
        ResultPayload resultPayload = ResultPayload.getInstance();
        byte[] resultPayloadBytes = resultPayload.getBytes(terminateString);

        Results results = createResults(completeResult, processingComplete);
        results.setResultPayload(resultPayloadBytes);
        results.setResultType(ResultPayload.STRING);

        writeResults(results);
    }


    private synchronized void processResultsFromMap(NBEvent nbEvent)
        throws IOException, MarshallingException, MapReduceException {

        String synopsis = (String) nbEvent.getContentSynopsis();
        byte[] resultBytes = nbEvent.getContentPayload();
        Results results = ResultsFactory.getInstance().getResults(resultBytes);
        byte[] resultPayloadBytes = results.getResultPayload();

        ResultPayload resultPayload = ResultPayload.getInstance();

        if (results.getResultType() != ResultPayload.INT) {
            System.out.println("StageOne: Incorrect ResultType received");
            return;
        }
        /** Processing assumes that the resultType is ResultPayload.INT */

        /**
         * We should only get one value here. This is the final combine stage.
         */
        KMVectorData newCentroids =
            (KMVectorData) ObjectMarshaller.fromBytes(resultPayloadBytes);

        setReceivedOutputFrom(results.getInstanceIdentifier());
        System.out.println("StageOne: Received result on [" + synopsis
            + "] from MAP (" + results.getInstanceIdentifier() + ")");

        double totalError = getError(initCData, newCentroids);

        if (!isAwaitingOutputs()) {
            resetOutputTracker();
            iteration++;
            if (totalError < 1) {
                System.out.println("KMStageOne: Termination condition with the   reached at iteration ("
                    + iteration + ")");
                setTeminationConditionReached();
                writeTermination(true);
            } else {

                writeResult(ObjectMarshaller.getBytes(newCentroids), false);
                this.initCData = newCentroids;
            }
        }
    }


    /**
     * Calculates the differance between the current cluster centers and the new
     * cluster centers.
     * 
     * @param cData
     *            Current cluster centers.
     * @param newCData
     *            New cluster centers.
     * @return differance or the current error.
     */
    private double getError(KMVectorData cData, KMVectorData newCData) {
        double totalError = 0;
        int numCentroids = cData.getNumVectors();

        double[][] centroids = cData.getData();
        double[][] newCentroids = newCData.getData();

        for (int i = 0; i < numCentroids; i++) {
            totalError +=
                getEuclidean(centroids[i], newCentroids[i], cData.getVecLen());
        }
        return totalError;
    }


    /**
     * Returns the euclidean distance (square value) of given two vectors.
     * 
     * @param v1
     *            - Vector one
     * @param v2
     *            - Vectpr tw0
     * @param vecLen
     *            - Length of the vectors.
     * @return - Square value of the euclidean.
     */
    public double getEuclidean(double[] v1, double[] v2, int vecLen) {
        double sum = 0;
        for (int i = 0; i < vecLen; i++) {
            sum += ((v1[i] - v2[i]) * (v1[i] - v2[i]));
        }
        return sum;
    }

}
