package ds.granules.samples.kmeans;

import java.io.IOException;

import cgl.narada.event.NBEvent;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * Kmeans Clustering - StageThree. Performs a reduction operation on passed
 * partial clustre centers. More information of the MapReduce algorithm for
 * Kmeans clustering can be found here.
 * http://www.cs.indiana.edu/~jekanaya/cglmr.html#Kmeans_Clustering
 * 
 * @author Jaliya Ekanayake
 * @author Shrideep Pallickara
 * 
 */
public class KMStageThree extends MapReduceBase {

    private ResultPayload resultPayload;

    String workerType = "Unknown ";

    private double[][] newCentroidsData = null;

    private double[] counts = null;


    public KMStageThree() {
        setDomain("Generic");
        setVersionInformation(200);
        setOperationIdentifier("Kmeans-Stage-Three");
        resultPayload = ResultPayload.getInstance();
    }


    public void execute() {
        try {
            StreamingAccess mapReduceDataset = getMapReduceDataset();

            while (mapReduceDataset.isDataAvailable()) {
                NBEvent nbEvent = mapReduceDataset.getStreamingData();
                processResultsFromMap(nbEvent);
            }

        } catch (DatasetException e) {
            e.printStackTrace();
        } catch (IOException e) {
            System.out.println("Problems converting bytes to longs");
            e.printStackTrace();
        } catch (MarshallingException e) {
            System.out.println("Problems Marshalling: " + e.getMessage());
            e.printStackTrace();
        } catch (MapReduceException e) {
            System.out.println("Problems writing Results: " + e.getMessage());
            e.printStackTrace();
        }

    }


    private synchronized void processResultsFromMap(NBEvent nbEvent)
        throws IOException, MarshallingException, MapReduceException {

        String synopsis = (String) nbEvent.getContentSynopsis();
        byte[] resultBytes = nbEvent.getContentPayload();
        Results results = ResultsFactory.getInstance().getResults(resultBytes);
        byte[] resultPayloadBytes = results.getResultPayload();

        ResultPayload resultPayload = ResultPayload.getInstance();

        if (results.getResultType() == ResultPayload.STRING) {
            String receivedString = resultPayload.getString(resultPayloadBytes);
            if (receivedString.equals("TERMINATE")) {
                setTeminationConditionReached();
                System.out.println("StageThree: Terminate command from "
                    + "one of the StageTwos processed");
            }
            return;
        }

        if (results.getResultType() != ResultPayload.INT) {
            System.out.println("StageOne: Incorrect ResultType received");
            return;
        }

        /**
         * Remainder of processing assumes that the resultPayload type is of
         * type ResultPayload.INT but actually the data is in composite form.
         */

        KMVectorData cData =
            (KMVectorData) ObjectMarshaller.fromBytes(resultPayloadBytes);
        int numDataPoints = cData.getNumVectors();
        /**
         * One additional location carries the count.
         */
        int vecLen = cData.getVecLen() - 1;

        if (newCentroidsData == null) {
            // First time. Initialize the memeber variables.
            this.newCentroidsData = new double[numDataPoints][vecLen];
            this.counts = new double[numDataPoints];
        }

        double[][] currentCentroids = cData.getData();

        for (int j = 0; j < numDataPoints; j++) {
            for (int k = 0; k < vecLen; k++) {
                newCentroidsData[j][k] += currentCentroids[j][k];
            }
            counts[j] += currentCentroids[j][vecLen];
            // small hack. say data length is 2, then the tmpCentroid has 3 data
            // points including the counts. from points 0,1,2 , 2 is the count.
        }

        setReceivedOutputFrom(results.getInstanceIdentifier());
        System.out.println("StageThree: Received results  on [" + synopsis
            + "] from MAP (" + results.getInstanceIdentifier() + ")");

        if (!isAwaitingOutputs()) {
            resetOutputTracker();

            for (int i = 0; i < numDataPoints; i++) {
                for (int j = 0; j < vecLen; j++) {
                    if (counts[i] != 0) {
                        newCentroidsData[i][j] =
                            (newCentroidsData[i][j]) / counts[i];
                    }
                }
            }
            KMVectorData newCData =
                new KMVectorData(numDataPoints, vecLen, newCentroidsData);

            writeResult(ObjectMarshaller.getBytes(newCData), false);
            // setTeminationConditionReached();

            // Cleanup
            this.newCentroidsData = null;
            this.counts = null;
        }

    }


    private void writeResult(byte[] resultPayloadBytes,
                             boolean processingComplete)
        throws MarshallingException, MapReduceException {
        boolean completeResult = true;

        Results results = createResults(completeResult, processingComplete);
        results.setResultPayload(resultPayloadBytes);
        results.setResultType(ResultPayload.INT);

        writeResults(results);
    }

}
