package ds.granules.samples.kmeans;

import java.io.IOException;
import java.util.Enumeration;

import cgl.narada.event.NBEvent;
import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.FileAccess;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * Kmeans Clustering - StageTwo. This is the Map task for the Kmeans clustering
 * application. The operation, processResultsFromMap execute the following
 * algorithm.
 * 
 * 1. Load the vector data from the data split assigned to this task. 2. For
 * each vector, calculate the distance between it and the cluster centers 3.
 * Assign the vector to a cluster center with a minimum distance 4. Increment
 * the counter for each cluster center as vectors are added to them 5. Pass the
 * new cluster centers and the respective counts to the reduce stage
 * (StageThree)
 * 
 * More information of the MapReduce algorithm for Kmeans clustering can be
 * found here. http://www.cs.indiana.edu/~jekanaya/cglmr.html#Kmeans_Clustering
 * 
 * @author Jaliya Ekanayake
 * @author Shrideep Pallickara
 * 
 */
public class KMStageTwo extends MapReduceBase {
    private ResultPayload resultPayload;

    private String datasetIdentifier = "FileDataset-ID-KmeansStageTwo";

    private DatasetCollection datasetCollection;

    private FileAccess access = null;

    private boolean initialized = false;

    private KMVectorData vData = null;


    public KMStageTwo() {
        setDomain("Generic");
        setVersionInformation(200);
        setOperationIdentifier("Iterative-Stage-Two");
        resultPayload = ResultPayload.getInstance();
    }


    public void execute() {

        try {
            if (!initialized) {
                initializeAccessToFiles();
                initialized = true;
            }

            StreamingAccess mapReduceDataset = getMapReduceDataset();

            while (mapReduceDataset.isDataAvailable()) {
                NBEvent nbEvent = mapReduceDataset.getStreamingData();
                processResultsFromMap(nbEvent);
            }

        } catch (DatasetException e) {
            e.printStackTrace();
        } catch (IOException e) {
            System.out.println("Problems converting bytes to longs");
            e.printStackTrace();
        } catch (MarshallingException e) {
            System.out.println("Problems Marshalling: " + e.getMessage());
            e.printStackTrace();
        } catch (MapReduceException e) {
            System.out.println("Problems writing Results: " + e.getMessage());
            e.printStackTrace();
        }

    }


    private void initializeAccessToFiles() throws DatasetException, IOException {
        datasetCollection = getDatasetCollection();

        if (!datasetCollection.hasDataset(datasetIdentifier)) {
            System.out.println("Unable to locate dataset: " + datasetIdentifier);
            return;
        }
        Dataset dataset = datasetCollection.getDataset(datasetIdentifier);

        if (dataset.getDatasetType() == Dataset.FILES) {
            access = (FileAccess) dataset;
        } else {
            System.out.println("Incorrect datatset: "
                + dataset.getDatasetType() + " initialized. Returning ...");
            return;
        }

        if (access == null) {
            System.out.println("Unable to initialize the dataset");
            return;
        }

        Enumeration filenames = access.getFilenames();
        int count = 0;
        while (filenames.hasMoreElements()) {
            String filename = (String) filenames.nextElement();
            vData = new KMVectorData(filename);
            if (count > 0) {
                throw new DatasetException(
                                           "KMStageTwo accpet only a single file per task.");
            }
            count++;

        }
    }


    private synchronized void processResultsFromMap(NBEvent nbEvent)
        throws IOException, MarshallingException, MapReduceException {

        String synopsis = (String) nbEvent.getContentSynopsis();
        byte[] resultBytes = nbEvent.getContentPayload();
        Results results = ResultsFactory.getInstance().getResults(resultBytes);
        byte[] resultPayloadBytes = results.getResultPayload();

        ResultPayload resultPayload = ResultPayload.getInstance();

        if (results.getResultType() == ResultPayload.STRING) {
            String receivedString = resultPayload.getString(resultPayloadBytes);
            if (receivedString.equals("TERMINATE")) {
                setTeminationConditionReached();
                writeTermination(true);
                System.out.println("StageTwo: Terminate command from StageOne "
                    + "processed ");
            }
            return;
        }

        if (results.getResultType() != ResultPayload.INT) {
            System.out.println("StageOne: Incorrect ResultType received");
            return;
        }

        /**
         * Remainder of processing assumes that the resultPayload type is of
         * type ResultPayload.INT
         */

        KMVectorData currentCentroids =
            (KMVectorData) ObjectMarshaller.fromBytes(resultPayloadBytes);
        setReceivedOutputFrom(results.getInstanceIdentifier());

        System.out.println("StageTwo: Received results on [" + synopsis
            + "] from MAP (" + results.getInstanceIdentifier() + ")");

        if (!isAwaitingOutputs()) {
            resetOutputTracker();
            int numData = vData.getNumVectors();
            int vecLen = vData.getVecLen();
            int numCentroids = currentCentroids.getNumVectors();
            double[][] centroids = currentCentroids.getData();
            double[][] data = vData.getData();
            double[][] newCentroids = new double[numCentroids][vecLen + 1]; // One
            // additional
            // column
            // carries
            // the
            // counts.

            for (int i = 0; i < numData; i++) {
                double min = 0;
                double dis = 0;
                int minCentroid = 0;
                for (int j = 0; j < numCentroids; j++) {
                    dis = getEuclidean2(data[i], centroids[j], vecLen);
                    if (j == 0) {
                        min = dis;
                    }
                    if (dis < min) {
                        min = dis;
                        minCentroid = j;
                    }
                }
                for (int k = 0; k < vecLen; k++) {
                    newCentroids[minCentroid][k] += data[i][k];
                }
                newCentroids[minCentroid][vecLen] += 1;
            }
            // Note the additional column
            writeResult(
                        ObjectMarshaller.getBytes(new KMVectorData(
                                                                   numCentroids,
                                                                   vecLen + 1,
                                                                   newCentroids)),
                        false);
        }

    }


    /**
     * Returns the euclidean distance (square value) of given two vectors.
     * 
     * @param v1
     *            - Vector one
     * @param v2
     *            - Vectpr tw0
     * @param vecLen
     *            - Length of the vectors.
     * @return - Square value of the euclidean.
     */
    public double getEuclidean2(double[] v1, double[] v2, int vecLen) {
        double sum = 0;
        for (int i = 0; i < vecLen; i++) {
            sum += ((v1[i] - v2[i]) * (v1[i] - v2[i]));
        }
        return sum; // No need to use the sqrt

    }


    private void writeResult(byte[] resultPayloadBytes,
                             boolean processingComplete)
        throws MarshallingException, MapReduceException {
        boolean completeResult = true;

        Results results = createResults(completeResult, processingComplete);
        results.setResultPayload(resultPayloadBytes);
        results.setResultType(ResultPayload.INT);

        writeResults(results);

    }


    private void writeTermination(boolean processingComplete)
        throws MarshallingException, MapReduceException {
        boolean completeResult = true;
        String terminateString = "TERMINATE";
        ResultPayload resultPayload = ResultPayload.getInstance();
        byte[] resultPayloadBytes = resultPayload.getBytes(terminateString);

        Results results = createResults(completeResult, processingComplete);
        results.setResultPayload(resultPayloadBytes);
        results.setResultType(ResultPayload.STRING);

        writeResults(results);
    }
}
