/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.samples.pythonColl;

/***
 * This is a basic deployer for the fibonacci bridging sample code.
 * It was built based off of the Flip example
 * 
 * @author K Ericson
 */

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Properties;
import java.util.StringTokenizer;

import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.DatasetFactory;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.CommunicationsException;
import ds.granules.exception.DeploymentException;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.InstanceDeployer;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProgressTracker;

public class PyCollDeployerRep extends InstanceDeployer {

    private ProgressTracker progressTracker;


    /**
     * Initializes the deployer.
     * 
     * @param streamingProperties
     *            The properties for streaming
     * @throws CommunicationsException
     *             If there are network problems during communications
     * @throws IOException
     *             If there are IO problems
     * @throws MarshallingException
     *             If there are marshalling errors
     * @throws DeploymentException
     *             If there are deployment related problems
     */
    public PyCollDeployerRep(Properties streamingProperties)
        throws CommunicationsException, IOException, MarshallingException,
        DeploymentException {
        initialize(streamingProperties);
    }


    /**
     * Initialize a specific <code>FibMap</code> instance.
     * 
     * @param pcr
     * 
     * @param postFix
     *            The instance ID for the instance in question.
     * @return The initialized <code>FibMap</code> instance.
     * @throws DatasetException
     *             If there are problems initializing the dataset.
     * @throws IOException
     * @throws MapReduceException
     */
    public PyCollMap initializePyCollMap(PyCollReduceRep pcr, int postFix,
                                         String pyProg, int total)
        throws DatasetException, IOException, MapReduceException {
        System.out.println("Starting to initialize!!!");
        PyCollMap operation = new PyCollMap();
        Properties props = new Properties();
        String value = "(" + postFix + ")";
        String property = "Worker Number";
        props.put(property, value);
        props.put("Program", pyProg);
        props.put("Number of Workers", total + "");
        operation.setProcessingDirectives(props);

        DatasetFactory datasetFactory = DatasetFactory.getInstance();
        StreamingAccess streamingAccess =
            datasetFactory.createStreamingDataset("PyCollStreamDatasetID",
                                                  "Testcase");

        String streamSynopsis = "WorkerStreams/PyColl";

        streamingAccess.addInputStream(streamSynopsis,
                                       StreamingAccess.STRING_SYNOPSIS);

        DatasetCollection datasetCollection = operation.getDatasetCollection();

        datasetCollection.addDataset(streamingAccess);
        operation.setDatasetCollection(datasetCollection);
        operation.setAsExecuteWhenDataAvailable();

        operation.addReduce(pcr);

        return operation;
    }


    /**
     * This method deploys <code>FibMap</code> instances over the set of
     * available resources.
     * 
     * @param numOfOperations
     *            The total number of instances that need to be initialized.
     * @param prog
     *            The program to be executed, including the path to reach it.
     * @param fout
     * @param zoom
     * @param height
     * @param width
     * @param fName
     * @param pyProg2
     * @param description
     * @throws DatasetException
     *             If there are problems initializing the datasets
     * @throws CommunicationsException
     *             If there are network problems during communications
     * @throws DeploymentException
     *             If there are deployment related problems
     * @throws MarshallingException
     *             If there are marshalling problems.
     * @throws IOException
     *             If there is a problem connecting to the c program passed in
     *             as a parameter
     * @throws MapReduceException
     */
    public void prepareInstancesAndDeploy(int numOfOperations, String prog,
                                          String pyProg2, String fName,
                                          int width, int height, int zoom,
                                          String fout, String description)
        throws DatasetException, CommunicationsException, DeploymentException,
        MarshallingException, IOException, MapReduceException {
        // only one Reducer, so initialize it first, then attatch it to the maps
        // after it's been initialized
        PyCollReduceRep pcr =
            initializePyCollReduce(pyProg2, numOfOperations, fName, width,
                                   height, zoom, fout, description);
        PyCollMap[] operations = new PyCollMap[numOfOperations];
        for (int i = 0; i < numOfOperations; i++) {
            operations[i] = initializePyCollMap(pcr, i, prog, numOfOperations);
        }

        PyCollReduceRep[] reducers = { pcr };
        progressTracker = deployOperations(reducers);

        String jobId = progressTracker.getIdentifier();
        deployOperations(operations, jobId);
    }


    private PyCollReduceRep initializePyCollReduce(String prog,
                                                   int numOfOperations,
                                                   String inFile, int bwidth,
                                                   int bheight, int zoom,
                                                   String outFile,
                                                   String description)
        throws DatasetException {

        PyCollReduceRep operation = new PyCollReduceRep();
        Properties props = new Properties();

        props.put("Output File", outFile);
        props.put("Zoom", zoom + "");
        props.put("Height", bheight + "");
        props.put("Width", bwidth + "");
        props.put("Input File", inFile);
        props.put("Program", prog);
        props.put("Number of Workers", numOfOperations + "");
        props.put("Description", description);
        operation.setProcessingDirectives(props);

        operation.setAsExecuteWhenDataAvailable();
        DatasetFactory datasetFactory = DatasetFactory.getInstance();

        StreamingAccess streamingAccess =
            datasetFactory.createStreamingDataset("PyCollStreamDatasetID",
                                                  "Testcase");

        String streamSynopsis = "Reducer/Initialize";

        streamingAccess.addInputStream(streamSynopsis,
                                       StreamingAccess.STRING_SYNOPSIS);

        DatasetCollection datasetCollection = operation.getDatasetCollection();

        datasetCollection.addDataset(streamingAccess);
        operation.setDatasetCollection(datasetCollection);
        operation.setAsExecuteWhenDataAvailable();

        return operation;
    }


    /**
     * @param args
     */
    public static void main(String[] args) {
        String module = "PyCollDeployer.main()->";
        if (args.length < 2) {
            System.out.println("Usage: java ds.granules.samples.pythonColl."
                + "PyCollDeployer <hostname> <port>");
            System.exit(0);
        }
        String hostname = args[0];
        String portnum = args[1];
        int numOfMaps = 2;
        String pyProg1 = "/tmp/ericson/python/mainMap.py";
        String pyProg2 = "/tmp/ericson/python/mainMap.py";
        String fName = "/tmp/ericson/python/img.jpg";
        int width = 10;
        int height = 10;
        int zoom = 1;
        String fout = "/tmp/ericson/out.jpg";
        String description = "car";

        Properties streamingProperties = new Properties();
        streamingProperties.put("hostname", hostname);
        streamingProperties.put("portnum", portnum);

        String helpInfo =
            "\n\nThis is a simple example that demonstrates bridging behavior "
                + "via unnamed pipes using Granules. The current list of commands include: \n"
                + "submit <num_of_maps> <python_mapper_main> <python_reducer_main> <input_image_file> <pixel_box_width> <pixel_box_height> <zoom> <output_file> <image_description>\n"
                + "status        : Lists the status of the all Jobs\n"
                + "h             : List the set of commands \n\n";
        try {
            PyCollDeployerRep deployer =
                new PyCollDeployerRep(streamingProperties);
            StringTokenizer tok;

            InputStreamReader isr = new InputStreamReader(System.in);
            BufferedReader commandLine = new BufferedReader(isr);

            System.out.println(helpInfo);

            while (true) {
                String s = commandLine.readLine();
                if (s == null) {
                    System.out.println(module + "String is null!!!");
                    break;
                }

                if (s.startsWith("submit")) {
                    System.out.println(module
                        + "Processing the [submit] command");

                    tok = new StringTokenizer(s, " ");
                    int numOfTokens = tok.countTokens();
                    tok.nextToken();/* skip the command */

                    if (numOfTokens > 1) {
                        numOfMaps = Integer.parseInt(tok.nextToken());
                    }
                    if (numOfTokens > 2) {
                        pyProg1 = tok.nextToken();
                    }
                    if (numOfTokens > 3) {
                        pyProg2 = tok.nextToken();
                    }
                    if (numOfTokens > 4) {
                        fName = tok.nextToken();
                    }
                    if (numOfTokens > 5) {
                        width = Integer.parseInt(tok.nextToken());
                    }
                    if (numOfTokens > 6) {
                        height = Integer.parseInt(tok.nextToken());
                    }
                    if (numOfTokens > 7) {
                        zoom = Integer.parseInt(tok.nextToken());
                    }
                    if (numOfTokens > 8) {
                        fout = tok.nextToken();
                    }
                    if (numOfTokens > 9)
                        description = tok.nextToken();

                    deployer.prepareInstancesAndDeploy(numOfMaps, pyProg1,
                                                       pyProg2, fName, width,
                                                       height, zoom, fout,
                                                       description);

                    continue;
                }

                if (s.startsWith("status")) {
                    System.out.println(deployer.getStatus());
                    continue;
                }

                if (s.equals("h")) {
                    System.out.println(helpInfo);
                    continue;
                }

                if (s.equals("q")) {
                    System.out.println(module
                        + "Exiting the sample instance deployer");
                    break;
                }

                if (s.equals("p")) {
                    System.out.println(deployer.getStatus());
                    continue;
                }

            }

        } catch (Exception e) {
            System.out.println(e.getMessage());
            e.printStackTrace();
        }
        System.exit(0);

    }

}
