/**
 * Software License, Version 1.0
 * 
 * Copyright 2008 The Trustees of Indiana University.  All rights reserved.
 * 
 *
 *Redistribution and use in source and binary forms, with or without 
 *modification, are permitted provided that the following conditions are met:
 *
 *1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 *2) All redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 *3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Community Grids Lab. For 
 * further information contact the Community Grids Lab at 
 * http://communitygrids.iu.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 *4) The name Indiana University or Community Grids Lab or Granules, 
 * shall not be used to endorse or promote products derived from this software 
 * without prior written permission from Indiana University.  For written 
 * permission, please contact the Advanced Research and Technology Institute 
 * ("ARTI") at 351 West 10th Street, Indianapolis, Indiana 46202.
 *5) Products derived from this software may not be called Granules, 
 * nor may Indiana University or Community Grids Lab or Granules appear
 * in their name, without prior written permission of ARTI.
 * 
 *
 * Indiana University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Indiana University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.samples.sort;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Properties;
import java.util.StringTokenizer;

import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.DatasetFactory;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.CommunicationsException;
import ds.granules.exception.DeploymentException;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.InstanceDeployer;
import ds.granules.operation.MapReduceException;
import ds.granules.operation.ProgressTracker;

/**
 * The deployer for the periodicity example.
 * 
 * @author Shrideep Pallickara
 * 
 */
public class SortDeployer extends InstanceDeployer {
    private ProgressTracker progressTracker;


    /**
     * Initializes the deployer.
     * 
     * @param streamingProperties
     *            The properties for streaming
     * @throws CommunicationsException
     *             If there are network problems during communications
     * @throws IOException
     *             If there are IO problems
     * @throws MarshallingException
     *             If there are marshalling erros
     * @throws DeploymentException
     *             If there are deployment related problems
     */
    public SortDeployer(Properties streamingProperties)
        throws CommunicationsException, IOException, MarshallingException,
        DeploymentException {
        initialize(streamingProperties);
    }


    /**
     * Initialize a specific <code>SortingOperation</code> instance.
     * 
     * @param postFix
     *            The instance ID for the instance in question.
     * @return The initialized <code>SortingOperation</code> instance.
     * @throws DatasetException
     *             If there are problems initializing the dataset.
     */
    public SortingOperation initializeSortingOperation(int postFix)
        throws DatasetException {
        SortingOperation operation = new SortingOperation();
        Properties props = new Properties();
        String value = "(" + postFix + ")";
        String property = "Worker Number";
        props.put(property, value);
        operation.setProcessingDirectives(props);

        DatasetFactory datasetFactory = DatasetFactory.getInstance();
        StreamingAccess streamingAccess =
            datasetFactory.createStreamingDataset("DatasetID", "Testcase");

        String streamSynopsis1 = "NumberStreams/Resource" + postFix;
        String streamSynopsis2 = "NumberStreams/Compute";

        streamingAccess.addInputStream(streamSynopsis1,
                                       StreamingAccess.STRING_SYNOPSIS);
        streamingAccess.addInputStream(streamSynopsis2,
                                       StreamingAccess.STRING_SYNOPSIS);

        DatasetCollection datasetCollection =
            datasetFactory.createDatasetCollection();

        datasetCollection.addDataset(streamingAccess);
        operation.setDatasetCollection(datasetCollection);
        operation.setAsExecuteWhenDataAvailable();

        return operation;
    }


    /**
     * This method deploys vanilla <code>SortingOperation</code> instances over
     * the set of available resources. This is different from the other
     * equivalent method in this class which deploys a set of Map/Reduce roles.
     * 
     * @param numOfOperations
     *            The total number of instances that need to be initialized.
     * @throws DatasetException
     *             If there are problems initializing the datasets
     * @throws CommunicationsException
     *             If there are network problems during communications
     * @throws DeploymentException
     *             If there are deployment related problems
     * @throws MarshallingException
     *             If there are marshalling problems.
     */
    public void prepareBaseInstancesAndDeploy(int numOfOperations)
        throws DatasetException, CommunicationsException, DeploymentException,
        MarshallingException {
        SortingOperation[] operations = new SortingOperation[numOfOperations];
        for (int i = 0; i < numOfOperations; i++) {
            operations[i] = initializeSortingOperation(i);
        }
        progressTracker = deployOperations(operations);
    }


    /**
     * Initializes a <code>SortMap</code> instance based on the specified
     * arguments.
     * 
     * @param reduce
     *            The reducer that would be associated with the map.
     * @param postFix
     *            The map instance number
     * @return The initialized <code>SortMap</code> instance
     * @throws DatasetException
     *             If there are problems initializing/using the datasets
     * @throws MapReduceException
     *             If there are problems related to Map/Reduce roles within the
     *             application.
     */
    public SortMap initializeSortMap(SortReduce reduce, int postFix)
        throws DatasetException, MapReduceException {
        SortMap map = new SortMap();

        Properties props = new Properties();
        String value = "(" + postFix + ")";
        String property = "Worker Number";
        props.put(property, value);
        map.setProcessingDirectives(props);

        DatasetFactory datasetFactory = DatasetFactory.getInstance();
        StreamingAccess streamingAccess =
            datasetFactory.createStreamingDataset("NumberStreams-DatasetID",
                                                  "Testcase");

        String streamSynopsis1 = "NumberStreams/Resource" + postFix;
        String streamSynopsis2 = "NumberStreams/Compute";
        streamingAccess.addInputStream(streamSynopsis1,
                                       StreamingAccess.STRING_SYNOPSIS);
        streamingAccess.addInputStream(streamSynopsis2,
                                       StreamingAccess.STRING_SYNOPSIS);

        DatasetCollection datasetCollection = map.getDatasetCollection();
        datasetCollection.addDataset(streamingAccess);

        map.setAsExecuteWhenDataAvailable();
        map.addReduce(reduce);

        return map;
    }


    /**
     * Prepares the set of Map/Reduce instances to be deployed on a set of
     * machines.
     * 
     * @param numOfMaps
     *            The number of Maps that need to be deployed.
     * @throws DatasetException
     *             If there are problems initializing the datasets
     * @throws MapReduceException
     *             If there is a MapReduceExcecution because of problems when
     *             setting up the roles.
     * @throws IOException
     *             If there are IO problems.
     * @throws CommunicationsException
     *             If there are network problems during communications
     * @throws DeploymentException
     *             If there are deployment related problems
     * @throws MarshallingException
     *             If there are marshalling problems.
     * @throws InterruptedException
     *             If there are problems sleeping
     */
    public void prepareInstancesAndDeploy(int numOfMaps)
        throws MapReduceException, CommunicationsException,
        DeploymentException, MarshallingException, DatasetException,
        InterruptedException {

        SortReduce reduce = new SortReduce();
        reduce.setAsExecuteWhenDataAvailable();

        SortMap[] maps = new SortMap[numOfMaps];
        for (int i = 0; i < numOfMaps; i++) {
            maps[i] = initializeSortMap(reduce, i);
        }

        SortReduce[] reducers = { reduce };
        progressTracker = deployOperations(reducers);

        String jobId = progressTracker.getIdentifier();
        deployOperations(maps, jobId);
    }


    /**
     * @param args
     */
    public static void main(String[] args) {
        String module = "SortDeployer.main()->";
        if (args.length < 2) {
            System.out.println("Usage: java ds.granules.samples.sort."
                + "SortDeployer <hostname> <port>");
            System.exit(0);
        }
        String hostname = args[0];
        String portnum = args[1];
        int numOfMaps = 2;

        Properties streamingProperties = new Properties();
        streamingProperties.put("hostname", hostname);
        streamingProperties.put("portnum", portnum);

        String helpInfo =
            "\n\nThis is example demonstrates Parallel-sort functionality "
                + "using Granules. The current list of commands include: \n"
                + "submit <num_of_maps>  \n"
                + "status        : Lists the status of the all Jobs"
                + "h             : List the set of commands \n\n";
        try {
            SortDeployer deployer = new SortDeployer(streamingProperties);
            StringTokenizer tok;

            InputStreamReader isr = new InputStreamReader(System.in);
            BufferedReader commandLine = new BufferedReader(isr);

            System.out.println(helpInfo);

            while (true) {
                String s = commandLine.readLine();
                if (s == null) {
                    System.out.println(module + "String is null!!!");
                    break;
                }

                if (s.startsWith("submit")) {
                    System.out.println(module
                        + "Processing the [submit] command");

                    tok = new StringTokenizer(s, " ");
                    int numOfTokens = tok.countTokens();
                    tok.nextToken();/* skip the command */

                    if (numOfTokens > 1) {
                        numOfMaps = Integer.parseInt(tok.nextToken());
                    }

                    deployer.prepareInstancesAndDeploy(numOfMaps);
                    continue;
                }

                if (s.startsWith("status")) {
                    System.out.println(deployer.getStatus());
                    continue;
                }

                if (s.equals("h")) {
                    System.out.println(helpInfo);
                    continue;
                }

                if (s.equals("q")) {
                    System.out.println(module
                        + "Exiting the sample instance deployer");
                    break;
                }

                if (s.equals("p")) {
                    System.out.println(deployer.getStatus());
                    continue;
                }

            }

        } catch (Exception e) {
            System.out.println(e.getMessage());
            e.printStackTrace();
        }
    }
}
