/**
 * Software License, Version 1.0
 * 
 * Copyright 2008 The Trustees of Indiana University.  All rights reserved.
 * 
 *
 *Redistribution and use in source and binary forms, with or without 
 *modification, are permitted provided that the following conditions are met:
 *
 *1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 *2) All redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 *3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Community Grids Lab. For 
 * further information contact the Community Grids Lab at 
 * http://communitygrids.iu.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 *4) The name Indiana University or Community Grids Lab or Granules, 
 * shall not be used to endorse or promote products derived from this software 
 * without prior written permission from Indiana University.  For written 
 * permission, please contact the Advanced Research and Technology Institute 
 * ("ARTI") at 351 West 10th Street, Indianapolis, Indiana 46202.
 *5) Products derived from this software may not be called Granules, 
 * nor may Indiana University or Community Grids Lab or Granules appear
 * in their name, without prior written permission of ARTI.
 * 
 *
 * Indiana University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Indiana University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.scheduler;

import java.io.File;
import java.util.Properties;
import java.util.Random;

import cgl.narada.util.UUIDRetriever;
import ds.granules.exception.CommunicationsException;
import ds.granules.util.ParamsReader;
import ds.granules.util.Version;

/**
 * This is a class which encapsulates the functionality of a computational
 * resource. In most cases, one instance of this class needs to run every
 * machine within the computational cloud.
 *
 * The Resource is implemented as a singleton; only one instance should exist
 * in a given Java VM, but may provide multiple computational threads.
 * 
 * @author Shrideep Pallickara
 * 
 */
public final class Resource {

    private static Resource instance = null;

    private ExchangeProcessor exchangeProcessor;
    private Diagnostics diagnostics;
    private StreamingService streamingService;
    private UUIDRetriever retriever;

    /**
     * Retrieves singleton instance of the Resource.
     *
     * @throws NullPointerException if the Resource has not be initialized
     *
     * @return Reference to the singleton Resource instance.
     */
    public static Resource getInstance() {
        if (instance == null) {
            throw new NullPointerException(
                    "Resource singleton has not been initialized.");
        }

        return instance;
    }

    /**
     * Provides diagnostic information for this Resource.
     *
     * @return Diagnostics for the Resource
     */
    public Diagnostics getDiagnostics() {
        return diagnostics;
    }

    /**
     * 
     * @param props
     *            The properties for streaming
     * @param commType
     *            The communications type
     * @param numOfThreads
     *            The maximum number of threads that should be spawned
     * @throws CommunicationsException
     *             If there are problems sending messages over the network
     */
    private Resource(Properties props, String commType, int numOfThreads)
    throws CommunicationsException {

        if (instance != null) {
            return;
        }

        streamingService = StreamingService.getInstance();
        streamingService.initialize("config/ServiceConfiguration.txt",
                                    new Random().nextInt(), props, commType);
        retriever = UUIDRetriever.getInstance();

        String deploymentStream = retriever.getRandomBasedUUIDAsString();
        diagnostics = new Diagnostics(deploymentStream, numOfThreads);
        exchangeProcessor =
            new ExchangeProcessor(diagnostics, deploymentStream, numOfThreads);
        exchangeProcessor.initialize();

        instance = this;

        System.out.println("\n\tStarted " + Version.getVersion() + " Resource");
    }

    /**
     * Main entry point for a Granules computational resource.
     *
     * Configuring a resource can be done in three ways:
     *
     * 1. Edit the config/ResourceConfig.txt file
     * 2. Specify an alternate config file as the first argument to the Resource
     * 3. Use System Properties to specify resource parameters.  These can
     *    include the following:
     *        - hostname
     *        - portnum
     *        - transport
     *        - number-of-threads
     *    If System Properties are specified, they override any parameters
     *    specified in the configuration files.
     *
     * @param args
     */
    public static void main(String[] args) {
        ParamsReader paramsReader = ParamsReader.getInstance();
        String configLocation = "config/ResourceConfig.txt";

        if (args.length != 0) {
            configLocation = args[0];
        }

        try {
            Properties resourceProps = new Properties();

            /* Read properties from config file, if it exists. */
            File configFile = new File(configLocation);
            if (configFile.exists()) {
                resourceProps = paramsReader.getProperties(configLocation);
            }

            /* Use System properties, if available. These overwrite values
             * specified in the config file. */
            String[] propertyNames = { "hostname", "portnum",
                                       "transport", "number-of-threads" };

            for (String propertyName : propertyNames) {
                String propertyValue = System.getProperty(propertyName);

                if (propertyValue != null) {
                    resourceProps.setProperty(propertyName, propertyValue);
                }
            }

            /* If execution reaches this point with no hostname and port number
             * for the broker, we cannot continue. */
            if (resourceProps.getProperty("portnum")  == null ||
                resourceProps.getProperty("hostname") == null) {
                System.out.println("Please configure the hostname and port " +
                    "number of the broker to communicate with.");

                return;
            }

            String commType = resourceProps.getProperty("transport", "niotcp");
            int numOfThreads = Integer.parseInt(
                resourceProps.getProperty("number-of-threads", "4"));

            new Resource(resourceProps, commType, numOfThreads);

        } catch (Exception e) {
            String report =
                "Encountered (" + e.getClass().getCanonicalName()
                    + ") exception: " + e.getMessage();
            System.out.println(report);
        }
    }
}
