/**
 * Software License, Version 1.0
 * 
 * Copyright 2008 The Trustees of Indiana University.  All rights reserved.
 * 
 *
 *Redistribution and use in source and binary forms, with or without 
 *modification, are permitted provided that the following conditions are met:
 *
 *1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 *2) All redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 *3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Community Grids Lab. For 
 * further information contact the Community Grids Lab at 
 * http://communitygrids.iu.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 *4) The name Indiana University or Community Grids Lab or Granules, 
 * shall not be used to endorse or promote products derived from this software 
 * without prior written permission from Indiana University.  For written 
 * permission, please contact the Advanced Research and Technology Institute 
 * ("ARTI") at 351 West 10th Street, Indianapolis, Indiana 46202.
 *5) Products derived from this software may not be called Granules, 
 * nor may Indiana University or Community Grids Lab or Granules appear
 * in their name, without prior written permission of ARTI.
 * 
 *
 * Indiana University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Indiana University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */
package ds.granules.scheduler;

import java.io.IOException;

/**
 * This is an interface that snapshots the lifecyle of a task. This interface
 * can be used to retrieve the status of a task, its CPU bound time and the
 * number of times that it has been scheduled for execution.
 * 
 * @author Shrideep Pallickara
 * 
 */
public interface TaskLifecycleMetrics {
	public static final int NOT_SET = 0;
	public static final int INITIALIZING = 1;
	public static final int INITIALIZED = 2;
	public static final int READY = 3;
	public static final int COMPLETE = 4;
	public static final int WAITING_FOR_DATA = 5;
	public static final int EXECUTING = 6;
	public static final int QUEUED_FOR_EXECUTION = 7;
	public static final int ABNORMAL_TERMINATION = 8;
	public static final int ABORTED = 9;

	/**
	 * Retrieve the instance identifier of the operation that was encapsulated
	 * within a Task. This will allow us to determine which taskId was connected
	 * to which instanceId.
	 * 
	 * @return The instance identifier of the operation.
	 */
	public String getInstanceIdentifier();

	/**
	 * Retrieve the number of iterations performed for the operation so far
	 * 
	 * @return The number of times that the computation has been inovoked so
	 *         far.
	 */
	public int getNumberofIterationsSoFar();

	/**
	 * Retrieve the CPU-bound time of the computation.
	 * 
	 * @return The CPU-bound time in milliseconds.
	 */
	public long getCPUBoundTime();

	/**
	 * Retrieve the time at which the operation arrived at this node.
	 * 
	 * @return The arrival time in milliseconds.
	 */
	public long getArrivalTime();

	/**
	 * Retrieve the time this operation was first launched
	 * 
	 * @return The startTime in milliseconds.
	 */
	public long getStartTime();

	/**
	 * Retrieve the time at which this task was completed
	 * 
	 * @return The completion time in milliseconds.
	 */
	public long getCompletionTime();

	/**
	 * Retrieve the time this operation spent residing in memory
	 * 
	 * @return The memory-resident time in milliseconds.
	 */
	public long getTimeResidentInMemory();

	/**
	 * Retrieve the execution status associated with the operation.
	 * 
	 * @return The execution status for the task
	 */
	public int getExecutionStatus();

	/**
	 * Retrieve information about the completion of a task. In the case of
	 * errors and aborts of executions this will include information about why a
	 * task was prematurely terminated.
	 * 
	 * @return The task completion information as a String
	 */
	public String getTaskCompletionInfo();

	/**
	 * Create the serialized representation of this class.
	 * 
	 * @return The byte[] stream for this object.
	 */
	public byte[] getBytes() throws IOException;
}
