/**
 * Software License, Version 1.0
 * 
 * Copyright 2010 Colorado State University.  All rights reserved.
 * 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 *  the list of authors in the original source code, this list of conditions and
 *  the disclaimer listed in this license;
 * 2) All redistributions in binary form must reproduce the above copyright 
 *  notice, this list of conditions and the disclaimer listed in this license in
 *  the documentation and/or other materials provided with the distribution;
 * 3) Any documentation included with all redistributions must include the 
 * following acknowledgment:
 *
 *"This product includes software developed by the Department of Computer Science 
 * at Colorado State University. For further information contact the department at 
 * http://www.cs.colostate.edu/."
 *
 * Alternatively, this acknowledgment may appear in the software itself, and 
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name Colorado State University or Granules shall not be used to endorse
 * or promote products derived from this software without prior written permission
 * from Colorado State University.  For written permission, please contact the 
 * Computer Science Department at 1100 Center Avenue, Fort Collins, CO 80523-1873.
 * 5) Products derived from this software may not be called Granules, 
 * nor may Colorado State University or Granules appear in their name, without prior
 * written permission of ARTI.
 * 
 *
 * Colorado State University provides no reassurances that the source code provided 
 * does not infringe the patent or any other intellectual property rights of 
 * any other entity.  Colorado State University disclaims any liability to any 
 * recipient for claims brought by any other entity based on infringement of 
 * intellectual property rights or otherwise.  
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO 
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF 
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. 
 * COLORADO STATE UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", 
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.  
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR 
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION 
 * GENERATED USING SOFTWARE.
 */

package ds.granules.learning.clustering.fuzzykmeans;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.FileUtil;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.fs.PathFilter;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.io.Writable;
import org.apache.mahout.clustering.ClusterObservations;
import org.apache.mahout.clustering.fuzzykmeans.SoftCluster;
import org.apache.mahout.common.IOUtils;
import org.apache.mahout.common.distance.DistanceMeasure;
import org.apache.mahout.math.Vector;
import org.apache.mahout.math.VectorWritable;

import cgl.narada.event.NBEvent;
import cgl.narada.service.ServiceException;
import ds.granules.dataset.Dataset;
import ds.granules.dataset.DatasetCollection;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.FileAccess;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;


/**
 * A Mapper for the Mahout implementation of Fuzzy K-Means clustering.
 * 
 * @author ericson
 * 
 */
public class FuzzyKMeansMapper extends MapReduceBase {
  //from fuzzykmeans
  private static final double MINIMAL_VALUE = 0.0000000001;
  private double m = 2.0; // default value

  private String datasetIdentifier = "FileDataset-ID-MKMeans";
  private String streamingIdentifier = "StreamingDataset-ID-MKMeans";
  private int WorkerNumber;
  private boolean initialized = false;
  private String initClusters; //filename where initial clusters are located
  double distanceThreshold; //will be entered with list of points/clusters
  DistanceMeasure measure; //will be entered with initialization
  List<Vector> points2Cluster = new ArrayList<Vector>(); //the points it's clustering
  long runTime; //quick and dirty timer...
  int runCount = 0; //loop and gather times
  int maxRun = 20; //how many times to run
  long totalRunTime;
  long compressionTime;
  long decompressionTime;

  private List<SoftCluster> clusters; 

  public FuzzyKMeansMapper() {
    setDomain("Generic");
    setVersionInformation(200);
    setOperationIdentifier("MahoutMapper");
    ResultPayload.getInstance();

  }

  /**
   * Initializes the mapper with values from the deployer.
   */
  private void initializeMapper()
  {
    clusters = new ArrayList<SoftCluster>();
    Properties props = this.getProcessingDirectives();
    String dm = props.getProperty("Distance Measure");
    ClassLoader ccl = Thread.currentThread().getContextClassLoader();
    try {
      measure = ccl.loadClass(dm).asSubclass(DistanceMeasure.class).newInstance();
    } catch (InstantiationException e) {
      e.printStackTrace();
    } catch (IllegalAccessException e) {
      e.printStackTrace();
    } catch (ClassNotFoundException e) {
      e.printStackTrace();
    }
    WorkerNumber = Integer.parseInt(props.getProperty("Worker Number"));
    distanceThreshold = Double.parseDouble(props.getProperty("Distance Threshold"));
    initClusters = props.getProperty("Initial Clusters");
    m = Double.parseDouble(props.getProperty("Fuzziness"));
    loadClusters(initClusters);
    System.out.println("Mapper done initializing!");
    initialized = true;
    runTime = 0;
    compressionTime = 0;
    decompressionTime = 0;
    totalRunTime = System.nanoTime();
  }

  /**
   * Load initial clusters.
   * @param initClusters2
   *             File containing a list of the initial clusters.
   */
  private void loadClusters(String initClusters2) {
    //from KMeansUtil.java: apache.mahout.clustering.kmeans
    // from input file, read in initial clusters
    Configuration conf = new Configuration();
    Path clusterPath = new Path(initClusters2, "*");
    Collection<Path> result = new ArrayList<Path>();

    // filter out the files
    PathFilter clusterFileFilter = new PathFilter() {
      @Override
      public boolean accept(Path path) {
        return path.getName().startsWith("part");
      }
    };

    try {
      // get all filtered file names in result list
      FileSystem fs = clusterPath.getFileSystem(conf);
      FileStatus[] matches = fs.listStatus(FileUtil.stat2Paths(fs.globStatus(clusterPath, clusterFileFilter)), clusterFileFilter);
      for (FileStatus match : matches) {
        result.add(fs.makeQualified(match.getPath()));
      }
      // iterate thru the result path list
      for (Path path : result) {
        SequenceFile.Reader reader = null;
        try {
          reader = new SequenceFile.Reader(path.getFileSystem(conf),path,conf);
          Writable key;
          key = reader.getKeyClass().asSubclass(Writable.class).newInstance();

          SoftCluster value = new SoftCluster();
          while (reader.next(key, value)) {
            // get the cluster info
            clusters.add(value);
            value = new SoftCluster();
          }

          reader.close();
        }
        catch(Exception e)
        {
          System.out.println("Error: " + e.getMessage());
          e.printStackTrace();
        }finally {
          IOUtils.quietClose(reader);}
      }

    } catch (IOException e) {
      throw new IllegalStateException(e);
    }
  }

  public void execute() {
    try {
      DatasetCollection dc = getDatasetCollection();
      if (!initialized)
      {
        System.out.println("Mapper not yet initialized...");
        FileAccess k = null;
        Dataset d = dc.getDataset(datasetIdentifier);
        if(d.getDatasetType() == Dataset.FILES)
          k = (FileAccess)d;
        else
        {
          System.out.println(WorkerNumber + ": Dataset issue: " + d.getDatasetType() + "initialized.  Returning...");
          return;
        }

        Enumeration fnames = k.getFilenames();

        while(fnames.hasMoreElements()){
          String f = (String)fnames.nextElement();
          getPoints(f);
        }

        initializeMapper();
        this.changeToExecuteWhenDataAvailable();

        HashMap<Integer, GranulesObservations> forReducer = DataforReducer(runIteration());
        writeResults(forReducer);
      }

      //now listen for streaming data
      StreamingAccess s = null;
      Dataset t = dc.getDataset(streamingIdentifier);
      if(t.getDatasetType() == Dataset.STREAMS)
        s = (StreamingAccess) t;
      else
      {
        System.out.println(WorkerNumber + ": Dataset issue: " + t.getDatasetType() + "initialized.  Returning...");
        return;
      }

      while (s.isDataAvailable()) {
        NBEvent nbEvent = s.getStreamingData();
        processResultsFromMap(nbEvent);
      }

    } catch (DatasetException e) {
      e.printStackTrace();
    } catch (IOException e) {
      System.out.println("Problems converting bytes to longs");
      e.printStackTrace();
    } catch (MarshallingException e) {
      System.out.println("Problems Marshalling: " + e.getMessage());
      e.printStackTrace();
    } catch (MapReduceException e) {
      System.out.println("Problems writing Results: " + e.getMessage());
      e.printStackTrace();
    } catch (NumberFormatException e) {
      e.printStackTrace();
    } catch (ServiceException e) {
      e.printStackTrace();
    }

  }

  /**
   * Loads points to cluster
   * @param f
   *             File containing points to cluster
   */
  private void getPoints(String f) {
    try {
      Configuration c = new Configuration();
      Path p = new Path(f);
      SequenceFile.Reader reader = new SequenceFile.Reader(p.getFileSystem(c),p,c);
      Writable key;
      key = (Writable)reader.getKeyClass().newInstance();
      VectorWritable value = new VectorWritable();
      while (reader.next(key, value)) {
        // get the point info
        points2Cluster.add(value.get());
        value = new VectorWritable();
      }
      reader.close();

    } catch (Exception e) {
      System.out.println("Problem reading in vector file");
      e.printStackTrace();
    }
  }

  private void startRun(NBEvent nbEvent) throws IOException, MarshallingException, NumberFormatException, ServiceException {
    if (!initialized)
      initializeMapper();
    byte[] data = nbEvent.getContentPayload();
    Results results = ResultsFactory.getInstance().getResults(data);
    byte[] resultPayloadBytes = results.getResultPayload();
    int originalSize = Integer.parseInt(results.getRespondTo());
    byte[] decompressed;
    if(resultPayloadBytes.length != originalSize) {
      decompressed = decompressPayload(resultPayloadBytes, originalSize);
    }
    else {
      decompressed = resultPayloadBytes;
    }
    ResultPayload rp = ResultPayload.getInstance();

    Object[] objArr = rp.getObjectArray(decompressed);

    clusters.clear();
    for(Object c : objArr)
      clusters.add(((GranulesSoftCluster)c).getCluster());
  }



  private synchronized void processResultsFromMap(NBEvent nbEvent)
  throws IOException, MarshallingException, MapReduceException, NumberFormatException, ServiceException {
    String synopsis = (String) nbEvent.getContentSynopsis();
    if (synopsis.startsWith("Initialize")||synopsis.startsWith("ClusterInfo"))	//contains clusters, maxIterations and distance measure,
    {
      startRun(nbEvent);
      HashMap<Integer, GranulesObservations> forReducer = DataforReducer(runIteration());
      writeResults(forReducer);
    }
    else //assume termination
    {
      //Code to handle timing
			long tmpTime = System.nanoTime() - totalRunTime;
			FileWriter fw = new FileWriter(new File("worker" + WorkerNumber + ".timing"), true);
			fw.write(runTime + "\t" + compressionTime + "\t" + decompressionTime + "\n");
			fw.flush();
			fw.close();
			if(WorkerNumber == 0) {
				fw = new FileWriter(new File("overall_fuzzy.timing"),true);
				fw.write(tmpTime + "\n");
				fw.flush();
				fw.close();
			}
      runTime = 0;

      if(runCount < maxRun)
      {//start again!
        initialized = false;
        clusters = null;
        System.gc();
        runCount++;
        initializeMapper();
        System.out.println("starting iteration on " + points2Cluster.size() + " points...");
        totalRunTime = System.nanoTime();
        HashMap<Integer, GranulesObservations> forReducer = DataforReducer(runIteration());
        System.out.println("Sending results to reducer");
        writeResults(forReducer);
      }
    }
  }

  /**
   * Perform an iteration of clustering
   * @return
   *             <code>List</code> of probabilities of every point belonging to every cluster
   */
  private List<SoftCluster> runIteration()
  {	
    long startTime = System.nanoTime();
    List<SoftCluster> results = new ArrayList<SoftCluster>();
    results.addAll(clusters);
    for(Vector point : points2Cluster)
    {
      List<Double> clusterDistanceList = new ArrayList<Double>();
      for(SoftCluster cluster : results)
      {
        double distance = measure.distance(cluster.getCenter(), point);
        clusterDistanceList.add(distance);
      }

      // clustering code from the Mahout implementation:fuzzykmeans/FuzzyKMeansClusterer.java
      for (int i = 0; i < results.size(); i++) {
        SoftCluster cluster = results.get(i);
        ClusterObservations value =
          new ClusterObservations(computeProbWeight(clusterDistanceList.get(i), clusterDistanceList,i),
              point,
              point.times(point));
        cluster.observe(value);
      }
    }
    runTime += System.nanoTime() - startTime;
    return results;
  }

  /** Computes the probability of a point belonging to a cluster */
  public double computeProbWeight(double clusterDistance, List<Double> clusterDistanceList, int start) {
    if (clusterDistance == 0) {
      clusterDistance = MINIMAL_VALUE;
    }
    double denom = 0.0;
    for (int i = 0; i < clusterDistanceList.size(); i++) {
      if(i == start)
        continue;
      double eachCDist = clusterDistanceList.get(i);
      if (eachCDist == 0.0) {
        eachCDist = MINIMAL_VALUE;
      }
      denom += Math.pow(clusterDistance / eachCDist, 2.0 / (m - 1));
    }
    return 1.0 / denom;
  }

  /**
   * Assembles clustering results for the reducer
   * @param list
   *             <code>List</code> of probabilities of every point belonging to every cluster
   * @return
   *             <code>HashMap</code> containing observations for every cluster
   */
  private HashMap<Integer, GranulesObservations> DataforReducer(List<SoftCluster> list)
  {
    System.out.println("in DataforReducer");
    HashMap<Integer, GranulesObservations> results = new HashMap<Integer, GranulesObservations>();
    for (SoftCluster c : list)
    {
      ClusterObservations o = c.getObservations();
      results.put(c.getId(), new GranulesObservations(o));
    }
    System.out.println("returning results");
    return results;
  }

  /**
   * Sends results to the reducer
   * @param map
   *             <code>HashMap</code> generated in <code>DataforReducer</code> function
   * @throws MarshallingException
   *             Problems marshalling the map
   * @throws MapReduceException
   *             Problems sending the data to the reducer
   */
  private void writeResults(HashMap<Integer, GranulesObservations> map)
  throws MarshallingException, MapReduceException {
    boolean completeResult = true;
    ResultPayload resultPayload = ResultPayload.getInstance();

    byte[] resultPayloadBytes = resultPayload.getBytes(map);
    byte[] compressedBytes = compressPayload(resultPayloadBytes);

    Results results = createResults(completeResult, false);

    results.setResultPayload(compressedBytes);
    results.setRespondTo(Integer.toString(resultPayloadBytes.length));

    results.setResultType(ResultPayload.HASHMAP);

    writeResults(results);
  }

  // Code to handle compressing dat a for transmission between the maps and reduces
  private byte[] compressPayload(byte[] payload) {
    long startTime = System.nanoTime();
    Deflater compresser = new Deflater(Deflater.BEST_SPEED);
    compresser.setInput(payload);
    compresser.finish();
    int payloadLength = payload.length;
    System.out.println("Payload length=" + payloadLength);

    byte[] output = new byte[payloadLength];
    int compressedDataLength = compresser.deflate(output);
    System.out.println("compressedDataLength=" + compressedDataLength);

    if (compressedDataLength == payloadLength) {
      return payload;
    }

    byte[] compressedBytes = new byte[compressedDataLength];
    System.arraycopy(output, 0, compressedBytes, 0, compressedDataLength);
    compressionTime += System.nanoTime() - startTime;
    return compressedBytes;
  }

  private byte[] decompressPayload(byte[] compressedPayload, int originalLength) throws ServiceException {
    long startTime = System.nanoTime();
    Inflater decompresser = new Inflater();
    decompresser.setInput(compressedPayload);
    byte[] originalPayload = new byte[originalLength];
    try { 
      int resultLength = decompresser.inflate(originalPayload);
    } catch (DataFormatException dfe) {
      throw new ServiceException("Reducer Decompression failed => " +
          dfe.toString());
    }
    decompressionTime += System.nanoTime() - startTime;
    return originalPayload;
  }

}
