/**
 * Software License, Version 1.0 Copyright 2010 Colorado State University. All
 * rights reserved. Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following conditions
 * are met: 1) All redistributions of source code must retain the above
 * copyright notice, the list of authors in the original source code, this list
 * of conditions and the disclaimer listed in this license; 2) All
 * redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the disclaimer listed in this license in the
 * documentation and/or other materials provided with the distribution; 3) Any
 * documentation included with all redistributions must include the following
 * acknowledgment: "This product includes software developed by the Department
 * of Computer Science at Colorado State University. For further information
 * contact the department at http://www.cs.colostate.edu/." Alternatively, this
 * acknowledgment may appear in the software itself, and wherever such
 * third-party acknowledgments normally appear. 4) The name Colorado State
 * University or Granules shall not be used to endorse or promote products
 * derived from this software without prior written permission from Colorado
 * State University. For written permission, please contact the Computer Science
 * Department at 1100 Center Avenue, Fort Collins, CO 80523-1873. 5) Products
 * derived from this software may not be called Granules, nor may Colorado State
 * University or Granules appear in their name, without prior written permission
 * of ARTI. Colorado State University provides no reassurances that the source
 * code provided does not infringe the patent or any other intellectual property
 * rights of any other entity. Colorado State University disclaims any liability
 * to any recipient for claims brought by any other entity based on infringement
 * of intellectual property rights or otherwise. LICENSEE UNDERSTANDS THAT
 * SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO WARRANTIES AS TO CAPABILITIES OR
 * ACCURACY ARE MADE. COLORADO STATE UNIVERSITY GIVES NO WARRANTIES AND MAKES NO
 * REPRESENTATION THAT SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT,
 * COPYRIGHT, OR OTHER PROPRIETARY RIGHTS. COLORADO STATE UNIVERSITY MAKES NO
 * WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES",
 * "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE. LICENSEE ASSUMES THE ENTIRE
 * RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS, AND TO
 * THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING SOFTWARE.
 */
package ds.granules.learning.clustering.fuzzykmeans;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.FileUtil;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.fs.PathFilter;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.mahout.clustering.fuzzykmeans.SoftCluster;
import org.apache.mahout.common.IOUtils;
import org.apache.mahout.common.distance.DistanceMeasure;

import cgl.narada.event.NBEvent;
import cgl.narada.service.ServiceException;
import ds.granules.dataset.DatasetException;
import ds.granules.dataset.StreamingAccess;
import ds.granules.exception.MarshallingException;
import ds.granules.operation.MapReduceBase;
import ds.granules.operation.MapReduceException;
import ds.granules.results.ResultPayload;
import ds.granules.results.Results;
import ds.granules.results.ResultsFactory;

/**
 * The reducer for Mahout's Fuzzy K-Means using Granules for communications
 * 
 * @author ericson
 */
public class FuzzyKMeansReducer extends MapReduceBase {
  private int iterations = 0;
  private int loop = 0;
  private int MaxIters;
  String workerType = "Reducer ";
  private DistanceMeasure measure;
  private double convergenceDelta;
  boolean initialized;
  HashMap<Integer, SoftCluster> currentClusters; // hashmap of clusterID,
  // cluster
  boolean converged;
  String clusterFile;
  String synopsis = "ClusterInfo";
  long runTime;
  long decompressionTime;
  long compressionTime;
  //fault tolerance
  boolean isFaultTolerant = false;
  int resultsCount = 0;

  public FuzzyKMeansReducer() {
    setDomain("Generic");
    setVersionInformation(200);
    setOperationIdentifier("KMeansMahoutReducer");
    initialized = false;
  }

  private void initialize() {
    System.out.println("in initialize()");
    runTime = 0;
    iterations = 0;
    compressionTime = 0;
    decompressionTime = 0;
    converged = false;
    Properties p = this.getProcessingDirectives();
    // get data for measure and convergence delta:
    // measure = ....;
    // convergenceDelta = ...;
    currentClusters = new HashMap<Integer, SoftCluster>();
    String dm = p.getProperty("Distance Measure");
    ClassLoader ccl = Thread.currentThread().getContextClassLoader();
    try {
      measure = ccl.loadClass(dm).asSubclass(DistanceMeasure.class)
      .newInstance();
    } catch (InstantiationException e) {
      e.printStackTrace();
    } catch (IllegalAccessException e) {
      e.printStackTrace();
    } catch (ClassNotFoundException e) {
      e.printStackTrace();
    }
    MaxIters = Integer.parseInt(p.getProperty("Max Iterations"));
    convergenceDelta = Double.parseDouble(p.getProperty("Convergence Delta"));
    clusterFile = p.getProperty("Initial Clusters");
    loadInitialClusters(clusterFile);
    initialized = true;
    System.out.println("Done initialize!");
  }

  /**
   * Loads initial clusters from file
   * 
   * @param clusterFile2
   *          File containing initial clusters
   */
  private void loadInitialClusters(String clusterFile2) {
    Collection<SoftCluster> clusters = new ArrayList<SoftCluster>();
    Configuration conf = new Configuration();
    Path clusterPath = new Path(clusterFile2, "*");
    Collection<Path> result = new ArrayList<Path>();

    // filter out the files
    PathFilter clusterFileFilter = new PathFilter() {
      @Override
      public boolean accept(Path path) {
        return path.getName().startsWith("part");
      }
    };

    try {
      // get all filtered file names in result list
      FileSystem fs = clusterPath.getFileSystem(conf);
      FileStatus[] matches = fs.listStatus(FileUtil.stat2Paths(fs.globStatus(
          clusterPath, clusterFileFilter)), clusterFileFilter);

      for (FileStatus match : matches) {
        result.add(fs.makeQualified(match.getPath()));
      }

      // iterate thru the result path list
      for (Path path : result) {
        // Check if we are restarting partway through: if the input file matches part-r-***, *** is the iteration we stopped on
        String s = path.getName();
        String[] broken = s.split("-");
        try {
          int tmp = Integer.parseInt(broken[broken.length-1]);
          iterations = tmp;
        }
        catch (NumberFormatException e) {
          // Not formatted for a reload, continue on
        }
        SequenceFile.Reader reader = null;
        try {
          reader = new SequenceFile.Reader(fs, path, conf);
          Class<?> valueClass = reader.getValueClass();
          Writable key;
          try {
            key = reader.getKeyClass().asSubclass(Writable.class).newInstance();
          } catch (InstantiationException e) { // Should not be possible
            throw new IllegalStateException(e);
          } catch (IllegalAccessException e) {
            throw new IllegalStateException(e);
          }
          if (valueClass.equals(SoftCluster.class)) {
            SoftCluster value = new SoftCluster();
            while (reader.next(key, value)) {
              // get the cluster info
              clusters.add(value);
              value = new SoftCluster();
            }
          }
        } finally {
          IOUtils.quietClose(reader);
        }
      }

    } catch (IOException e) {
      throw new IllegalStateException(e);
    }

    // Easier storage for Granules implementation
    for (SoftCluster c : clusters) {
      currentClusters.put(c.getId(), c);
    }

  }

  public void execute() {
    try {
      System.out.println("in execute()");
      if (!initialized)
        initialize();
      StreamingAccess mapReduceDataset = getMapReduceDataset();

      while (mapReduceDataset.isDataAvailable()) {
        NBEvent nbEvent = mapReduceDataset.getStreamingData();
        processResultsFromMap(nbEvent);
      }

    } catch (DatasetException e) {
      e.printStackTrace();
    } catch (IOException e) {
      System.out.println("Problems converting bytes to longs");
      e.printStackTrace();
    } catch (MarshallingException e) {
      System.out.println("Problems Marshalling: " + e.getMessage());
      e.printStackTrace();
    } catch (MapReduceException e) {
      System.out.println("Problems writing Results: " + e.getMessage());
      e.printStackTrace();
    } catch (NumberFormatException e) {
      e.printStackTrace();
    } catch (ServiceException e) {
      e.printStackTrace();
    }

  }

  private synchronized void processResultsFromMap(NBEvent nbEvent)
  throws IOException, MarshallingException, MapReduceException,
  NumberFormatException, ServiceException {
    if (!initialized)
      initialize();
    System.out.println("getting inputs");
    byte[] resultBytes = nbEvent.getContentPayload();
    Results results = ResultsFactory.getInstance().getResults(resultBytes);

    byte[] resultPayloadBytes = results.getResultPayload();

    int originalSize = Integer.parseInt(results.getRespondTo());
    byte[] decompressedBytes;
    if(resultPayloadBytes.length != originalSize) {
      decompressedBytes = decompressPayload(resultPayloadBytes, Integer
          .parseInt(results.getRespondTo()));
    }
    else {
      decompressedBytes = resultPayloadBytes;
    }

    ResultPayload resultPayload = ResultPayload.getInstance();
    System.out.println("Done getting inputs!");

    if (results.getResultType() != ResultPayload.HASHMAP) {
      System.out
      .println("Mahout KMeans Reducer: Incorrect ResultType received");
      return;
    }

    //System.out.println("Got a result!!");
    this.resultsCount++;
    /**
     * Remainder of processing assumes that the resultPayload type is of type
     * ResultPayload.INT
     */
    long startTime = System.nanoTime();
    HashMap<Integer, GranulesObservations> incomingMap = (HashMap<Integer, GranulesObservations>) resultPayload
    .getHashMap(decompressedBytes);


    setReceivedOutputFrom(results.getInstanceIdentifier());

    handleIncoming(incomingMap);
    runTime += System.nanoTime() - startTime;
    System.out.println("Have processed " + resultsCount + " results");

    if (!isAwaitingOutputs()) {
      resetOutputTracker();
      iterations++;
      resultsCount=0;

      List<SoftCluster> clusters = new ArrayList<SoftCluster>();
      for (SoftCluster c : currentClusters.values()) {
        clusters.add(c);
      }
      startTime = System.nanoTime();
      List<SoftCluster> newClusters = calcNewClusters(clusters, converged);
      // For comparison, comment converge check out -- When testing Mahout Fuzzy
      // K-Means code was broken
      // and didn't check for convergence.
      // System.out.println("testing convergence");
      // converged = testConvergence(newClusters);
      //System.out.println("Done w convergence check");
      currentClusters.clear();
      updateMap(newClusters);
      runTime += System.nanoTime() - startTime;
      if (!converged && iterations < MaxIters) {
        System.out.println("sending stuff to mappers");
        writeResult(newClusters, converged);
        System.out.println("all done");
        if(isFaultTolerant) {
          writeToFile();
        }
      }

      else {

        long tmp = System.nanoTime();
        long fileTime = System.nanoTime() - tmp;
        System.out.println("Done with a round of execution");
        // start again
        initialized = false;

        // Code to collect timing. 
        FileWriter fw = new FileWriter(new File("/tmp/ericson/Reducer.timing"), true);
        fw.write(runTime + "\t" +
            compressionTime + "\t" + decompressionTime + "\t" + fileTime + "\n");
        fw.flush(); fw.close();

        byte[] rpb = resultPayload.getBytes("Done");

        Results r = createResults(true, false);
        r.setResultPayload(rpb);
        r.setResultType(ResultPayload.STRING);

        // writeResults(synopsis,results);
        this.writeResults("Termination", r);
        System.out.println("sending termination");
        writeToFile();
        currentClusters = null;
        System.gc();
        System.out.println("Done @ reducer");
      }

    }

  }

  /**
   * Writes final clusters to file
   */
  private void writeToFile() {
    Configuration c = new Configuration();
    Path p = new Path("part-r-" + iterations + "-" + loop + "_"
        + System.currentTimeMillis());
    SequenceFile.Writer writer;
    try {
      writer = new SequenceFile.Writer(p.getFileSystem(c), c, p, Text.class,
          SoftCluster.class);
      Writable w;
      for (SoftCluster clus : currentClusters.values()) {
        w = new Text(clus.getIdentifier());
        writer.append(w, clus);
      }

      writer.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  private void updateMap(List<SoftCluster> newClusters) {
    for (SoftCluster c : newClusters) {
      currentClusters.put(c.getId(), c);
    }
  }

  private void writeResult(List<SoftCluster> data, boolean processingComplete)
  throws MarshallingException, MapReduceException, IOException {
    List<GranulesSoftCluster> sending = new ArrayList<GranulesSoftCluster>();
    for (SoftCluster c : data)
      sending.add(new GranulesSoftCluster(c));
    ResultPayload resultPayload = ResultPayload.getInstance();
    byte[] resultPayloadBytes = resultPayload.getBytes(sending.toArray());
    byte[] compressedBytes = compressPayload(resultPayloadBytes);

    Results results = createResults(true, processingComplete);
    results.setResultPayload(compressedBytes);
    results.setRespondTo(Integer.toString(resultPayloadBytes.length));
    results.setResultType(ResultPayload.OBJECT_ARRAY);

    writeResults(synopsis, results);
  }

  private List<SoftCluster> calcNewClusters(List<SoftCluster> oldClusters,
      boolean done) {
    int convergeCount = 0;
    List<SoftCluster> newClusters = new ArrayList<SoftCluster>();
    for (SoftCluster c : oldClusters) {
      boolean converged = computeConvergence(c);
      if (converged) {
        convergeCount++;
      }
      c.computeParameters();
      newClusters.add(c);
    }
    System.out.println(convergeCount + " Clusters have converged");
    if (convergeCount < newClusters.size())
      done = false;
    return newClusters;
  }

  private void handleIncoming(HashMap<Integer, GranulesObservations> data) {
    for (int key : data.keySet()) {
      SoftCluster t = currentClusters.get(key);
      if (t.equals(null))
        System.out.println("Error: can't find cluster with id: " + key);
      else {
        t.observe(data.get(key).toClusterObservations());
        currentClusters.put(key, t);
      }
    }
  }

  private boolean testConvergence(List<SoftCluster> newClusters) {
    boolean done = true;
    for (SoftCluster c : newClusters) {
      if (!c.isConverged())
        return false;
    }
    return done;
  }

  private boolean computeConvergence(SoftCluster c) {
    return c.computeConvergence(measure, convergenceDelta);
  }

  // General code to handle compressing/decompressing data
  private byte[] compressPayload(byte[] payload) {
    long startTime = System.nanoTime();
    Deflater compresser = new Deflater(Deflater.BEST_SPEED);
    compresser.setInput(payload);
    compresser.finish();
    int payloadLength = payload.length;
    System.out.println("Payload length=" + payloadLength);

    byte[] output = new byte[payloadLength];
    int compressedDataLength = compresser.deflate(output);
    System.out.println("compressedDataLength=" + compressedDataLength);

    if (compressedDataLength == payloadLength) {
      return payload;
    }

    byte[] compressedBytes = new byte[compressedDataLength];
    System.arraycopy(output, 0, compressedBytes, 0, compressedDataLength);
    compressionTime += System.nanoTime() - startTime;
    return compressedBytes;
  }

  private byte[] decompressPayload(byte[] compressedPayload, int originalLength)
  throws ServiceException {
    long startTime = System.nanoTime();
    Inflater decompresser = new Inflater();
    decompresser.setInput(compressedPayload);
    System.out.println("about to decompress");
    byte[] originalPayload = new byte[originalLength];
    try {
      int resultLength = decompresser.inflate(originalPayload);
    } catch (DataFormatException dfe) {
      System.out.println("DataFormatException!!!\n" + dfe.getLocalizedMessage());
      throw new ServiceException("Reducer Decompression failed => "
          + dfe.toString());
    }
    decompressionTime += System.nanoTime() - startTime;
    System.out.println("done!");
    return originalPayload;
  }

}
